/*-------------------------------------------------------------------
 * drivers/block/hd.c
 *-------------------------------------------------------------------*/
#include <weix/config.h>
#include <weix/sched.h>
#include <weix/fs.h>
#include <weix/kernel.h>
#include <weix/hdd_struct.h>
#include <weix/hdreg.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <weix/signal.h>
#include <weix/types.h>
#include <weix/hdd_struct.h>
#include <weix/kdebug.h>

#define MAJOR_NR 3
#include <weix/blk.h>

#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
static void recal_intr(void);
static void bad_rw_intr(void);
void unexpected_hd_interrupt(void);
static void hd_interrupt(int unused);

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
static int current_minor = 0;
static int recalibrate = 0;
static int reset = 0;

#ifdef HD_TYPE
static int NR_HD = 1;
#else
static int NR_HD = 0;
#endif

static int hd_sizes[5*MAX_HD] = {0, };

/* request_irq()   üԴϴ. */
static struct sigaction hd_sigaction = {
    hd_interrupt,
    0,
    SA_INTERRUPT,
    NULL
};

static struct hd_driveid hd_ident_info[MAX_HD] = {{0,}, {0,}, };

/*-------------------------------------------------------------------*
 * CMOS_READ
 *-------------------------------------------------------------------*/
static inline unsigned char CMOS_READ(unsigned char addr)
{
    outb_p(addr, 0x70);
    return inb_p(0x71);
}

/*-------------------------------------------------------------------*
 * hd_interrupt
 *-------------------------------------------------------------------*/
static void hd_interrupt(int unused)
{
    void (*handler)(void) = DEVICE_INTR;

    DEVICE_INTR = NULL;
    if (!handler)
        handler = unexpected_hd_interrupt;
    handler();
    sti();
}

/*-------------------------------------------------------------------*
 * sys_setup
 *-------------------------------------------------------------------*/
int sys_setup()
{
	static int callable = 1;
	int i, a, drive=0;
	int mask = (1 << 6) - 1;
#ifndef HD_TYPE
	int heads;
	unsigned char cmos_disks;
#endif
	unsigned long first_sector;
	struct partition *p;
	struct buffer_head * bh;

	kprintf("sys_setup: in sys_setup\n");
	if (!callable)
		return -1;
	callable = 0;

#ifndef HD_TYPE
    if((cmos_disks = CMOS_READ(0x12))) {
		kprintf("HDC: first IDE hard disk: ");
		if(cmos_disks & 0xf0) {
			current_gover->hd_info[0].cyl 	= *(unsigned short *)0xC20;
			current_gover->hd_info[0].head = *(unsigned char  *)0xC22;
			current_gover->hd_info[0].wpcom= *(unsigned short *)0xC25;
			current_gover->hd_info[0].dso	= *(unsigned char  *)0xC28;
			current_gover->hd_info[0].sect = *(unsigned char  *)0xC2E;
			kprintf("type %d", ((cmos_disks & 0xf0) == 0xf0) ? CMOS_READ(0x19) : (cmos_disks >> 4));
			kprintf("(%d/%d/%d)", current_gover->hd_info[0].cyl, current_gover->hd_info[0].head, current_gover->hd_info[0].sect);

			if((current_gover->hd_info[0].head > 16) && current_gover->hd_info[0].head % 2)
				kprintf("HDC: WARNING: invalid number of heads, disabled\n");
			else {
				for(heads = current_gover->hd_info[0].head ; heads > 16; heads /= 2)
					current_gover->hd_info[0].lbaf = current_gover->hd_info[0].head / heads;

				if(current_gover->hd_info[0].lbaf > 1)
					kprintf(" LBA (geometry: %d/%d/%d)\n",
						current_gover->hd_info[0].cyl 		* current_gover->hd_info[0].lbaf,
						current_gover->hd_info[0].head		/ current_gover->hd_info[0].lbaf,
						current_gover->hd_info[0].sect);
				else
					kprintf("\n");
			NR_HD++;
			}
		} else
			kprintf("not installed\n");

		if(cmos_disks & 0x0f) {
			current_gover->hd_info[1].cyl  = *(unsigned short *)0xC30;
			current_gover->hd_info[1].head = *(unsigned char  *)0xC32;
			current_gover->hd_info[1].wpcom= *(unsigned short *)0xC35;
			current_gover->hd_info[1].dso  = *(unsigned char  *)0xC28;
			current_gover->hd_info[1].sect = *(unsigned char  *)0xC3E;
			kprintf("type %d", ((cmos_disks & 0x0f) == 0x0f) ? CMOS_READ(0x1a) : (cmos_disks & 0x0f));
			kprintf("(%d/%d/%d)", current_gover->hd_info[1].cyl, current_gover->hd_info[1].head, current_gover->hd_info[1].sect);
            if((current_gover->hd_info[1].head > 16) && current_gover->hd_info[1].head % 2)
                kprintf("HDC: WARNING: invalid number of heads, disabled\n");
            else {
                for(heads = current_gover->hd_info[1].head ; heads > 16; heads /= 2)
                    current_gover->hd_info[1].lbaf = current_gover->hd_info[1].head / heads;

                if(current_gover->hd_info[1].lbaf > 1)
                    kprintf(" LBA (geometry: %d/%d/%d)\n",
                        current_gover->hd_info[1].cyl      * current_gover->hd_info[0].lbaf,
                        current_gover->hd_info[1].head     / current_gover->hd_info[0].lbaf,
                        current_gover->hd_info[1].sect);
                else
                    kprintf("\n");
			NR_HD++;
            }
        } else
            kprintf("not installed\n");
	}
#endif

    if (NR_HD) {
        if (irqaction(HD_IRQ,&hd_sigaction)) {
            kprintf("hd.c: unable to get IRQ%d for the harddisk driver\n",HD_IRQ);
            NR_HD = 0;
        }
    }

#if 0
    for(drive=0; drive<NR_HD; drive++) {
        kprintf("hd_info[%d].cyl=%d, hd_info[%d].head=%d\n", drive, current_gover->hd_info[drive].cyl, drive, current_gover->hd_info[drive].head);
        kprintf("hd_info[%d].wpcom=%d, hd_info[%d].ctl=%d\n", drive, current_gover->hd_info[drive].wpcom, drive, current_gover->hd_info[drive].ctl);
        kprintf("hd_info[%d].lzone=%d, hd_info[%d].sect=%d\n", drive, current_gover->hd_info[drive].lzone, drive, current_gover->hd_info[drive].sect);
    }
#endif

	/* current_gover->hd_info[0]  ù° ϵ  ŵϴ.
	 * current_gover->hd_info[1]  ι° ϵ  ŵϴ.
	 * current_gover->hd_info[].u.hd_part[0]  sector    ˴ϴ. */
	for (i=0 ; i<NR_HD ; i++) {
		current_gover->hd_info[i].u.hd_part[0].start_sect = 0;
		current_gover->hd_info[i].u.hd_part[0].nr_sects = current_gover->hd_info[i].head * current_gover->hd_info[i].sect * current_gover->hd_info[i].cyl;
		current_gover->hd_info[i].u.hd_part[0].special_op = 1;
		current_gover->hd_info[i].u.hd_part[0].identified = 0;
	}

	/*  ׸ ؼ 0 մϴ. */
	for (i = NR_HD ; i < 2 ; i++) {
		current_gover->hd_info[i].u.hd_part[0].start_sect = 0;
		current_gover->hd_info[i].u.hd_part[0].nr_sects = 0;
	}

	/* ó ׸ ´ ˻մϴ. */
	first_sector = current_gover->hd_info[0].u.hd_part[0].start_sect;
    if ((int)first_sector == -1) {
        current_gover->hd_info[0].u.hd_part[0].start_sect = 0;
        return -1;
    }

	for (drive=0 ; drive<NR_HD ; drive++) {
		/* ϵ  Super Block  о Դϴ. */
		if (!(bh = block_read(0x300 + drive*5, 0, BLOCK_SIZE))) {
			panic("Unable to read partition table of drive %d\n\r",
				drive);
		}
		if (*(unsigned short *)(bh->b_data+510) == 0xAA55) {
			p = (struct partition *) (0x1BE + bh->b_data);
			for (i=1;i<=4;i++,p++) {
				if(!p->nr_sects)
					continue;
				current_gover->hd_info[drive].u.hd_part[i].start_sect = first_sector+p->start_sect;
				current_gover->hd_info[drive].u.hd_part[i].nr_sects = p->nr_sects;
				if ((current_minor & 0x3f) >= 60)
                	continue;
				if (p->sys_ind == EXTENDED_PARTITION) {
					kprintf(" < FIX ME T_T >");
				}
				kprintf("hd%d ", i);
        		if (*(unsigned short *) (bh->b_data+0xfc) == 0x55AA) {
            		p = (struct partition *) (0x1BE + bh->b_data);
            		for (i = 4 ; i < 16 ; i++, current_minor++) {
                		p--;
                		if ((current_minor & mask) >= mask-2)
                   			break;
                		if (!(p->start_sect && p->nr_sects))
                    		continue;
						kprintf("Disk Manager partition talbe\n");
            		}
        		}
			}
		} else
			kprintf(" Bad partition table");
		kprintf("\n");
		brelse(bh);
	}
#if	0
	for(i=0; i<5; i++) {
        kprintf("hd[i].start_sect = %d, ", hd[i].start_sect);
        kprintf("hd[i].nr_sects = %d\n", hd[i].nr_sects);
	}
#endif

	for(a=0 ; a<MAX_HD ; a++)
		for (i=0 ; i<5 ; i++)
			hd_sizes[i] = current_gover->hd_info[0].u.hd_part[i].nr_sects>>1 ;

	block_dev_size[MAJOR_NR] = hd_sizes;

	mount_root();
	return (0);
}

/*-------------------------------------------------------------------*
 * controller_ready
 *-------------------------------------------------------------------*/
static int controller_ready(void)
{
	int retries = 100000;

	while (--retries && (inb_p(HD_STATUS)&0xc0)!=0x40);
	return (retries);
}

/*-------------------------------------------------------------------*
 * win_result
 *-------------------------------------------------------------------*/
static int win_result(void)
{
	int i=inb_p(HD_STATUS);

	if ((i & (BUSY_STAT | READY_STAT | WRERR_STAT | SEEK_STAT | ERR_STAT))
		== (READY_STAT | SEEK_STAT))
		return(0); /* ok */
	if (i&1) i=inb(HD_ERROR);
	return (1);
}

/*-------------------------------------------------------------------*
 * hd_out
 *-------------------------------------------------------------------*/
static void hd_out(unsigned int drive,unsigned int nsect,unsigned int sect,
		unsigned int head,unsigned int cyl,unsigned int cmd,
		void (*intr_addr)(void))
{
	unsigned short port;

#if 0
	kprintf("drive=%d, nsect=%d, sect=%d, head=%d\n", drive, nsect, sect, head);
	kprintf("cyl=%d, cmd=%d, intr_addr=%08X\n", cyl, cmd, intr_addr);
#endif
	if (drive>1 || head>15)
		panic("Trying to write bad sector");
	if (!controller_ready()) {
		reset = 1;
		return;
	}

	SET_INTR(intr_addr);

	outb_p(HD_CMD, current_gover->hd_info[drive].ctl);
	port=HD_DATA;
	outb_p(++port, current_gover->hd_info[drive].wpcom>>2);
	outb_p(++port, nsect);
	outb_p(++port, sect);
	outb_p(++port, cyl);
	outb_p(++port, cyl>>8);
	outb_p(++port, 0xA0|(drive<<4)|head);
	outb(++port, cmd);
}

/*-------------------------------------------------------------------
 * drive_busy
 *-------------------------------------------------------------------*/
static int drive_busy(void)
{
	unsigned int i;
	unsigned char c;

	for (i = 0; i < 50000; i++) {
		c = inb_p(HD_STATUS);
		c &= (BUSY_STAT | READY_STAT | SEEK_STAT);
		if (c == (READY_STAT | SEEK_STAT))
			return 0;
	}
	kprintf("HD controller times out\n\r");
	return(1);
}

/*-------------------------------------------------------------------*
 * reset_controller
 *-------------------------------------------------------------------*/
static void reset_controller(void)
{
	int	i;

	outb(HD_CMD, 4);
	for(i = 0; i < 1000; i++) nop();
	outb(HD_CMD, current_gover->hd_info[0].ctl & 0x0f );
	if (drive_busy())
		kprintf("HD-controller still busy\n\r");
	if ((i = inb(HD_ERROR)) != 1)
		kprintf("HD-controller reset failed: %02x\n\r",i);
}

/*-------------------------------------------------------------------*
 * reset_hd
 *-------------------------------------------------------------------*/
static void reset_hd(void)
{
	static int i;

repeat:
	if (reset) {
		reset = 0;
		i = -1;
		reset_controller();
	} else if (win_result()) {
		bad_rw_intr();
		if (reset)
			goto repeat;
	}
	i++;
	if (i < NR_HD) {
		hd_out(i,current_gover->hd_info[i].sect,current_gover->hd_info[i].sect,current_gover->hd_info[i].head-1,
			current_gover->hd_info[i].cyl,WIN_SPECIFY,&reset_hd);
	} else
		do_hd_request();
}

/*-------------------------------------------------------------------*
 * unexpected_hd_interrupt
 *-------------------------------------------------------------------*/
void unexpected_hd_interrupt(void)
{
	kprintf("Unexpected HD interrupt\n\r");
	reset = 1;
	do_hd_request();
}

/*-------------------------------------------------------------------*
 * bad_rw_intr
 *-------------------------------------------------------------------*/
static void bad_rw_intr(void)
{
	if (++CURRENT->errors >= MAX_ERRORS)
		end_request(0);
	if (CURRENT->errors > MAX_ERRORS/2)
		reset = 1;
}

/*-------------------------------------------------------------------*
 * fixstring
 *-------------------------------------------------------------------*/
static void fixstring (unsigned char *s, int bytecount)
{       
    unsigned char *p, *end = &s[bytecount &= ~1];   /* bytecount must be even */
    
    /* convert from big-endian to little-endian */
    for (p = end ; p != s;) {
        unsigned short *pp = (unsigned short *) (p -= 2);
        *pp = (*pp >> 8) | (*pp << 8);
    }       
            
    /* strip leading blanks */
    while (s != end && *s == ' ')
        ++s;

    /* compress internal blanks and strip trailing blanks */
    while (s != end && *s) { 
        if (*s++ != ' ' || (s != end && *s && *s != ' '))
            *p++ = *(s-1);
    }
    
    /* wipe out trailing garbage */
    while (p != end)    
        *p++ = '\0';
}
/*-------------------------------------------------------------------*
 * identify_intr
 *-------------------------------------------------------------------*/
static void identify_intr(void)
{
	unsigned int dev = DEVICE_NR(CURRENT->dev);
	struct hd_driveid *id = hd_ident_info;
	unsigned short stat = inb_p(HD_STATUS);

	if(stat & (BUSY_STAT | ERR_STAT))
		kprintf("BUSY -_- & ERROR -_- \n");
	else {
		port_read(HD_DATA, id, 256);
        if ((id->field_valid&1) && id->cur_cyls && id->cur_heads && (id->cur_heads <= 16) && id->cur_sectors) {
			kprintf("Make new CHS\n");
            current_gover->hd_info[0].cyl  = id->cur_cyls;
            current_gover->hd_info[0].head = id->cur_heads;
            current_gover->hd_info[0].sect = id->cur_sectors;
        }
		fixstring (id->model, sizeof(id->model));
        fixstring(id->serial_no, sizeof(id->serial_no));
        fixstring(id->fw_rev, sizeof(id->fw_rev));
        kprintf("  hd%c: %s %dMB w/%dKB Cache, CHS=%d/%d/%d, MaxMult=%d\n",
            (dev+'a'), id->model, (id->cyls*id->heads*id->sectors/2048),
            (id->buf_size/2), (current_gover->hd_info[dev].cyl), (current_gover->hd_info[dev].head),
            (current_gover->hd_info[dev].sect), (id->max_multsect));
	}
	do_hd_request();
	return;
}
/*-------------------------------------------------------------------*
 * read_intr
 *-------------------------------------------------------------------*/
static void read_intr(void)
{
	int i, retries = 100000;
	cli();
    do {
        i = (unsigned) inb_p(HD_STATUS);
        if (i & BUSY_STAT)
            continue;
        if (!OK_STATUS(i))
            break;
        if (i & DRQ_STAT)
            goto ok_to_read;
    } while (--retries > 0);
	bad_rw_intr();
	do_hd_request();
	sti();
	return;
ok_to_read:
	port_read(HD_DATA,CURRENT->buffer,256);
	CURRENT->errors = 0;
	CURRENT->buffer += 512;
	CURRENT->sector++;
	if (--CURRENT->nr_sectors) {
		SET_INTR(&read_intr);
		sti();
		return;
	}
	sti();
	end_request(1);
	do_hd_request();
}

/*-------------------------------------------------------------------*
 * write_intr
 *-------------------------------------------------------------------*/
static void write_intr(void)
{
	if (win_result()) {
		bad_rw_intr();
		do_hd_request();
		return;
	}
	if (--CURRENT->nr_sectors) {
		CURRENT->sector++;
		CURRENT->buffer += 512;
		SET_INTR(&write_intr);
		port_write(HD_DATA,CURRENT->buffer,256);
		return;
	}
	end_request(1);
	do_hd_request();
}

/*-------------------------------------------------------------------*
 * recal_intr
 *-------------------------------------------------------------------*/
static void recal_intr(void)
{
	if (win_result())
		bad_rw_intr();
	do_hd_request();
}

/*-------------------------------------------------------------------*
 * hd_times_out
 *-------------------------------------------------------------------*/
void hd_times_out(void)
{
	if (!CURRENT)
		return;
	kprintf("HD timeout");
	if (++CURRENT->errors >= MAX_ERRORS)
		end_request(0);
	SET_INTR(NULL);
	reset = 1;
	do_hd_request();
}

/*-------------------------------------------------------------------*
 * do_special_op
 *-------------------------------------------------------------------*/
int do_special_op(unsigned int dev)
{
    /* ⼭ hd_info[0]  ù° ϵ带 ŵϴ. ̷ 0
     * ھ   ߿ upgrade   ų  
     *   weix  ϳ ϵ常 Ѵٴ  Ÿ ݴϴ.
     * FIXME : ΰ ̻ ϵ带 νϵ ľմϴ. */
    /* 0xEC : COMMAND : get device information */
	if(!(current_gover->hd_info[0].u.hd_part[dev].identified)) {
		current_gover->hd_info[0].u.hd_part[dev].identified = 1;
		hd_out(dev,0,0,0,0,0xEC,&identify_intr);
		return reset;
	}
	current_gover->hd_info[0].u.hd_part[dev].special_op = 0;
	return 1;
}
/*-------------------------------------------------------------------*
 * do_hd_request
 *-------------------------------------------------------------------*/
void do_hd_request(void)
{
	int i,r;
	unsigned int block,dev;
	unsigned int sec,head,cyl, track;
	unsigned int nsect;

repeat:
	sti();
	INIT_REQUEST;
    if (reset) {
        recalibrate = 1;
        reset_hd();
        return;
    }
	
    dev = MINOR(CURRENT->dev);
    block = CURRENT->sector;
    nsect = CURRENT->nr_sectors;
									// hd_info[0]   weix ù° ϵ ϸ   
									//  Ÿϴ.  Ͻñ ٶϴ.
	if (dev >= 5*NR_HD || block+2 > current_gover->hd_info[0].u.hd_part[dev].nr_sects) {
		kprintf("-_- block > hd[dev].nr_sects\n");
		end_request(0);
		goto repeat;
	}

	/*  б  LBA  Ǿ ֽϴ. */
    block += current_gover->hd_info[0].u.hd_part[dev].start_sect;
    dev >>= 6;

//    /* identify_intr   Ȯ ۵ ʽϴ. */
//	if(current_gover->hd_info[0].u.hd_part[dev].special_op) {
//		if(do_special_op(dev)) {
//			goto repeat;
//		}
//		return;
//	}

    sec   = block % current_gover->hd_info[dev].sect + 1;
    track = block / current_gover->hd_info[dev].sect;
    head  = track % current_gover->hd_info[dev].head;
    cyl   = track / current_gover->hd_info[dev].head;

	//kprintf("CURRENT->sector = %d, ", CURRENT->sector);
	//kprintf("block += hd[dev].start_sect = %d, ", block);
	//kprintf("nsect = CURRENT->nr_sectors = %d, ", nsect);
	//kprintf("sec=%d, track=%d, head=%d, cyl=%d, ", sec, track, head, cyl);

	if (recalibrate) {
		recalibrate = 0;
		hd_out(dev,current_gover->hd_info[CURRENT_DEV].sect,0,0,0,
			WIN_RESTORE,&recal_intr);
		return;
	}	

	if (CURRENT->cmd == WRITE) {
		hd_out(dev,nsect,sec,head,cyl,WIN_WRITE,&write_intr);
		r=inb_p(HD_STATUS)&DRQ_STAT;
		for(i=0 ; i<10000 && !(r) ; i++)
			/* nothing */ ;
		if (!r) {
			bad_rw_intr();
			goto repeat;
		}
		port_write(HD_DATA,CURRENT->buffer,256);
		return;
	} else if (CURRENT->cmd == READ) {
		hd_out(dev,nsect,sec,head,cyl,WIN_READ,&read_intr);
		return;
	} else
		panic("unknown hd-command");
}

static int hd_ioctl(struct inode * inode, struct file * file,
    unsigned int cmd, unsigned long arg)
{
	kprintf("in hd_ioctl\n");
	return -1;
}
static int hd_open(struct inode * inode, struct file * filp)
{
	kprintf("in hd_open\n");
	return -1;
}

static void hd_release(struct inode * inode, struct file * file)
{
    kprintf("in hd_release\n");
}

int block_read_snd(struct inode * inode, struct file * filp, char * buf, int count)
{
	kprintf("in block_read\n");
	return -1;
}
int block_write(struct inode * inode, struct file * filp, char * buf, int count)
{
	kprintf("in block_write\n");
	return -1;
}
int block_fsync(struct inode *inode, struct file *filp)
{
	kprintf("in block_fsync\n");
	return -1;
}
static struct file_operations hd_fops = {
    NULL,           /* lseek - default */
    block_read_snd,     /* read - general block-dev read */
    block_write,        /* write - general block-dev write */
    NULL,           /* readdir - bad */
    NULL,           /* select */
    hd_ioctl,       /* ioctl */
    NULL,           /* mmap */
    hd_open,        /* open */
    hd_release,     /* release */
    block_fsync     /* fsync */
};

/*-------------------------------------------------------------------*
 * hd_init
 *
 * ϵ ũ ʱȭϴ κԴϴ.  weix 
 *  ϰ Ǿ  ʾ  ʴ κ
 * ֽϴ. 
 *-------------------------------------------------------------------*/
void hd_init(void)
{
	/* hd  մϴ. */
    if (register_blkdev(MAJOR_NR,"hd",&hd_fops)) {
        kprintf("Unable to get major %d for harddisk\n",MAJOR_NR);
        return;
    }

	/* DEVICE_REQUEST = do_hd_request */
	block_device[MAJOR_NR].request_fn = DEVICE_REQUEST;
}
