/*-------------------------------------------------------------------
 * drivers/char/console.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*-------------------------------------------------------------------
 *  NOTE!!! We sometimes disable and enable interrupts for a short while
 * (to put a word in video IO), but this will work even for keyboard
 * interrupts. We know interrupts aren't enabled when getting a keyboard
 * interrupt, as we use trap-gates. Hopefully all is well.
 *-------------------------------------------------------------------*/
#include <weix/sched.h>
#include <weix/tty.h>
#include <asm/io.h>
#include <asm/system.h>

#include <weix/extern_func_T_T.h>
/*-------------------------------------------------------------------
 *  κ
 *-------------------------------------------------------------------*/
#define SCREEN_START 	0xb8000
#define SCREEN_END   	0xc0000
#define LINES 			25
#define COLUMNS 		80
#define NPAR 			16
#define KEYBOARD_IRQ	1
/*
 * this is what the terminal answers to a ESC-Z or csi0c
 * query (= vt100 response).
 */
#define RESPONSE "\033[?1;2c"

/*-------------------------------------------------------------------
 *  void con_init(void);
 *
 * This routine initalizes console interrupts, and does nothing
 * else. If you want the screen to clear, call tty_write with
 * the appropriate escape-sequece.
 *-------------------------------------------------------------------*/
void con_init(void)
{
	register unsigned char a;

	kbd_init();

	outb_p(0x21, inb_p(0x21)&0xfd);
	a=inb_p(0x61);
	outb_p(0x61, a|0x80);
	outb(0x61, a);
}
