/*-------------------------------------------------------------------
 * drivers/char/tty_io.c
 *-------------------------------------------------------------------*/
#include <weix/errno.h>
#include <weix/signal.h>
#include <asm/io.h>
#include <weix/sched.h>
#include <weix/tty.h>
#include <weix/keyboard.h>
#include <weix/kd.h>
#include <asm/fs_func.h>
#include <asm/system.h>
#include <weix/ctype.h>
#include <weix/kdebug.h>
#include <weix/kernel.h>
#include <weix/fs.h>
#include <weix/types.h>
#include <weix/stat.h>
#include <weix/string.h>
#include <weix/dirent.h>
#include <weix/tty.h>
#include <asm/bitops.h>
#include <weix/kdebug.h>

#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------
 * global variables
 *-------------------------------------------------------------------*/
static int tty_count=1024*8;
int save_tty = 0;
static int save_tty_tail = 0;
static int save_tty_head = 0;
int fg_console = 0;
struct tty_struct tty_table[] = {
	{
		{0,
		OPOST|ONLCR,		/* change outgoing NL to CRNL */
		0,
		ICANON | ECHO | ECHOCTL | ECHOKE,
		0,					/* console termio */
		INIT_C_CC},
		0,					/* initial pgrp */
		0,					/* initial stopped */
		NULL,				/*  ̰ con_write  ־ ڸ */
		{0,0,0,0,""},		/* console read-queue */
		{0,0,0,0,""},		/* console write-queue */
		{0,0,0,0,""}		/* console secondary queue */
	},{
		{0, /*IGNCR*/
		OPOST | ONLRET,		/* change outgoing NL to CR */
		B2400 | CS8,
		0,
		0,
		INIT_C_CC},
		0,
		0,
		NULL,				/*  ̰ rs_write  ־ ڸ */
		{0x3f8,0,0,0,""},		/* rs 1 */
		{0x3f8,0,0,0,""},
		{0,0,0,0,""}
	},{
		{0, /*IGNCR*/
		OPOST | ONLRET,		/* change outgoing NL to CR */
		B2400 | CS8,
		0,
		0,
		INIT_C_CC},
		0,
		0,
		NULL,				/*  ̰ rs_write  ־ ڸ */
		{0x2f8,0,0,0,""},	/* rs 2 */
		{0x2f8,0,0,0,""},
		{0,0,0,0,""}
	}
};

static int tty_lseek(struct inode * inode, struct file * file, off_t offset, int orig)
{
	kprintf("in tty_lseek\n");
    return -ESPIPE;
}
static int tty_select(struct inode * inode, struct file * filp, int sel_type, select_table * wait)
{
	kprintf("in tty_select - fix me\n");
	return -1;
}
static int tty_open(struct inode * inode, struct file * f)
{
	kprintf("in tty_open\n");
	return 0;
}
static void tty_release(struct inode * inode, struct file * filp)
{
	kprintf("in tty_release\n");
}
int tty_ioctl(struct inode * inode, struct file * file,
    unsigned int cmd, unsigned long arg)
{
	kprintf("in tty_ioctl\n");
	return -1;
}

/*-------------------------------------------------------------------
 * is_need_save_tail
 *
 *  : tty_queue  head tail ׸  ʿ䰡 ִ ˻
 * մϴ.  Լ sys_read  Բ ˴ϴ.
 *-------------------------------------------------------------------*/
int is_need_save_tail(void)
{
	if(save_tty)
		return 1;
	return 0;
}

/*-------------------------------------------------------------------
 * tty_read_getchar
 *
 *  : Ű ۿ  Էµ ڸ ϳ  ɴϴ. 
 * Ű Է  ÿ Ű Էµ  ٸϴ.
 *-------------------------------------------------------------------*/
int tty_read_getchar(void)
{
	uchar key;
	ushort keycode;
	struct tty_struct * tty;

	tty = &(TTY_TABLE(0));

	while(!(key = (char) tty->read_q.buf[save_tty_tail++])) {
		/*  save_tty_tail  Ű   0̶
		 * ٽ 1 ν  ڸ  ˻ϰ Ѵ. */
		save_tty_tail--;
	}
	/*  (char) tty->read_q.buf[save_tty_tail++]  0 ƴ϶
	 *  Ѿ ȴ. */
	keycode = key_map[kbd_state][key];

	return ((int)keycode);
}

/*-------------------------------------------------------------------
 * tty_read
 *
 *  : ͹̳  count ŭ ڿ Է ޽ϴ.
 *-------------------------------------------------------------------*/
static int tty_read(struct inode * inode, struct file * file, char * buf, int count)
{
	int i;
	char keycode;
	char tmp_buf[256];

	/* TTY_QUEUE  TAIL, HEAD  ʿ䰡  մϴ.
	 * do_tty_interrupt Լ Ͻñ ٶϴ. */
	save_tty = 1;

	/*  ũ⸸ŭ  о Դϴ. */
	for(i=0; i<sizeof(tmp_buf); i++) {
		keycode = (char)tty_read_getchar();

		/* keycode   ˴ϴ. */
		if(keycode == -1) {
			kprintf("tty_read_getchar: error\n");
			return -1;
		}
		/*  ϰ  ´ٸ   ϴ.
		 * 0x01    (char)  keycode
		 * ȯ ؼ short  պκ  
		 * Դϴ.  ϰ 0x201 Դϴ. */
		if(keycode == 0x01)
			break;

		tmp_buf[i] = keycode;
	}

	memcpy(buf, tmp_buf, count);
	/*  TTY_QUEUE    ʿ䰡 . */
	save_tty = 0;
	return count;
}

/*-------------------------------------------------------------------
 * tty_write
 *
 *  :  ͹̳ο buf  count ŭ մϴ.
 *-------------------------------------------------------------------*/
static int tty_write(struct inode * inode, struct file * file, char * buf, int count)
{
	int i, n;
	va_list args;
	char buf_tmp[256];
	
#if 0
	kprintf("in tty_write\n");
#endif
	va_start(args, buf);
	n = vsprintf(buf_tmp, buf, args);
	va_end(args);

	for(i=0; i<n; i++) {
		switch(buf_tmp[i]) {
			/* ٸ ɼ  ó ,  ڸ  մϴ. */
			default		: console_putchar(buf_tmp[i]);
						  break;
		}
	}

	return n;
}

static struct file_operations tty_fops = {
    tty_lseek,
    tty_read,
    tty_write,
    NULL,       /* tty_readdir */
    tty_select,
    tty_ioctl,
    NULL,       /* tty_mmap */
    tty_open,
    tty_release
};

/*-------------------------------------------------------------------
 * char_device_init
 *
 *  : character ̽ ʱȭ մϴ.
 *-------------------------------------------------------------------*/
void char_device_init(void)
{
	con_init();

    if (register_chrdev(TTY_MAJOR,"tty",&tty_fops))
        panic("unable to get major %d for tty device", TTY_MAJOR);
}

/*-------------------------------------------------------------------
 * view_prev_screen
 *-------------------------------------------------------------------*/
void view_prev_screen(void)
{
	int i=0, return_num=0, tmp_i;
	struct current_tty_buffer * tt;

	tty_count -= 2000;
	tmp_i=tty_count;

	tt = tty_buffer;

	console_clear_screen();

	while(1)
		if(tt->buffer[tty_count++] == '\n')
			break;

	/* print characters in screen */
	for(i=tty_count; (return_num < 23) &&((i - tmp_i) < 2000); i++) {
		if(tt->buffer[i] == '\n') {
			console_cleartoline();
			return_num++;
			continue;
		}
		console_putchar(tt->buffer[i]);
	} 
}

/*-------------------------------------------------------------------
 * view_next_screen
 *-------------------------------------------------------------------*/
void view_next_screen(void)
{
	int i, return_num=0;
    struct current_tty_buffer * tt;

    tt = tty_buffer;

    console_clear_screen();
    /* print characters in screen */
    for(i=0; (return_num < 23) && (i < 2000); i++) {
        if(tt->buffer[i] == '\n') {
            console_cleartoline();
            return_num++;
            continue;
        }
        console_putchar(tt->buffer[i]);
    }
}

/*-------------------------------------------------------------------
 * do_tty_interrupt
 *
 *  :  Լ ȣǸ  ͹̳ queue  ִ ۵ ȭ
 *  մϴ. 
 *-------------------------------------------------------------------*/
void do_tty_interrupt(int tty)
{
	uchar c;
	ushort key_code;
	struct tty_struct * ts;
	
	ts = tty_table + 0;

	/* tty_queue  tail  ʿ䰡  ?
	 * sys_read ÿ ˴ϴ. */
	if(is_need_save_tail()) {
		/* ͷƮ diable ŵϴ.
		 * sys_read ϰ ִ  sys_read save_tty_tail
		 * save_tty_head  ϰ ֽϴ. ׷  
		 * ٸ ͷƮ    Ἲ ϵ
		 * մϴ. */
		cli();
		/*  մϴ. */
		save_tty_tail = ts->read_q.tail;
		save_tty_head = ts->read_q.head;
		/* ͷƮ enable ŵϴ. */
		sti();
	}

	/* tail head   ȭ鿡 ڸ մϴ. */
	while(ts->read_q.tail != ts->read_q.head) {
		
		c = (char) ts->read_q.buf[ts->read_q.tail];
		key_code = key_map[kbd_state][c];
		kprintf("%c", key_code);

		/* tail  1  ŵϴ. */
		if(ts->read_q.tail < (TTY_BUF_SIZE - 1)) {
			INC(ts->read_q.tail);
		} else {
			ts->read_q.tail = 0;
		}
	}
}
