/** io.h
 **
 ** Description:
 ** Low-level Input/Output Functions
 **
 **/

/*
 * Mach Operating System
 * Copyright (c) 1991,1990 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */

#ifndef __ASM_IO_H_
#define __ASM_IO_H_

#include <asm/types.h>

#define io_wait()   inb(0x80)

#define inl(port) \
({ dword _tmp__; \
	asm volatile("inl %1, %0" : "=a" (_tmp__) : "d" ((word)(port))); \
	_tmp__; })
#define inportl(port) inl(port)

#define inw(port) \
({ word _tmp__; \
	asm volatile(".byte 0x66; inl %1, %0" : "=a" (_tmp__) : "d" ((word)(port))); \
	_tmp__; })
#define inportw(port) inw(port)

#define inb(port) \
({ byte _tmp__; \
	asm volatile("inb %1, %0" : "=a" (_tmp__) : "d" ((word)(port))); \
	_tmp__; })
#define inportb(port) inb(port)
#define inb2(port) \
({ byte _tmp__; \
	asm volatile("inb %1, %0" : "=a" (_tmp__) : "g" ((word)port) ); \
	_tmp__; })
#define inportb2(port) inb2(port)

#define outl(port, val) \
({ asm volatile("outl %0, %1" : : "a" ((dword)val) , "d" ((word)(port))); })
#define outportl(port, val) outl(port, val)

#define outw(port, val) \
({asm volatile(".byte 0x66; outl %0, %1" : : "a" ((word)(val)) , "d" ((word)(port))); })
#define outportw(port, val) outw(port, val)

#define outb(port, val) \
({ asm volatile("outb %0, %1" : : "a" ((byte)(val)) , "d" ((word)(port))); })
#define outportb(port, val) outb(port, val)

#define outb_p(port,value) \
__asm__ ("outb %%al,%%dx\n" \
        "\tjmp 1f\n" \
        "1:\tjmp 1f\n" \
        "1:"::"a" (value),"d" (port))

#define inb_p(port) ({ \
unsigned char _v; \
__asm__ volatile ("inb %%dx,%%al\n" \
    "\tjmp 1f\n" \
    "1:\tjmp 1f\n" \
    "1:":"=a" (_v):"d" (port)); \
_v; \
})

#endif /* __ASM_IO_H_ */
