#ifndef __ASM_SYSTEM_H
#define __ASM_SYSTEM_H

#include <asm/types.h>
#include <asm/segment.h>

/* interrupt control.. */
#define __save_flags(x)		__asm__ __volatile__("pushfl ; popl %0":"=g" (x): /* no input */)
#define __restore_flags(x) 	__asm__ __volatile__("pushl %0 ; popfl": /* no output */ :"g" (x):"memory")
#define __cli() 		__asm__ __volatile__("cli": : :"memory")
#define __sti()			__asm__ __volatile__("sti": : :"memory")
/* used in the idle loop; sti takes one instruction cycle to complete */
#define safe_halt()		__asm__ __volatile__("sti; hlt": : :"memory")

/* For spinlocks etc */
#define local_irq_save(x)	__asm__ __volatile__("pushfl ; popl %0 ; cli":"=g" (x): /* no input */ :"memory")
#define local_irq_restore(x)	__restore_flags(x)
#define local_irq_disable()	__cli()
#define local_irq_enable()	__sti()

#define nop() __asm__ __volatile__ ("nop")
#define cli() __cli()
#define sti() __sti()
#define save_flags(x) __save_flags(x)
#define restore_flags(x) __restore_flags(x)
/* again with other names */
#define get_flags(x) __save_flags(x)
#define set_flags(x) __restore_flags(x)


/* paging */
extern inline dword get_CR0 (void)
{
  register dword x;
  __asm__ __volatile__ ("movl %%cr0,%0" : "=r" (x));
  return (x);
}

extern inline dword get_CR3 (void)
{
  register dword x;
  __asm__ __volatile__ ("movl %%cr3,%0" : "=r" (x));
  return (x);
}

#define set_CR0(c)  __asm__ __volatile__ ("movl %%eax, %%cr0"::"a" (c))
#define set_CR3(c)  __asm__ __volatile__ ("movl %%eax, %%cr3"::"a" (c))

extern inline void flush_tlb(void) { set_CR3(get_CR3()); }

/* REP NOP (PAUSE) is a good thing to insert into busy-wait loops. */
extern inline void rep_nop(void) { __asm__ __volatile__("rep;nop"); }

#define move_to_user_mode() \
__asm__ ("movl %%esp,%%eax\n\t" \
    "pushl %0\n\t" \
    "pushl %%eax\n\t" \
    "pushfl\n\t" \
    "pushl %1\n\t" \
    "pushl $next_step\n\t" \
    "iret\n" \
    "next_step: \n" \
	"movl %0,%%eax\n\t" \
    "movw %%ax,%%ds\n\t" \
    "movw %%ax,%%es\n\t" \
    "movw %%ax,%%fs\n\t" \
    "movw %%ax,%%gs" \
    ::"i" (USER_DS), "i" (USER_CS):"ax")

#endif
