/** kernel.h
 **
 ** Description:
 ** JijOS kernel specific functions
 **
 **/

#ifndef __WEIX_KERNEL_H_
#define __WEIX_KERNEL_H_

#include <weix/types.h>

void halt(void) __attribute__ ((noreturn));
void reboot(void) __attribute__ ((noreturn));
void waitKey(void);

#define KERN_EMERG  "<0>"   /* system is unusable           */
#define KERN_ALERT  "<1>"   /* action must be taken immediately */
#define KERN_CRIT   "<2>"   /* critical conditions          */
#define KERN_ERR    "<3>"   /* error conditions         */
#define KERN_WARNING    "<4>"   /* warning conditions           */
#define KERN_NOTICE "<5>"   /* normal but significant condition */
#define KERN_INFO   "<6>"   /* informational            */
#define KERN_DEBUG  "<7>"   /* debug-level messages         */

#if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 5)
# define NORET_TYPE    __volatile__
# define ATTRIB_NORET  /**/
# define NORET_AND     /**/
#else
# define NORET_TYPE    /**/
# define ATTRIB_NORET  __attribute__((noreturn))
# define NORET_AND     noreturn,
#endif

#define MULTIBOOT_MEMORY	(1L<<0)
#define MULTIBOOT_BOOT_DEVICE	(1L<<1)
#define MULTIBOOT_CMDLINE	(1L<<2)
#define MULTIBOOT_MODS		(1L<<3)
#define MULTIBOOT_AOUT_SYMS	(1L<<4)
#define MULTIBOOT_ELF_SHDR	(1L<<5)
#define MULTIBOOT_MEM_MAP	(1L<<6)


typedef struct multiboot_module
{
  /* physical start and end addresses of the module data itself.  */
  dword start;
  dword end;

  /* arbitrary ASCII-Z string associated with the module.  */
  char *string;

  /* boot loader must set to 0; OS must ignore.  */
  dword reserved;
} __attribute__ ((packed)) multiboot_module;

typedef struct memory_map
{
	dword	size;
	dword	base_addr_low;
	dword	base_addr_high;
	dword	length_low;
	dword	length_high;
	dword	type;
} __attribute__ ((packed)) memory_map;

typedef struct drives_map
{
	dword	size;
	uchar	drive_number;
	uchar	drive_mode;
	ushort	drive_cylinders;
	uchar	drive_heads;
	uchar	drive_sectors;
	ushort	drive_ports;
} __attribute__ ((packed)) drives_map;

typedef struct multiboot_info
{
	/* these flags indicate which parts of the multiboot_info are valid;
	 * see below for the actual flag bit definitions.  */
	dword flags;

	/* lower/upper memory installed in the machine.
	 * valid only if MULTIBOOT_MEMORY is set in flags word above.  */
	dword memory_lower;
	dword memory_upper;

	/* BIOS disk device the kernel was loaded from.
	 * Valid only if MULTIBOOT_BOOT_DEVICE is set in flags word above.  */
	byte boot_device[4];

	/* command-line for the OS kernel: a null-terminated ASCII string.
	 * valid only if MULTIBOOT_CMDLINE is set in flags word above.  */
	char *command_line;

	/* list of boot modules loaded with the kernel.
	 * valid only if MULTIBOOT_MODS is set in flags word above.  */
	dword modules_count;
	multiboot_module *modules;

	/* info about kernel ELF file */
	dword num;
	dword size;
	dword addr;
	dword reserved;

	/* memory map -- int 0x15 BIOS call (e820) */
	dword mmap_length;
	memory_map *mmap_addr;

	dword		drives_length;
	drives_map	*drives_map;
} __attribute__ ((packed)) multiboot_info;

extern int do_exit(long code);

#endif /* __WEIX_KERNEL_H_ */
