#ifndef _WEIX_PAGE_H
#define _WEIX_PAGE_H

#include <weix/mm.h>
#include <weix/idt.h> /* for struct gate */
#include <weix/gdt.h> /* for struct descriptior */

#define PDE_POSITION ((GDT_SIZE * sizeof(descriptor)) + (IDT_SIZE * sizeof(gate)) + 0x800 + 0x70000)

			/* PAGE_SHIFT determines the page size */
#define PAGE_SHIFT			12
#define PAGE_SIZE			((unsigned long)1<<PAGE_SHIFT)

			/* number of bits that fit into a memory pointer */
#define BITS_PER_PTR			(8*sizeof(unsigned long))
			/* to mask away the intra-page address bits */
#define PAGE_MASK			(~(PAGE_SIZE-1))
			/* to align the pointer to the (next) page boundary */
#define PAGE_ALIGN(addr)		(((addr)+PAGE_SIZE-1)&PAGE_MASK)
			/* to align the pointer to a pointer address */
#define PTR_MASK			(~(sizeof(void*)-1))

					/* sizeof(void*)==1<<SIZEOF_PTR_LOG2 */
					/* 64-bit machines, beware!  SRB. */
#define SIZEOF_PTR_LOG2			2

			/* to find an entry in a page-table-directory */
#define PAGE_DIR_OFFSET(base,address)	((unsigned long*)((base)+\
  ((unsigned long)(address)>>(PAGE_SHIFT-SIZEOF_PTR_LOG2)*2&PTR_MASK&~PAGE_MASK)))
			/* to find an entry in a page-table */
#define PAGE_PTR(address)		\
  ((unsigned long)(address)>>(PAGE_SHIFT-SIZEOF_PTR_LOG2)&PTR_MASK&~PAGE_MASK)
			/* the no. of pointers that fit on a page */
#define PTRS_PER_PAGE			(PAGE_SIZE/sizeof(void*))

#define PAGE_PRESENT    0x001
#define PAGE_RW     0x002
#define PAGE_USER   0x004
#define PAGE_PWT    0x008   /* 486 only - not used currently */
#define PAGE_PCD    0x010   /* 486 only - not used currently */
#define PAGE_ACCESSED   0x020
#define PAGE_DIRTY  0x040
#define PAGE_COW    0x200   /* implemented in software (one of the AVL bits) */

#define PAGE_PRIVATE    (PAGE_PRESENT | PAGE_RW | PAGE_USER | PAGE_ACCESSED | PAGE_COW)
#define PAGE_SHARED (PAGE_PRESENT | PAGE_RW | PAGE_USER | PAGE_ACCESSED)
#define PAGE_COPY   (PAGE_PRESENT | PAGE_USER | PAGE_ACCESSED | PAGE_COW)
#define PAGE_READONLY   (PAGE_PRESENT | PAGE_USER | PAGE_ACCESSED)
#define PAGE_TABLE  (PAGE_PRESENT | PAGE_RW | PAGE_USER | PAGE_ACCESSED)

#endif /* _WEIX_PAGE_H */
