/********************************************************
 * include/weix/sched.h
 *
 * this is schedule header
 ********************************************************/
#ifndef _WEIX_SCHED_H
#define _WEIX_SCHED_H

#include <asm/processor.h>
#include <asm/segment.h>
#include <asm/types.h>
#include <weix/types.h>
#include <weix/params.h>
#include <weix/head.h>
#include <weix/idt.h>
#include <weix/gdt.h>
#include <weix/hdd_struct.h>
#include <weix/linkage.h>
#include <weix/fs.h>
#include <weix/page.h>
#include <weix/config.h>

#define HZ 100

#define TASK_SIZE   	0x04000000

#if (TASK_SIZE & 0x3fffff)
#error "TASK_SIZE must be multiple of 4M"
#endif

#if (((TASK_SIZE>>16)*NR_TASKS) != 0x10000)
#error "TASK_SIZE*NR_TASKS must be 4GB"
#endif

#define FIRST_TASK current_gover->citizen[0]
#define LAST_TASK current_gover->citizen[NR_TASKS-1]

/* task  ¸ ŵϴ. */
#define TASK_RUNNING        	0
#define TASK_INTERRUPTIBLE  	1
#define TASK_UNINTERRUPTIBLE    2
#define TASK_ZOMBIE     		3
#define TASK_STOPPED        	4

#ifndef NULL
#define NULL ((void *) 0)
#endif

typedef int (*fn_ptr)();

/* Intel CPU  ڵԴϴ. */
struct i387_struct {
    long    cwd;
    long    swd;
    long    twd;
    long    fip;
    long    fcs;
    long    foo;
    long    fos;
    long    st_space[20];   /* 8*10 bytes for each FP-reg = 80 bytes */
};

/* TSS ü Դϴ. */
struct tss_struct {
    ushort	back_link;
	ushort	tmp_1;
    ulong	esp0;
    ushort	ss0;
	ushort	tmp_2;
    ulong	esp1;
    ushort	ss1;
	ushort	tmp_3;
    ulong	esp2;
    ushort	ss2;
	ushort	tmp_4;
    ulong	cr3;
    ulong	eip;
    ulong	eflags;
    ulong	eax,ecx,edx,ebx;
    ulong	esp;
    ulong	ebp;
    ulong	esi;
    ulong	edi;
    ushort	es;
	ushort	tmp_5;
    ushort	cs;
	ushort	tmp_6;
    ushort	ss;
	ushort	tmp_7;
    ushort	ds;
	ushort	tmp_8;
    ushort	fs;
	ushort	tmp_9;
    ushort	gs;
	ushort	tmp_10;
    ushort	ldt;
	ushort	tmp_11;
    long    trace_bitmap;
    struct i387_struct i387;
};

/*  my_life_schedule  ũ 564bytes  Ůϴ. */
struct my_life_schedule {
/* hard coding  Ǿ ֱ    Ǳ ǽñ ٶϴ. ( :  :-)) */
	long	state			/* μ :	TASK_RUNNING
												TASK_INTERRUPTIBLE
												TASK_UNINTERRUPTIBLE
												TASK_ZOMEBIE */
	,		counter			/*  ð */
	,		my_priority		/* μ  */
	/*  weix signal ó κ
	 * ϼ ʾҽϴ.  Űµ 
	 *  ֽñ ٶϴ. */
	,		signal;			/* ñ׳ */
	fn_ptr	sig_restorer;
	fn_ptr	sig_fn[32];

/* پ  */
	int 	exit_code;
	ulong	start_of_code
	,		start_of_data
	,	 	end_of_code
	,		end_of_data
	,		brk
	,		start_stack;
	
/* PID ... */
	struct my_life_schedule		*my_goverment, *my_father, *my_mother	/* θ   */
	,							*my_brothers, *my_children
	,							*my_friends;
	long	my_occupation	/*
							 *   - μ  -
							 * 1. Ϲ 
							 *		 (level)
							 * 2. ΰںе( )
							 * 		 ڻ (level)
							 */
	,    	pid				/* μ ID */
	,		pgrp
	,		session
	,		leader;

/* uid ... */
	ushort 	uid				/*  id */
	, 		euid			/* effective user id */
	,	 	suid			
	,		fsuid

	, 		gid				/* ׷ id */
	, 		egid			/* effective group id */
	, 		sgid
	,		fsgid;

/* ֶ
 *  ֶ  ǥ  ǰ ø 
 *  ּ. */
	int		my_happy_degree		// 
	,		my_agree_degree		// 
	,		my_sad_degree		// 
	,		my_enjoy_degree;	// 

/* ڸ ... */
	long	alarm			/* time of next alarm in ticks, or 0 */
	,	 	utime			/* user time in ticks */
	,		stime			/* sys time in ticks */
	,		cutime 			/* cumulative user time of children */
	,		cstime	 		/* cumulative sys time of children */
	,		start_time;

/*  */
	ushort 	used_math;

/*  ý   */
	int 			tty;
	ushort 			umask;
	struct inode 	*pwd;
	struct inode 	*root;
	ulong 			close_on_exec;
	struct file 	* filp[NR_OPEN]; 

/*  task  ldt ==  0-zero, 1-cs, 2-ds&ss */
	struct descriptor 	ldt[3];

/*  task  tss */
    struct tss_struct 	tss;
	struct my_life_schedule	*next_schedule		/*  μ double linked-list */
	,						*prev_schedule;
};

#define OUR_CITIZEN		my_life_schedule
#define OUR_GOVER_STAFF	my_life_schedule
#define NR_GOVER_STAFF	10		/*  ںе ִ ο 10Դϴ. */

/*  Ȱ մϴ. */
struct weix_goverment_operations {
	/*  task  ִ  մϴ. */
	int (*   up_process_priority) (struct OUR_CITIZEN * p, long priority);
	/*  task  ִ  մϴ. */
	int (* down_process_priority) (struct OUR_CITIZEN * p, long priority);
	/*  task   , ΰںе   մϴ. */
	int (*set_process_occupation) (struct OUR_CITIZEN * p, long occupation);
	/* ε ֶ  Ŀ ȯ濡 ݿϴ  մϴ.
	 * ̷ ɿ ϱ⿡   ൿ     մϴ. */
	int (*          reflect_hnek) (struct OUR_CITIZEN * p);
};

/*
 * δ ε   ֽϴ. ׸ α(?) θ
 *  Ҽ ֽϴ.
 */
	 //ͽ_
struct weix_goverment {
	/* ε 繫 */
    struct OUR_CITIZEN * citizen[NR_TASKS];		/* 츮 ùε */
	struct OUR_GOVER_STAFF * gover_staff[NR_GOVER_STAFF];	/*  ںе */
	/* ٷ */
	struct wait_queue run_queue_list[MAX_NR_RUN_QUEUE]; /* μ   ΰ մϴ. */

	/*  ڵ   ֵ ߰մϴ. */

	/* ùε ֶ Ͽ հ  ֽϴ. */
	/* "" "" ٸ δ ùε鿡   Դϴ. */
    int     citizen_happy_degree	//
    ,       citizen_agree_degree	//
    ,       citizen_sad_degree		//
    ,       citizen_enjoy_degree;	//

	/*  Ȱ */
	struct weix_goverment_operations * g_op;

	/* ΰ   */
		/* 1)  ̽   */
	struct hdd_struct hd_info[MAX_HD];

	/*  ο  θ ŵϴ.
	 *  1,  2   Դϴ.  weix 󿡼 
	 * ϳ θ ϵ Ͽϴ. */
    struct weix_goverment * prev, * next;
};

/*  INIT_TASK  ó task struct  ϱ  ˴ϴ. */
#define INIT_TASK \
					{ 														\
/*  */ 			0, 15, 15, 											\
/* ñ׳  	*/   		0, NULL, {(fn_ptr) 0,}, 							\
/* ec,brk...*/ 			0, 0, 0, 0, 0, 0, 0, 								\
/* pid etc..*/ 			NULL, NULL, NULL, NULL, NULL, NULL, 0, 0, 0, 0, 0, 	\
/* uid etc 	*/   		0, 0, 0, 0, 0, 0, 0, 0, 							\
/* ֶ	*/			0, 0, 0, 0, 										\
/* alarm 	*/ 			0, 0, 0, 0, 0, 0, 									\
/* math 	*/ 			0, 													\
/* fs info 	*/   		-1, 0133, NULL, NULL, 0, 							\
/* filp 	*/  		{NULL,}, 											\
/* ldt 		*/    		{ 													\
							/* dummy */										\
        					{     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 	\
							/* code segment */								\
   							{0x9fff, 0, 0, 8, 1, 3, 1, 0, 0, 0, 1, 0, 0},	\
							/* data segment */								\
        					{0x9fff, 0, 0, 2, 1, 3, 1, 0, 0, 0, 1, 0, 0}, 	\
    					}, 													\
/* tss		*/			{													\
							0, 0, PAGE_SIZE+(long)&init_task, KERNEL_DS, 0,	\
							0, 0, 0, 0, 0, 0, (ulong)(PDE_POSITION),		\
     						0, 0, 0, 0, 0, 0, 0, 0, 						\
     						0, 0, USER_DS, 0, USER_DS, 0, USER_DS, 0,		\
							USER_DS, 0, USER_DS, 0, USER_DS, 0,				\
     						_LDT(0), 0, 0x80000000, 						\
        					{} 												\
    					}, 													\
						NULL, NULL											\
					}

/* INIT_GOVER  ó goverment  ϱ  ˴ϴ.
 * Editÿ Ǳ ñ ٶϴ. */
#define INIT_GOVER \
    {                                               \
/*        */	{&(init_task.task), NULL, },    \
/*ΰڵ*/	{NULL, },                       \
/*RUN_QUEUE   */	{{NULL}, },                     \
/*ֶ    */	0, 0, 0, 0,                     \
/*Ȱ    */	NULL,                           \
    /*  HD_TYPE(ϵ Ÿ ) Ǿ ִٸ װ      \
     * ϰ ׷   0 ʱȭմϴ.  weix    \
     * ڵ ν Ҿ Դϴ. ׷Ƿ HD_TYPE        \
     * ֽô  Դϴ. */                 \
/*ϵ        */	HD_TYPE,                        \
/*prevnext    */	NULL, NULL                      \
    }

/* Bits for p_flags in my_life_schedule[]. A process is runnable if p_flags == 0 */
#define P_SLOT_FREE		001	/* set when slot is not in use. */
#define NO_MAP			002 /* keeps unmapped forked child from running */
#define SENDING			004 /* set when process blocked trying to send */
#define RECEIVING		010 /* set when process blocked trying to recv */
#define PENDING			020 /* set when inform() of signal pending */
#define SIG_PENDING		040 /* keep to-be signalled proc from running */
#define P_STOP		   0100 /* set when process is neing traced */

/* extern functions */
//extern struct my_life_schedule *task[NR_TASKS];
extern struct my_life_schedule *last_task_used_math;
extern struct my_life_schedule *current;
extern struct weix_goverment * current_gover;
extern long volatile jiffies;
extern long startup_time;
extern int timer_interrupt(void);
extern int request_irq(unsigned int irq,void (*handler)(int));
extern void free_irq(unsigned int irq);
extern asmlinkage int system_call(void);
extern volatile void panic(const char * fmt, ...);

#define CURRENT_TIME (startup_time+jiffies/HZ)

/*
 * Entry into gdt where to find first TSS. 
 * 0-nul
 * 1-KERNEL_CS
 * 2-KERNEL_DS
 * 3-USER_CS
 * 4-USER_DS
 * 5-System_call 
 * 6-TSS0
 * 7-LDT0
 * 8-TSS1 etc ...
 */
#define FIRST_TSS_ENTRY 6
#define FIRST_LDT_ENTRY (FIRST_TSS_ENTRY+1)
#define _TSS(n) ((((unsigned long) n)<<4)+(FIRST_TSS_ENTRY<<3))
#define _LDT(n) ((((unsigned long) n)<<4)+(FIRST_LDT_ENTRY<<3))
#define ltr(n) __asm__("ltr %%ax"::"a" (_TSS(n)))
#define lldt(n) __asm__("lldt %%ax"::"a" (_LDT(n)))
#define str(n) \
__asm__("str %%ax\n\t" \
    "subl %2,%%eax\n\t" \
    "shrl $4,%%eax" \
    :"=a" (n) \
    :"a" (0),"i" (FIRST_TSS_ENTRY<<3))

/*
 *  switch_to(n) should switch tasks to task nr n, first
 * checking that n isn't the current task, in which case it does nothing.
 * This also clears the TS-flag if the task we switched to has used
 * tha math co-processor latest.
 */
#define switch_to(n) {							\
struct {long a,b;} __tmp; 						\
__asm__("cmpl %%ecx,current\n\t" 				\
    	"je 1f\n\t" 							\
		"cli\n\t" 								\
    	"movw %%dx,%1\n\t" 						\
    	"xchgl %%ecx,current\n\t" 				\
    	"ljmp %0\n\t" 							\
		"sti\n\t" 								\
    	"cmpl %%ecx,last_task_used_math\n\t" 	\
    	"jne 1f\n\t" 							\
    	"clts\n" 								\
    	"1:" 									\
    	::"m" (*&__tmp.a),"m" (*&__tmp.b), 		\
    	"d" (_TSS(n)),"c" ((long) current_gover->citizen[n])); 	\
}

extern inline void set_base(struct descriptor * addr, unsigned long base)
{
  addr->base0_15 = base;
  addr->base16_23 = base >> 16;
  addr->base24_31 = base >> 24;
}

extern inline void set_limit(struct descriptor * addr, unsigned long limit)
{
  /* calculate limit and granularity */
  if (limit > LIMIT_1MB)
    {
      limit = limit >> 12;
      addr->g = 1;
    }
  else
    addr->g = 0;
  addr->limit0_15 = limit;
  addr->limit16_19 = limit >> 16;
}

extern inline unsigned long get_base(struct descriptor * addr)
{
	unsigned long g_base, tmp;

	g_base = addr->base24_31 << 24;
    tmp    = addr->base16_23 << 16;
    g_base += tmp;
    g_base += addr->base0_15;

	return g_base;
}

static inline unsigned long get_limit(unsigned long segment)
{
    unsigned long __limit;
    __asm__("lsll %1,%0"
        :"=r" (__limit):"r" (segment));
    return __limit+1;
}

#endif
