/*-------------------------------------------------------------------*
 * init/main.c
 *
 *   weix  ߽ Դϴ. 
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define __LIBRARY__
#include <weix/config.h>
#include <weix/unistd.h>
#include <weix/time.h>
#include <weix/errno.h>
#include <weix/fcntl.h>
#include <weix/types.h>
#include <asm/io.h>				 /* for outb() */
#include <weix/gdt.h>
#include <weix/idt.h>
#include <weix/malloc.h>
#include <weix/delay.h>
#include <weix/equip.h>
#include <weix/pic.h>
#include <weix/exception.h>
#include <weix/processor.h>
#include <weix/page.h>
#include <weix/mm.h>
#include <weix/kernel.h>
#include <weix/init.h>
#include <weix/linkage.h>
#include <weix/sched.h>
#include <weix/fs.h>
#include <weix/string.h>
#include <weix/version.h>
#include <weix/kdebug.h>

#include <weix/extern_func_T_T.h>
/*-------------------------------------------------------------------*
 * SYSCALL_TYPE* ũ   忡 Ǵ init()
 * fork, pause, setup, sync  ȭϴ Դϴ.
 *-------------------------------------------------------------------*/
inline SYSCALL_TYPE0(int, fork);	 		/* fork */
inline SYSCALL_TYPE0(int, pause);	 		/* pause */
inline SYSCALL_TYPE0(int, setup);	/* setup */
inline SYSCALL_TYPE0(int, sync);	 		/* sync */

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
#ifdef CONFIG_WEIX_BOOT_HDD
static ulong get_available_memory(multiboot_info *);
static ulong get_used_memory(multiboot_info *);
static void show_memory_map(multiboot_info *);
#endif
void halt(void);
void reboot(void);
static void time_init(void);
volatile void panic(const char * fmt, ...);
void init();
extern void get_hd0(void);

/*-------------------------------------------------------------------*
 * gloval variables
 *-------------------------------------------------------------------*/
multiboot_info *bootinfo;
int root_mountflags = 0;
static char * argv[] = { "-",NULL };
static char * envp[] = { "HOME=/root", NULL };
unsigned long loops_per_sec = 1;

#define CMOS_READ(addr) ({ \
    outb_p(0x70, 0x80|addr); \
    inb_p(0x71); \
})

#define BCD_TO_BIN(val) ((val)=((val)&15) + ((val)>>4)*10)
struct cpuinfo_x86 boot_cpu_data = { 0, 0, 0, 0, -1, 1, 0, 0, -1 };

#ifndef CONFIG_WEIX_BOOT_HDD
unsigned long init_user_stack[1024] = { 0, };
#endif

/*-------------------------------------------------------------------*
 * start_weix
 *
 * weix   ԼԴϴ. ϵ  ϰ Ǹ GRUB
 *  κ ֱ  ѷ ϴ. ̰ɷ
 * ؼ ҽ   ϴ.
 *
 * FIXME : GRUB  ľ   ϴ.
 *-------------------------------------------------------------------*/
#ifdef CONFIG_WEIX_BOOT_HDD
int start_weix(multiboot_info *volatile_bootinfo)
#else
int start_weix()
#endif
{
#ifdef CONFIG_WEIX_BOOT_HDD
	ulong avail_mem, used_mem, total_pages;
#endif
	struct equip_list equip_com;

	/* weix  1M ~ 8M  ޸𸮸 մϴ. */
	ulong start_mem	=0x100000;
	ulong end_mem	=0x800000;

	/* initial screen */
	/* ȭ ʱȭ մϴ. ȭ  ϴ. */
	console_init(COL_BLACK, COL_GRAY);
	kprintf("= WEIX = -> v%s\n\n", CURRENT_VERSION);

#ifdef CONFIG_WEIX_BOOT_HDD
	/*
	 * ̿밡 ޸𸮿,  ޸𸮸 մϴ. 
     * GRUB  κԴϴ. 
	 */
    avail_mem = get_available_memory(volatile_bootinfo);
    used_mem  = get_used_memory(volatile_bootinfo);
#endif
    get_BIOS_equiplist(0, &equip_com);

    /*  CPU  ˾ ϴ.  ڵ忡 ߷Ƚϴ. */
    identify_CPU(&boot_cpu_data);

#ifdef CONFIG_WEIX_BOOT_HDD
	kprintf("GRUB: Kernel command line: %s\n",volatile_bootinfo->command_line);
	kprintf("GRUB: Memory -- (used/avail) %ld/%ld Kbytes\n",
	used_mem/1024, avail_mem/1024);
	kprintf("GRUB: Memory -- (lower/upper) %ld/%ld Kbytes\n",
	volatile_bootinfo->memory_lower,volatile_bootinfo->memory_upper);
	show_memory_map(volatile_bootinfo);
#endif
	display_CPU_info(&boot_cpu_data);
	kprintf("\n");

#ifdef CONFIG_WEIX_BOOT_HDD
	/* PIC մϴ.(programmable interrupt controller(?)) */
	/* ÷Ƿ ý boot/weix_seting.S κп PIC ϱ 
     * ϵø شմϴ.*/
	reprogram_PIC();
	kprintf("WEIX: PIC master at 0x20, slave at 0x28\n");
#endif
	kprintf("WEIX: Initializing hardware tables: |");

	/* GDT մϴ. */
	setup_GDT();
	kprintf("GDT|");
	/* IDT ʱȭմϴ. */
	setup_IDT();
	kprintf("IDT|");
	/* exception  κ մϴ. */
	setup_exceptions();
	kprintf("32 Exceptions\n");

#ifdef CONFIG_WEIX_BOOT_HDD
	/* ޸𸮸 ¾մϴ.  ⼭ ¾ϴ κ  Ŀο
     *  ʽϴ.  -_-  ̰  ϴ ̷ Ǿϴ.
	 *  Ŀο mem_map[]  ؼ ޸𸮸 մϴ. */
	total_pages = setup_memory(avail_mem);
	kprintf("WEIX: Memory -- %ld total, %ld free physical pages\n", total_pages, total_pages);
	show_free_memory();
#endif

#ifdef CONFIG_WEIX_BOOT_HDD
	if (bootinfo->flags & MULTIBOOT_MODS) {
		int i;
		for(i = 0; i < bootinfo->modules_count; i++)
			kprintf("Module %d: %s\n", i, bootinfo->modules[i].string);
	} else
		kprintf("WEIX: No modules loaded\n");
#endif

	/*  ý read-only  մϴ.   test  Ⱦҽϴ. */
	root_mountflags |= MS_RDONLY;

	/*  ¡ ʱȭմϴ. weix 0M~8M  ϱ 
	 * paging 0M~8M  մϴ. ߿ fork() Լ  ڽĿ ش
	 * Ǵ μ page directory, page entry   ˴ϴ. */
	setup_paging(start_mem, end_mem);
	kprintf("WEIX: Initializing - paging|");

	/* ð ʱȭ մϴ. */
	time_init();
	kprintf("time|");

	/* ٷ ʱȭ մϴ. */
	sched_init();
	kprintf("schedule|");

	/* ޸𸮸 ʱȭմϴ. */
	start_mem = mem_init(start_mem, end_mem);
	kprintf("memory|");

	/*  ĳ ʱȭ մϴ. */
	buffer_init();
	kprintf("buffer|");

    /* character ̽ ʱȭ մϴ. */
    char_device_init();
    kprintf("character|");

	/* block ̽ ʱȭ մϴ. */
	block_device_init();
	kprintf("block|");

	kprintf("\n");

	/*   ȯմϴ. */
	move_to_user_mode();
	kprintf("WEIX: move to user mode.\n");

	/* fork  մϴ. θ μ idle · 
	 * ڽ μ init() Լ ϰ ˴ϴ. */
	if(!fork()) {
		init();
	}
	/* idle task  ƹ͵  task   ϰ 
	 * ˴ϴ. */
	kprintf("in idle status\n");
	for(;;) {
		pause();
	}
	kprintf("WEIX: Maybe exited from WEIX ;-O\n");
	return 0;
}

/*-------------------------------------------------------------------*
 * get_used_memory
 *
 *  ǰ ִ ޸ ũ⸦ Ͽ մϴ.
 *-------------------------------------------------------------------*/
#ifdef CONFIG_WEIX_BOOT_HDD
static ulong get_used_memory(multiboot_info *bootinfo)
{
    int i;
    extern void end;
    dword top = (dword)&end;

    if (bootinfo->flags & MULTIBOOT_MODS)
    {
        for (i=0; i<bootinfo->modules_count; i++)
            if (bootinfo->modules[i].end > top)
                top = bootinfo->modules[i].end;
    }

    top = (top + 4095) & 0xFFFFF000;
    return (ulong)top;
}
#endif

#ifdef CONFIG_WEIX_BOOT_HDD
/*-------------------------------------------------------------------*
 * show_memory_map
 *
 *  ޸  ݴϴ. ⼭ µǴ 
 * GRUB BIOS     ϰ ֽϴ.
 *-------------------------------------------------------------------*/
static void show_memory_map(multiboot_info *bootinfo)
{
    int i,parts;
    if (bootinfo->flags & MULTIBOOT_MEM_MAP)
    {
        kprintf("GRUB: BIOS provided physical RAM map\n");
            parts=bootinfo->mmap_length/sizeof(memory_map);
            for (i=0;i<parts;i++)
        {
            kprintf("GRUB: addr: %08x%08x  len: %08x%08x ",
                bootinfo->mmap_addr[i].base_addr_high,
                bootinfo->mmap_addr[i].base_addr_low,
                bootinfo->mmap_addr[i].length_high,
                bootinfo->mmap_addr[i].length_low);
                kprintf("%s\n",(1==bootinfo->mmap_addr[i].type)?"(usable)":"(reserved)");
        }
        kprintf("\n");
    }
}
#endif

/*-------------------------------------------------------------------*
 * get_available_memory
 * 
 * ̿밡 ޸ ũ⸦ մϴ.  Լ  GRUB
 *  ϰ ֽϴ.
 *-------------------------------------------------------------------*/
#ifdef CONFIG_WEIX_BOOT_HDD
static ulong get_available_memory(multiboot_info *bootinfo)
{
    if (bootinfo->flags & MULTIBOOT_MEMORY)
        return (ulong)((bootinfo->memory_upper + 1024) << 10);
    else {
        /* this shouldn't happen - the bootloader has
         * to provide information on available memory */
#ifdef DEBUG
        kprintf("No memory information!\n");
#endif
        halt();
    }
    return 0;
}
#endif

/*-------------------------------------------------------------------*
 * time_init - linux 0.0.1 code
 *
 * CMOS  ð  оͼ time ü ֽϴ.
 *-------------------------------------------------------------------*/
static void time_init(void)
{
    struct tm time;

    do {
        time.tm_sec 	= CMOS_READ(0);
		time.tm_min 	= CMOS_READ(2);
		time.tm_hour 	= CMOS_READ(4);
		time.tm_mday 	= CMOS_READ(7);
		time.tm_mon 	= CMOS_READ(8)-1;
		time.tm_year 	= CMOS_READ(9);
    } while (time.tm_sec != CMOS_READ(0));

    BCD_TO_BIN(time.tm_sec);
    BCD_TO_BIN(time.tm_min);
    BCD_TO_BIN(time.tm_hour);
    BCD_TO_BIN(time.tm_mday);
    BCD_TO_BIN(time.tm_mon);
    BCD_TO_BIN(time.tm_year);

    startup_time = kernel_mktime(&time);
}

/*-------------------------------------------------------------------*
 * panic
 *
 * Ŀο ġ  ߻, ޼ ϰ
 * ߰ ˴ϴ.
 *-------------------------------------------------------------------*/
volatile void panic(const char * fmt, ...)
{
    static char buf[1024];
    va_list args;

    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    va_end(args);
    kprintf("Kernel panic: %s\n",buf);
    if (current == current_gover->citizen[0])
        kprintf("In swapper task - not syncing\n");
    else
        sys_sync();
    for(;;);
}

/*-------------------------------------------------------------------*
 * halt
 *
 * ý halt ŵϴ.
 *-------------------------------------------------------------------*/
void halt(void)
{
#ifdef DEBUG
    kprintf("\nhalt() is called. System halted.");
#endif
    while (TRUE)                 /* do nothing */
        asm("cli;hlt");          /* give the CPU some rest */
}

/*-------------------------------------------------------------------*
 * reboot
 *
 * ý reboot ŵϴ.
 *-------------------------------------------------------------------*/
void reboot(void)
{
    int k;
#ifdef DEBUG
    kprintf("\nRebooting...");
#endif
    outb(0x3f2,0x00);
    outb(0x64,0xfe);
    for (k=0; k<4000; k++);
    outb(0x64,0xff);
    asm("cli;hlt");
    while (1) ;
}

/*-------------------------------------------------------------------*
 * init
 *
 * Ŀ ʱȭ ̳  忡 Ǵ ù
 * ԼԴϴ. ϵ  νϰ, shell  оͼ
 * װ Ͽ,   ѱ  ˴ϴ.
 *-------------------------------------------------------------------*/
void init(void)
{
	int i;

	kprintf("in init functions\n");
	
	setup();
	(void) open("/dev/tty0",O_RDWR,0);
	(void) dup(0);
	(void) dup(0);
    if ((i=fork())<0)
        kprintf("Fork failed in init\r\n");
    else if (!i) {
        close(0);close(1);close(2);
        setsid();
        (void) open("/dev/tty0",O_RDWR,0);
        (void) dup(0);
        (void) dup(0);
		/* ..   /dev/sh  ġ ִ ǰ? T_T */
		/* FIXME :  /bin/sh  ּ. */
        _exit(execve("/dev/w_sh",argv,envp));
    }
	/* T_T  loop ų ֽϴ. :-[p] */
	for(;;);
}
