/*-------------------------------------------------------------------*
 * kernel/equip.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/equip.h>
#include <weix/types.h>

/*-------------------------------------------------------------------*
 * get_BIOS_equiplist from junOS
 *-------------------------------------------------------------------*/
void get_BIOS_equiplist(char is_IBMPS, struct equip_list *equip)
{
	/*
	   Background info:
	   The equipment in a computer is represented by one dword in the BIOS
	   data area, at 0x040:0x010.  We take this and convert it to our own
	   equipment list format.  Note that you cannot do a BIOS call to fetch
	   the info (int 0x11) because we are in protected mode !!!
	   EQAD is the absolute address of the BIOS equipment data dword
	*/

 	ulong BIOS_equip = *((ulong *)EQAD); /* The BIOS equipment info  */

	/* Check for floppies - this is bit 0 of the equipment dword     */
	/* Bits 6+7 contain the amount of floppy drives in the system    */

	if(BIOS_equip & 0x1)
	{
		equip->floppies = ((BIOS_equip & 0x0c0) >> 6) + 1; /* How many ?  */
	} else equip->floppies = 0;

	/* Is there a coprocessor ?  This is bit 1               */

	equip->has_coprocessor = (BIOS_equip & 0x02) != 0;

	/* Now we make a distinction between an IBM PS/1 or PS/2 computer    */
	/* (maybe even Aptiva ?) and a _real_ PC ;)              */
	/* IBMs store the availability of a mouse here, PCs the RAM banks    */
	/* Get bits 2 and 3 */
	if(is_IBMPS)
	{
		equip->mix.has_mouse = (BIOS_equip & 0x04) != 0;
	} else equip->mix.memory_banks = (BIOS_equip & 0x0c) >> 2;

	/* Get the initial video mode (video mode at boot -- bit 4 and 5)    */

	equip->initial_videomode = (BIOS_equip & 0x030) >> 4;

	/* Does this computer support DMA transfers ?  (sure hope so !!!)    */
	/* This is bit 8                         */
	equip->has_DMA_support = (BIOS_equip & 0x0100) != 0;

	/* Now we get the amount of RS232 serial ports (bits 9-11)       */
	equip->serial_ports = (BIOS_equip & 0x0e00) >> 9;

	/* Does this computer have a gameport ?  (bit 12)            */
	equip->has_gameport = (BIOS_equip & 0x01000) != 0;

	/* Does this computer have an internal modem ? (bit 13)      */
	equip->has_internal_modem = (BIOS_equip & 0x02000) != 0;

	/* How many printer ports do we have ?               */
	equip->printer_ports = (BIOS_equip & 0x0c000) >> 14;
}
