/*-------------------------------------------------------------------
 * exception_isr.s
 * Description:
 * Exception Interrupt Service Routines.
 * Errorcode 0 is pushed onto stack if not done by hardware.
 * Then the exception number is pushed onto the stack. From that
 * point all exceptions are handled identically.
 *
 * Segment registers are set up and general exception service
 * routine is called.
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/segment.h>

.text

.globl __divide_error, __debug_exception, __non_maskable_interrupt, __breakpoint, __overflow, __bounds_check, __invalid_opcode, __coprocessor_not_available, __double_fault, __coprocessor_segment_overrun, __invalid_TSS, __segment_not_present, __stack_exception, __general_protection, __page_fault	, __reserved, __coprocessor_error, __alignment_check

/*-------------------------------------------------------------------*
 *                           __divide_error
 *-------------------------------------------------------------------*/
	.align 	4
__divide_error:
	pushl	$0
	pushl	$0
	jmp	__all_exceptions
/*-------------------------------------------------------------------*
 *                           __debug_exception
 *-------------------------------------------------------------------*/
	.align	4
__debug_exception:
	pushl	$0
	pushl	$1
	jmp	__all_exceptions
/*-------------------------------------------------------------------*
 *                           __non_maskable_interrupt
 *-------------------------------------------------------------------*/
	.align	4
__non_maskable_interrupt:
	push	$0
	push	$2
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                           __breakpoint
 *-------------------------------------------------------------------*/
	.align	4
__breakpoint:
	push	$0
	push	$3
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                           __overflow
 *-------------------------------------------------------------------*/
	.align	4
__overflow:
	push	$0
	push	$4
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                           __bounds_check
 *-------------------------------------------------------------------*/
	.align	4
__bounds_check:
	push	$0
	push	$5
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                        __invalid_opcode
*-------------------------------------------------------------------*/
	.align	4
__invalid_opcode:
	push	$0
	push	$6
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                  __coprocessor_not_available
 *-------------------------------------------------------------------*/
	.align	4
__coprocessor_not_available:
	push	$0
	push	$7
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                           __double_fault
 *-------------------------------------------------------------------*/
	.align	4
__double_fault:
	push	$8
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                __coprocessor_segment_overrun
 *-------------------------------------------------------------------*/
	.align	4
__coprocessor_segment_overrun:
	push	$0
	push	$9
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                           __invalid_TSS
 *-------------------------------------------------------------------*/
	.align	4
__invalid_TSS:
	push	$10
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                           __segment_not_present
 *-------------------------------------------------------------------*/
	.align	4
__segment_not_present:
	push	$11
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                           __stack_exception
*-------------------------------------------------------------------*/
	.align	4
__stack_exception:
	push	$12
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                           __general_protection
 *-------------------------------------------------------------------*/
	.align	4
__general_protection:	
	push	$13
	jmp	__all_exceptions


/*-------------------------------------------------------------------*
 *                           __page_fault
 *-------------------------------------------------------------------*/
	.align	4
__page_fault:	
	push	$14
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                           __reserved
 *-------------------------------------------------------------------*/
	.align	4
__reserved:
	push	$0
	push	$15
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                           __coprocessor_error
 *-------------------------------------------------------------------*/	
	.align	4
__coprocessor_error:
	push	$0
	push	$16
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                           __alignment_check
 *-------------------------------------------------------------------*/
	.align	4
__alignment_check:
	push	$0
	push	$17
	jmp	__all_exceptions

/*-------------------------------------------------------------------*
 *                           __all_exceptions
 *-------------------------------------------------------------------*/
__all_exceptions:
	push	%ds
	push	%es
	push	%fs
	push	%gs
	pusha

	movw	$(USER_DS), %ax		 /* reset ds to default */
	movw	%ax, %ds

	/* exception_handler  ȣմϴ. */
	call	exception_handler

	popa
	pop		%gs
	pop		%fs
	pop		%es
	pop		%ds

	add		$8, %esp
	iret
