/*-------------------------------------------------------------------*
 *  exception.c
 *
 * Description:
 * General exception handler. Called by all exception ISR's.
 * The number of the original exception and errorcode are passed as
 * parameters.
 *-------------------------------------------------------------------*/
/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/types.h>
#include <asm/segment.h>
#include <weix/kernel.h>
#include <weix/idt.h>
#include <weix/exception.h>

#ifdef DEBUG
#include <weix/kdebug.h>
#endif

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
typedef char ExceptionMessage[32];

/*-------------------------------------------------------------------*
 * setup_exceptions
 *
 * Exception  κ IDT  ϴ κԴϴ.
 *-------------------------------------------------------------------*/
void setup_exceptions(void)
{
  interrupt_IDT(&IDT[0], KERNEL_CS, (dword)&__divide_error, 0);
  interrupt_IDT(&IDT[1], KERNEL_CS, (dword)&__debug_exception, 0);
  interrupt_IDT(&IDT[2], KERNEL_CS, (dword)&__non_maskable_interrupt, 0);
  interrupt_IDT(&IDT[3], KERNEL_CS, (dword)&__breakpoint, 3); /* can be called from user mode */
  interrupt_IDT(&IDT[4], KERNEL_CS, (dword)&__overflow, 0);
  interrupt_IDT(&IDT[5], KERNEL_CS, (dword)&__bounds_check, 0);
  interrupt_IDT(&IDT[6], KERNEL_CS, (dword)&__invalid_opcode, 0);
  interrupt_IDT(&IDT[7], KERNEL_CS, (dword)&__coprocessor_not_available, 0);
  interrupt_IDT(&IDT[8], KERNEL_CS, (dword)&__double_fault, 0);
  interrupt_IDT(&IDT[9], KERNEL_CS, (dword)&__coprocessor_segment_overrun, 0);
  interrupt_IDT(&IDT[10], KERNEL_CS, (dword)&__invalid_TSS, 0);
  interrupt_IDT(&IDT[11], KERNEL_CS, (dword)&__segment_not_present, 0);
  interrupt_IDT(&IDT[12], KERNEL_CS, (dword)&__stack_exception, 0);
  interrupt_IDT(&IDT[13], KERNEL_CS, (dword)&__general_protection, 0);
  interrupt_IDT(&IDT[14], KERNEL_CS, (dword)&__page_fault, 0);
  interrupt_IDT(&IDT[15], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[16], KERNEL_CS, (dword)&__coprocessor_error, 0);
  interrupt_IDT(&IDT[17], KERNEL_CS, (dword)&__alignment_check, 0);
  interrupt_IDT(&IDT[18], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[19], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[20], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[21], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[22], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[23], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[24], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[25], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[26], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[27], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[28], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[29], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[30], KERNEL_CS, (dword)&__reserved, 0);
  interrupt_IDT(&IDT[31], KERNEL_CS, (dword)&__reserved, 0);
}

static const ExceptionMessage exceptionDescription[N_EXCEPTIONS] =
{ "Divide Error",
  "Debug Exception",
  "Non-Maskable Interrupt",
  "Breakpoint",
  "Overflow",
  "Bounds Check",
  "Invalid Opcode",
  "Coprocessor Not Available",
  "Double Fault",
  "Coprocessor Segment Overrun",
  "Invalid TSS",
  "Segment Not Present",
  "Stack Exception",
  "General Protection Fault",
  "Page Fault",
  "Reserved",
  "Coprocessor Error",
  "Alignment Check",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved",
  "Reserved"};

/*-------------------------------------------------------------------*
 * exception_handler
 *
 *  exception  Ͼ ȴٸ, װ óϴ
 * handler Դϴ.  weix page fault  κ
 * Ǿ  , ⺻ exception ߻ halt
 *  մϴ.
 *-------------------------------------------------------------------*/
void exception_handler(dword edi, dword esi, dword ebp, dword esp,
                             dword ebx, dword edx, dword ecx, dword eax,
                             dword gs, dword fs, dword es, dword ds,
                             dword nr, dword errorCode,
                             dword eip, dword cs, dword eflags)
{
  int i;
  dword *p = (dword *)(esp + (cs & 0x03 ? 44 : 36));
  kprintf("\nBlue Screen of Death :0(\n");
  kprintf("%s", exceptionDescription[nr]);
  switch (nr)
    {
    case 14:
      if (!(errorCode & 0x01))
	kprintf(" : non-present page\n");
      else
	kprintf(" : protection violation\n");
    default:
      kprintf("\nERROR CODE=%08X\n", errorCode);
      kprintf("EAX=%08X  EBX=%08X  ECX=%08X  EDX=%08X\n", \
		  eax, ebx, ecx, edx);
      kprintf("EBP=%08X  ESP=%08X  ESI=%08X  EDI=%08X\n", \
		  ebp, esp + (cs & 0x03 ? 44 : 36), esi, edi);
      kprintf("DS =%04X      ES =%04X      FS =%04X      GS =%04X\n", \
	 	  ds & 0xffff, es & 0xffff, fs & 0xffff, gs & 0xffff);
      kprintf("CS =%04X      EIP=%08X  EFLAGS=%08X\n", \
		  cs & 0xffff, eip, eflags);
      for(i=0;i<6;i+=6) {
          kprintf("STACK[%02d]=%08X %08X %08X %08X %08X %08X\n", i, p[i], p[i+1],
					p[i+2], p[i+3], p[i+4], p[i+5]);
      }
      break;
    }
	show_CPU_reg();
  if (nr != 3) {
     halt();
  }
}
