/*-------------------------------------------------------------------*
 * kernel/fork.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/sched.h>
#include <weix/errno.h>
#include <weix/string.h>
#include <weix/kernel.h>
#include <weix/ptrace.h>
#include <weix/linkage.h>
#include <asm/fs_func.h>
#include <asm/segment.h>
#include <weix/page.h>
#include <asm/system.h>
#include <weix/kdebug.h>

#include <weix/extern_func_T_T.h>
/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
int find_empty_process(void);
struct wait_queue * find_empty_run_queue(void);

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
long last_pid=0;

/*-------------------------------------------------------------------*
 * verify_area
 *-------------------------------------------------------------------*/
void verify_area(void * addr,int size)
{
	unsigned long start;

	start = (unsigned long) addr;
	size += start & 0xfff;
	start &= 0xfffff000;
	start += get_base(&(current->ldt[2]));
	while (size>0) {
		size -= 4096;
		write_verify(start);
		start += 4096;
	}
}

/*-------------------------------------------------------------------*
 * copy_mem
 *-------------------------------------------------------------------*/
int copy_mem(int nr,struct my_life_schedule * p)
{
	unsigned long old_data_base,new_data_base,data_limit;
	unsigned long old_code_base,new_code_base,code_limit;

	code_limit=get_limit(0x0f);
	data_limit=get_limit(0x17);

#if 0
	 kprintf("code_limit=%08X data_limit=%08X\n", code_limit, data_limit);
#endif

	/* current->ldt[0] dummy ŵϴ.
	 * current->ldt[1] code ׸Ʈ ŵϴ.
	 * current->ldt[2] data/stack  ŵϴ. */
	old_code_base = get_base(current->ldt+1);
	old_data_base = get_base(current->ldt+2);

#if 0
	kprintf("old_code_base=%08X old_data_base=%08X\n", old_code_base, old_data_base);
#endif

	if (old_data_base != old_code_base)
		panic("We don't support separate I&D");
	if (data_limit < code_limit)
		panic("Bad data_limit");
#if 0
	kprintf("TASK_SIZE = %d, ", TASK_SIZE);
#endif
	new_data_base = new_code_base = nr * TASK_SIZE;
	set_base(p->ldt+1, new_code_base);
	set_base(p->ldt+2, new_data_base);

#if 0
	kprintf("old_data_base = %08X, new_data_base = %08X\n", old_data_base, new_data_base);
#endif
	if (copy_page_tables(old_data_base,new_data_base,data_limit)) {
		free_page_tables(new_data_base,data_limit);
		return -ENOMEM;
	}
	return 0;
}

/*-------------------------------------------------------------------*
 * sys_fork
 *-------------------------------------------------------------------*/
asmlinkage int sys_fork(struct pt_regs r)
{
    int nr, i;
    struct file *f;
    struct my_life_schedule *p;
	struct wait_queue *r_q;

    if ((nr=find_empty_process()) < 0)
        return -EAGAIN;

#if 0
	kprintf("in process - first\n");
    kprintf("current=%08X, *current=%08X\n", current, *current);
	kprintf("nr=%X, ebx=%08X, ecx=%08X, edx=%08X, esi=%08X\n", nr, r.ebx, r.ecx, r.edx, r.esi);
	kprintf("edi=%08X, ebp=%08X, eax=%08X, ds=%08X\n", r.edi, r.ebp, r.eax, r.ds);
	kprintf("es=%08X, fs=%08X, gs=%08X, o_eax=%08X, eip=%08X\n", r.es, r.fs, r.gs, r.orig_eax, r.eip);
	kprintf("cs=%08X, eflags=%08X, esp=%08X, ss=%08X\n", r.cs, r.eflags, r.esp, r.ss);

	dump_memory((ulong *)current, 32);
#endif

	/* μ ̺   ɴϴ. */
	if ((p=(struct my_life_schedule *) get_free_page()) == 0) {
		/* insert swapping functions in later version */
		//kprintf("no free page");
		return -EAGAIN;
	}

	memset((void *)p, 0, PAGE_SIZE);

	/* θ μ ڽĿ մϴ. */
	*p = *current;		/* NOTE!  κп   ʽϴ. */

	/* ڽ  κп   մϴ. */
	p->state = TASK_UNINTERRUPTIBLE;
	p->pid = last_pid;
	p->counter = p->my_priority;
	p->signal = 0;
	p->alarm = 0;
	p->leader = 0;		/* μ leadership  ڽĿ  ʽϴ. */
	p->utime = p->stime = 0;
	p->cutime = p->cstime = 0;
	p->start_time = jiffies;
	p->tss.back_link = 0;
	p->tss.esp0 = PAGE_SIZE + (long) p;
	p->tss.ss0 = KERNEL_DS;
	p->tss.eip = r.eip;
	p->tss.eflags = r.eflags;
	p->tss.eax = 0;
	p->tss.ecx = r.ecx;
	p->tss.edx = r.edx;
	p->tss.ebx = r.ebx;
	p->tss.esp = r.esp;
	p->tss.ebp = r.ebp;
	p->tss.esi = r.esi;
	p->tss.edi = r.edi;
	p->tss.es = r.es & 0xffff;
	p->tss.cs = r.cs & 0xffff;
	p->tss.ss = r.ss & 0xffff;
	p->tss.ds = r.ds & 0xffff;
	p->tss.fs = r.fs & 0xffff;
	p->tss.gs = r.gs & 0xffff;
#if 0
	kprintf("FORK_NR = %d, ", nr);
#endif
	p->tss.ldt = _LDT(nr);
	p->tss.trace_bitmap = 0x80000000;

#if 0
	dump_memory((ulong *)p, 32);
	kprintf("in process - second\n");
#endif

	if (copy_mem(nr,p)) {
		current_gover->citizen[nr] = NULL;
		free_page((long) p);
		return -EAGAIN;
	}

	for (i=0; i< NR_OPEN ; i++) {
		if ((f = p->filp[i])) {
			f->f_count++;
		}
	}

	if (current->pwd)
		current->pwd->i_count++;
	if (current->root)
		current->root->i_count++;
#if 0
	kprintf("in process - third\n");
#endif

	/* ڽ μ TSS, LDT descriptor  GDT ߰մϴ. */
    setup_descriptor(&GDT[(nr<<1)+FIRST_TSS_ENTRY], (ulong )(&(p->tss)), sizeof(p->tss), 9, 0, 0, 0);
    setup_descriptor(&GDT[(nr<<1)+FIRST_LDT_ENTRY], (ulong )(&(p->ldt)), sizeof(p->ldt), 2, 0, 0, 0);

    p->state = TASK_RUNNING;
	current_gover->citizen[nr] = p;	/* ڽ μ  繫 ȣ ȣ մϴ. */

	/*  ڽ μ run_queue  Դϴ. */
	cli();
	r_q = find_empty_run_queue();
	if(r_q == NULL)
		panic("find_empty_run_queue return -1, No empty run_queue\n");
	r_q->task = p;
	sti();

#if 0
	show_task_struct(p);
#endif

#if 0
	kprintf("exited in sys_fork -> last_pid=%d\n", last_pid);
#endif
	return last_pid;
}

/*-------------------------------------------------------------------*
 * find_empty_run_queue
 *
 * run_queue_ptr  Ͽ task ׸ NULL  ׸
 * ã մϴ.
 *-------------------------------------------------------------------*/
struct wait_queue * find_empty_run_queue(void)
{
	int i=0;
	struct wait_queue * t=run_queue_ptr;

	while(1) {
		if(t->task == NULL)
			return t;

		if(i>MAX_NR_RUN_QUEUE)
			panic("NO EMPTY RUN QUEUE\n");

		t = t->next;
		i++;
	}

	return NULL;
}

/*-------------------------------------------------------------------*
 * find_empty_process
 *
 *  繫  ã   °ΰ
 * մϴ.
 *-------------------------------------------------------------------*/
int find_empty_process(void)
{
	int i;

	repeat:
		if ((++last_pid)<0) last_pid=1;
		for(i=0 ; i<NR_TASKS ; i++) {
			if (current_gover->citizen[i] && current_gover->citizen[i]->pid == last_pid) {
				goto repeat;
			}
		}
	for(i=1 ; i<NR_TASKS ; i++) {
		if (!current_gover->citizen[i]) {
			// kprintf("empty_process = %d\n", i);
			return i;
		}
	}
	return -EAGAIN;
}
