/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/mm.h>
#include <weix/fs.h>
#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/string.h>
#include <asm/system.h>

#include <weix/kdebug.h>

#include <weix/extern_func_T_T.h>

#if (BUFFER_END & 0xfff)
#error "Bad BUFFER_END value"
#endif

#if (BUFFER_END > 0xA0000 && BUFFER_END <= 0x100000)
#error "Bad BUFFER_END value"
#endif

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
struct buffer_head * start_buffer = (struct buffer_head *) &end;
/* NR_HASH = 307 */
struct buffer_head * hash_table[NR_HASH];
static struct buffer_head * free_list;
static struct my_life_schedule * buffer_wait = NULL;
int nr_buffers = 0;
/* 
 * this variable is used point last used buffer
 * if this broken, WEIX is goto hell. :-) 
 */
static void * last_used_buffer_ptr = (void *) BUFFER_END;

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
inline void wait_on_buffer(struct buffer_head * bh);
static inline void remove_from_queues(struct buffer_head * bh);

/*-------------------------------------------------------------------*
 * wait_on_buffer
 *-------------------------------------------------------------------*/
inline void wait_on_buffer(struct buffer_head * bh)
{
	cli();
	while (bh->b_lock) {
		kprintf("wait_on_buffer: oh~ baby. don't sleep plz\n");
		sleep_on(&bh->b_wait);
	}
	sti();
}

/*-------------------------------------------------------------------*
 * sys_sync
 *-------------------------------------------------------------------*/
int sys_sync(void)
{
    return 0;
}

#define _hashfn(dev,block) (((unsigned)(dev^block))%NR_HASH)
/* if dev=300, block=0, value of _hashfn is 300 */
#define hash(dev,block) hash_table[_hashfn(dev,block)]

/*-------------------------------------------------------------------*
 * minix_old_sync_dev
 *-------------------------------------------------------------------*/
static int minix_old_sync_dev(int dev)
{
    int i;
    struct buffer_head * bh;

    bh = start_buffer;
    for (i=0 ; i<nr_buffers ; i++,bh++) {
        if (bh->b_dev != dev)
            continue;
        wait_on_buffer(bh);
        if (bh->b_dirt)
            ll_rw_block(WRITE,1,&bh);
    }
    return 0;
}

/*-------------------------------------------------------------------*
 * remove_from_queues
 *-------------------------------------------------------------------*/
static inline void remove_from_queues(struct buffer_head * bh)
{
/* remove from hash-queue */
	if (bh->b_next)
		bh->b_next->b_prev = bh->b_prev;
	if (bh->b_prev)
		bh->b_prev->b_next = bh->b_next;
	if (hash(bh->b_dev,bh->b_blocknr) == bh)
		hash(bh->b_dev,bh->b_blocknr) = bh->b_next;
/* remove from free list */
	if (!(bh->b_prev_free) || !(bh->b_next_free))
		panic("Free block list corrupted");
	bh->b_prev_free->b_next_free = bh->b_next_free;
	bh->b_next_free->b_prev_free = bh->b_prev_free;
	if (free_list == bh)
		free_list = bh->b_next_free;
}

/*-------------------------------------------------------------------*
 * insert_into_queues
 *-------------------------------------------------------------------*/
static inline void insert_into_queues(struct buffer_head * bh)
{
/* put at end of free list */
	//kprintf("insert_into_queues: in functions\n");
	bh->b_next_free = free_list;
	bh->b_prev_free = free_list->b_prev_free;
	free_list->b_prev_free->b_next_free = bh;
	free_list->b_prev_free = bh;
/* put the buffer in new hash-queue if it has a device */
	bh->b_prev = NULL;
	bh->b_next = NULL;
	if (!bh->b_dev) {
		return;
	}
	bh->b_next = hash(bh->b_dev,bh->b_blocknr);
	//kprintf("bh->b_dev=%x, bh->b_blocknr=%x\n", bh->b_dev, bh->b_blocknr);
	/*
	 * if bh->b_dev = 300, bh->b_blocknr = 0 
	 * hash(bh->b_dev, bh->b_blocknr) is
	 * hash_table[300] = bh;
	 */
	hash(bh->b_dev,bh->b_blocknr) = bh;
	if (bh->b_next)
        bh->b_next->b_prev = bh;
}

/*-------------------------------------------------------------------*
 * find_buffer
 *-------------------------------------------------------------------*/
static struct buffer_head * find_buffer(int dev, int block, int size)
{		
	struct buffer_head * tmp;

	//kprintf("find_buffer: in functions\n");
	/* hash(dev, block) = struct buffer_head hash_table[_hashfn(dev, block] */
	for (tmp = hash(dev,block) ; tmp != NULL ; tmp = tmp->b_next)
		if (tmp->b_dev==dev && tmp->b_blocknr==block)
			return tmp;
	return NULL;
}

/*-------------------------------------------------------------------*
 * get_hash_table
 *-------------------------------------------------------------------*/
struct buffer_head * get_hash_table(int dev, int block, int size)
{
	struct buffer_head * bh;

	//kprintf("get_hash_table : in function\n");
repeat:
	if (!(bh=find_buffer(dev,block, size))) {
    	//kprintf("bh->b_dev=%x, bh->b_blocknr=%x, bh->uptodate=%x\n", bh->b_dev, bh->b_blocknr, bh->b_uptodate);
    	//kprintf("bh->dirt=%x, bh->b_count=%x, bh->b_lock=%x\n", bh->b_dirt, bh->b_count, bh->b_lock);
		return NULL;
	}
	bh->b_count++;
	wait_on_buffer(bh);
	if (bh->b_dev != dev || bh->b_blocknr != block) {
		brelse(bh);
		goto repeat;
	}
	return bh;
}


/*-------------------------------------------------------------------*
 * get_block
 *-------------------------------------------------------------------*/
#define BADNESS(bh) (((bh)->b_dirt<<1)+(bh)->b_lock)
struct buffer_head * get_block(int dev, int block, int size)
{
    struct buffer_head * tmp, * bh;

repeat:
	//kprintf("get_block : \n");
	/* 
	 * in first time, get_hash_table can't find buffer because have no
	 * datas equal with buffer_head->b_dev, buffer_head->block in
	 * hash table
	 */ 
	bh = get_hash_table(dev,block, size);
	//kprintf("get_block: bh = %08X\n", bh);
    if (bh)
        return bh;

    tmp = free_list;
	//kprintf("get_block: tmp = %08X\n", tmp);
    do {
        if (tmp->b_count)
            continue;
        if (!bh || BADNESS(tmp)<BADNESS(bh)) {
            bh = tmp;
            if (!BADNESS(tmp))
                break;
        }
/* and repeat until we find something good */
    } while ((tmp = tmp->b_next_free) != free_list);

    if (!bh) {
		//kprintf("process : sleep_on\n");
		sleep_on(&buffer_wait);
		goto repeat;
    }

    wait_on_buffer(bh);
    if (bh->b_count)
        goto repeat;

    while (bh->b_dirt) {
        minix_old_sync_dev(bh->b_dev);
        wait_on_buffer(bh);
        if (bh->b_count)
            goto repeat;
    }
/* NOTE!! While we slept waiting for this block, somebody else might */
/* already have added "this" block to the cache. check it */
    if (find_buffer(dev,block, size))
        goto repeat;

/* OK, FINALLY we know that this buffer is the only one of it's kind, */
/* and that it's unused (b_count=0), unlocked (b_lock=0), and clean */
    bh->b_count=1;
    bh->b_dirt=0;
    bh->b_uptodate=0;
    remove_from_queues(bh);
    bh->b_dev=dev;
    bh->b_blocknr=block;
	last_used_buffer_ptr -= size;
/* inserted */
#ifdef DEBUG2
	kprintf("last_used_buffer_ptr=0x%08X\n", last_used_buffer_ptr);
#endif
	bh->b_data = (char *) last_used_buffer_ptr;
/* -------- */
    insert_into_queues(bh);
    return bh;
}

/*-------------------------------------------------------------------*
 * block_read == bread
 *-------------------------------------------------------------------*/
struct buffer_head * block_read(int dev, int block, int size)
{
    struct buffer_head * bh;

	//kprintf("block_read : block=%d, ", block);

	/* find buffer block. exist or not*/
    if (!(bh=get_block(dev,block, size)))
        panic("block_read: get_block returned NULL\n");

	//kprintf("get_block: bh->b_data = %08X\n", bh->b_data);

	bh->b_size=size;

	/* if buffer block exist, bh->b_uptodate == 1 */
    if (bh->b_uptodate)
        return bh;

	/* read real block */
    ll_rw_block(READ,1, &bh);

	/* yeah. i agree this is ugly. -_- */
you_must_wait_stupid_weix:
    if (!(bh->b_uptodate))
		/* in current, not realize hard disk timer so use this method */
		goto you_must_wait_stupid_weix;
	else 
        return bh;

	kprintf("oh~ run to your brelse. ");
    brelse(bh);
    return (NULL);
}

/*-------------------------------------------------------------------*
 * brelse
 *-------------------------------------------------------------------*/
void brelse(struct buffer_head * buf)
{
	//kprintf("brealse: bh->b_data=%08X\n", buf->b_data);
	if (!buf)
		return;
	//last_used_buffer_ptr += buf->b_size;
	wait_on_buffer(buf);
	if (!(buf->b_count--))
		panic("Trying to free free buffer");
	wake_up(&buffer_wait);
}


#define COPYBLK(from,to) \
__asm__("cld\n\t" \
    "rep\n\t" \
    "movsl\n\t" \
    ::"c" (BLOCK_SIZE/4),"S" (from),"D" (to) \
    )

/*
 * bread_page reads four buffers into memory at the desired address. It's
 * a function of its own, as there is some speed to be got by reading them
 * all at the same time, not waiting for one to be read, and then another
 * etc.
 */
/*-------------------------------------------------------------------*
 * bread_page
 *-------------------------------------------------------------------*/
void bread_page(unsigned long address,int dev,int b[4])
{
    struct buffer_head * bh[4];
    int i;

    for (i=0 ; i<4 ; i++)
        if (b[i]) {
            if ((bh[i] = get_block(dev,b[i], 1024)))
                if (!bh[i]->b_uptodate)
                    ll_rw_block(READ,1, &bh[i]);
        } else
            bh[i] = NULL;
    for (i=0 ; i<4 ; i++,address += BLOCK_SIZE)
        if (bh[i]) {
            wait_on_buffer(bh[i]);
            if (bh[i]->b_uptodate)
                COPYBLK((unsigned long) bh[i]->b_data,address);
            brelse(bh[i]);
        }
}

/*
 * Ok, breada can be used as bread, but additionally to mark other
 * blocks for reading as well. End the argument list with a negative
 * number.
 */
/*-------------------------------------------------------------------*
 * breada
 *-------------------------------------------------------------------*/
struct buffer_head * breada(int dev,int first, ...)
{
    va_list args;
    struct buffer_head * bh, *tmp;

    va_start(args,first);
    if (!(bh=get_block(dev,first, 1024)))
        panic("bread: get_block returned NULL\n");
    if (!bh->b_uptodate)
        ll_rw_block(READ,1, &bh);
    while ((first=va_arg(args,int))>=0) {
        tmp=get_block(dev,first, 1024);
        if (tmp) {
            if (!tmp->b_uptodate)
                ll_rw_block(READA, 1, &bh);
            tmp->b_count--;
        }
    }
    va_end(args);
    wait_on_buffer(bh);
    if (bh->b_uptodate)
        return bh;
    brelse(bh);
    return (NULL);
}

/*-------------------------------------------------------------------*
 * buffer_init
 *-------------------------------------------------------------------*/
void buffer_init(void)
{
	struct buffer_head * h = start_buffer;
	void * b = (void *) BUFFER_END;
	int i;

	while ( (b -= BLOCK_SIZE) >= ((void *) (h+1)) ) {
		h->b_dev = 0;
		h->b_dirt = 0;
		h->b_count = 0;
		h->b_lock = 0;
		h->b_uptodate = 0;
		h->b_wait = NULL;
		h->b_next = NULL;
		h->b_prev = NULL;
	/* deleted
		h->b_data = (char *) b;
	 ***************************/
		h->b_prev_free = h-1;
		h->b_next_free = h+1;
		h++;
		nr_buffers++;
		if (b == (void *) 0x100000)
			b = (void *) 0xA0000;
	}
	h--;
	free_list = start_buffer;
	free_list->b_prev_free = h;
	h->b_next_free = free_list;
	for (i=0;i<NR_HASH;i++)
		hash_table[i]=NULL;
}	
