/*-------------------------------------------------------------------*
 * kernel/fs/exec.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/sched.h>
#include <weix/errno.h>
#include <weix/stat.h>
#include <weix/fs.h>
#include <weix/elf.h>
#include <weix/kernel.h>
#include <weix/page.h>
#include <weix//mm.h>
#include <asm/fs_func.h>
#include <asm/segment.h>
#include <weix/types.h>
#include <weix/string.h>

#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
/*   ӽ÷  ־ϴ. */
unsigned long u_stack[1024];

/*-------------------------------------------------------------------*
 * read_elf_block
 *-------------------------------------------------------------------*/
int read_elf_block(struct inode * inode, struct Elf32_Phdr * ph, struct Elf32_Hdr * ex)
{
	struct buffer_head * tmp_bh;
	ulong c_offset=0, t_offset=0, d_offset=0;
    ulong start_text, start_data, end_text, end_data;
	ulong tmp_vaddr;
	int i, a, f_l=0, s_l=0, is_t_first_read=1, is_d_first_read=1;

    /* ó load  α׷ Header մϴ. */
    for(f_l = 0; f_l<ex->e_phnum; f_l++) {
        if(ph[f_l].p_type == PT_LOAD)
            break;
    }

    /*   մϴ. */
    for(a=0; a<ex->e_phnum; a++) {
        kprintf("p_type = %d, ", ph[a].p_type);
        kprintf("p_offset = %d, ", ph[a].p_offset);
        kprintf("p_vaddr = %08X, ", ph[a].p_vaddr);
        kprintf("p_filesz = %d, ", ph[a].p_filesz);
        kprintf("p_align = %X, ", ph[a].p_align);
    }
    /* ؽƮ  Ѵ. */
    end_text = ph[f_l].p_offset+ph[f_l].p_filesz;
	tmp_vaddr = ph[f_l].p_vaddr;
	kprintf(", text tmp_vdaddr = %x, ", tmp_vaddr);
	/* start_text  t_offset ʱȭ մϴ. */
	start_text = t_offset = ph[f_l].p_offset;
	c_offset = (t_offset/BLOCK_SIZE) * BLOCK_SIZE;
	/* Text  о Դϴ. */
	while(c_offset < end_text) {
		if(inode->u.ext2_i.i_data[(t_offset/BLOCK_SIZE)]) {
			kprintf(", c_offset = %d, t_offset = %d, text i_data = %d, ", c_offset, t_offset, (c_offset/BLOCK_SIZE));
    		if(!(tmp_bh = block_read(inode->i_dev, inode->u.ext2_i.i_data[(c_offset/BLOCK_SIZE)], BLOCK_SIZE))) {
        		iput(inode);
        		return -EACCES;
    		}
			c_offset += BLOCK_SIZE;
			/*    Text  Ѵٸ
			 * װ ޸𸮿 εմϴ.
			 * NOTE :  weix ø̼ text , data 
			 * 8M  ġ ϵ Ͽϴ.  8M  
			 * ҷ ERROR ߻  ֽϴ. */
			kprintf("t_offset = %d, c_offset = %d, ", t_offset, c_offset);
			if(t_offset < c_offset) {
				/*    ؽƮ    д ٸ,    */
				if(c_offset < end_text) {
					kprintf("[ta]");
					memcpy((void *)tmp_vaddr, (void *)(start_text+tmp_bh->b_data), (BLOCK_SIZE-start_text));
					tmp_vaddr += (BLOCK_SIZE - start_text);
					if(start_text)
						start_text = 0;
					if(is_t_first_read)
						is_t_first_read=0;
				} else {
					if(start_text) {
						kprintf("[tb]");
						/*  ̰ óд ̰, i_data[0]  ƴ϶ ELF  α׷ 忡
						 * شϴ κ д κп ܵǾ մϴ. ׷  ̻ 
						 *     Դϴ. Ʒ ڵ start_text 0 
						 * Ÿ īϴµ ־ ̻  ϴ. */
						if(is_t_first_read)
							start_text=0;
                    	/*    ؽƮ    ִٸ,    */
                    	memcpy((void *)tmp_vaddr, (void *)(start_text+tmp_bh->b_data), ph[f_l].p_filesz);
						brelse(tmp_bh);
                    	goto elf_read_data;
					} else {
						kprintf("[tc]");
                    	/*   ̶ */
                    	memcpy((void *)tmp_vaddr, (void *)(tmp_bh->b_data), (ph[f_l].p_filesz%BLOCK_SIZE));
					}
				}
			}
			brelse(tmp_bh);
		} else
			/* inode->u.ext2_i.i_data[]  0̸ ̻    ǹմϴ. */
			/*  11 ̻ ƴϱ⸦... */
			goto Oomm;
	}

/* ------------------------------------------------
 * Data  ޸𸮿 εմϴ.
 * ------------------------------------------------ */
elf_read_data:

	/*  Loadable α׷ Header ãϴ. */
    for(i=(f_l+1); i<ex->e_phnum; i++) {
        if(ph[i].p_type == PT_LOAD) {
			s_l = i;
            break;
		}
    }
    end_data = ph[s_l].p_offset+ph[s_l].p_filesz;
	tmp_vaddr = ph[s_l].p_vaddr;
	kprintf(", data tmp_vdaddr = %x, ", tmp_vaddr);
	start_data = d_offset = ph[s_l].p_offset;
	c_offset = (d_offset/BLOCK_SIZE) * BLOCK_SIZE;
	while(c_offset < end_data) {
		if(ph[s_l].p_filesz == 0)
			break;
        if(inode->u.ext2_i.i_data[(d_offset/BLOCK_SIZE)]) {
			kprintf(", c_offset = %d, data i_data = %d, ", c_offset, (c_offset/BLOCK_SIZE));
            if(!(tmp_bh = block_read(inode->i_dev, inode->u.ext2_i.i_data[(c_offset/BLOCK_SIZE)], BLOCK_SIZE))) {
                iput(inode);
                return -EACCES;
            }
            c_offset += BLOCK_SIZE;
			kprintf("d_offset = %d, c_offset = %d, ", d_offset, c_offset);
            if(d_offset < c_offset) {
                /*       д ٸ,    */
                if(c_offset < end_data) {
					kprintf("[da]");
                    memcpy((void *)tmp_vaddr, (void *)(start_data+tmp_bh->b_data), (BLOCK_SIZE-start_data));
                    tmp_vaddr += (BLOCK_SIZE - start_data);
                    if(start_data)
                        start_data = 0;
                    if(is_d_first_read)
                        is_d_first_read=0;
                } else {
                    if(start_data) {
                        /*       ִٸ,    */
						kprintf("[db]");
						/*  ̰ ó д Ÿ ̶ start_data 0 ϴ. 
						 *   ҽϴ.  . */
						if(is_d_first_read)
							start_data=0;
                        memcpy((void *)tmp_vaddr, (void *)(start_data+tmp_bh->b_data), ph[s_l].p_filesz);
						brelse(tmp_bh);
                        goto Oomm;
                    } else {
                        /*   ̶ */
						kprintf("[dc]");
                        memcpy((void *)tmp_vaddr, (void *)(tmp_bh->b_data), (ph[s_l].p_filesz%BLOCK_SIZE));
                    }
                }
            }
            brelse(tmp_bh);
        } else
            /* inode->u.ext2_i.i_data[]  0̸ ̻    ǹմϴ. */
            /*  11 ̻ ƴϱ⸦... */
            goto Oomm;
	}

Oomm:
	//dump_memory((void *)0x300000, 32);
	return 1;
}

/*-------------------------------------------------------------------*
 * do_execve
 *-------------------------------------------------------------------*/
int do_execve(unsigned long * eip, long tmp, char * filename, char ** argv, char ** envp)
{
	struct inode * inode;
	struct buffer_head * bh;
	struct Elf32_Hdr ex;
	struct Elf32_Phdr ph[6];
	int a, i, argc, envc;

	kprintf("in do_execve, ");

	/* ̸ ڷ ָ inode ü մϴ. */
	if (!(inode=namei(filename)))		/*  inode մϴ. */
		return -ENOENT;

	if (!S_ISREG(inode->i_mode)) {	/* Ϲ ̿ մϴ. */
		kprintf("must be regular file\n");
		iput(inode);
		return -EACCES;
	}

	/* 1024 Ʈŭ Ÿ нϴ. */
	if (!(bh = block_read(inode->i_dev,inode->u.ext2_i.i_data[0], BLOCK_SIZE))) {
		iput(inode);
		return -EACCES;
	}

	ex = *((struct Elf32_Hdr *) bh->b_data);

	if(ex.e_ident[EI_MAG0] == 0x7f && ex.e_ident[EI_MAG1] == 'E' &&
		ex.e_ident[EI_MAG2] == 'L' && ex.e_ident[EI_MAG3] == 'F') {
		kprintf("This file is ELF file\n");
	} else {
		kprintf("This file not supported in weix\n");
		return -ENOEXEC;
	}

	/*  մϴ. */
	kprintf("e_type = %d,", ex.e_type);
	kprintf("e_machine = %d, ", ex.e_machine);
	kprintf("e_version = %d, ", ex.e_version);
	kprintf("e_ehsize = %d, ", ex.e_ehsize);
	kprintf("e_phentsize = %d, ", ex.e_phentsize);
	kprintf("e_phnum = %d, ", ex.e_phnum);

	/*    Ŵϴ. */
	for(a=0; a<ex.e_phnum; a++)
		ph[a] = *((struct Elf32_Phdr *) (bh->b_data + ex.e_ehsize + (a*ex.e_phentsize)));

    for(i=0; i<ex.e_phnum; i++)
        if(ph[i].p_type == PT_LOAD)
            break;

	/*  Ÿ о Դϴ. */
	i = read_elf_block(inode, ph, &ex);
	if (!i)
		panic("Error read_elf_block\n");

	kprintf("e_entry = %08X\n", ex.e_entry);
	/* ⿡ ÿ ִ EIP ٲϴ. ׷  ɱ?
	 * iret  ϰ Ǹ ÿ ִ ټ ׸ Ʈ ˴ϴ.
	 * H +----------------+
	 *   |    USER_DS     |
	 *   |   STACK_PTR    |
	 *   |     EFLAGS     |
	 *   |    USER_CS     |
	 *   |      EIP       |
	 * L +----------------+
	 * ׷ EIP ٲ Ǹ, ư Ǵ ġ ޶ ϴ. ,
	 * ex.e_entry  Ű ġ  ϰ ˴ϴ. */
	
	/* ..  ȹ  EIP[0]  EIP  ־ ϰ
	 * EIP[3]  ESP   ־ մϴ. ٵ     𸣰
	 *  EIP[1]  EIP  , EIP[4] ESP   ֽϴ. */
	eip[1] = ex.e_entry;
	eip[4] = (ulong )(u_stack+4096);
	return 0;
}
