/*-------------------------------------------------------------------*
 * kernel/fs/ext2/dir.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/fs.h>
#include <weix/ext2_fs.h>
#include <weix/types.h>
#include <weix/stat.h>
#include <weix/string.h>
#include <weix/dirent.h>
#include <asm/bitops.h>

#include <weix/kdebug.h>
#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * defines 
 *-------------------------------------------------------------------*/
#define ENOENT           2  
#define EACCES          13   
#define EIO              5      /* I/O error */
#define ENOSPC          28      /* No space left on device */
#define EFBIG           27      /* File too large */

#define suser() (current->euid == 0)
#define NAMEI_RA_CHUNKS  2
#define NAMEI_RA_BLOCKS  4
#define NAMEI_RA_SIZE        (NAMEI_RA_CHUNKS * NAMEI_RA_BLOCKS)
#define NAMEI_RA_INDEX(c,b)  (((c) * NAMEI_RA_BLOCKS) + (b))
#define clear_block(addr,size) \
    __asm__("cld\n\t" \
        "rep\n\t" \
        "stosl" \
        : \
        :"a" (0), "c" (size / 4), "D" ((long) (addr)) \
        )
#define in_range(b, first, len)     ((b) >= (first) && (b) <= (first) + (len) - 1)

/*-------------------------------------------------------------------*
 * ext2_match
 *-------------------------------------------------------------------*/
static int ext2_match (int len, const char * const name,
               struct ext2_dir_entry * de)
{
    unsigned char same;

    if (!de || !de->inode || len > EXT2_NAME_LEN)
        return 0;
    /*
     * "" means "." ---> so paths like "/usr/lib//libc.a" work
     */
    if (!len && de->name_len == 1 && (de->name[0] == '.') &&
       (de->name[1] == '\0'))
        return 1;
    __asm__("cld\n\t"
        "repe ; cmpsb\n\t"
        "setz %0"
        :"=q" (same)
        :"S" ((long) name), "D" ((long) de->name), "c" (len)
        );
    return (int) same;
}

/*-------------------------------------------------------------------*
 * ext2_mknod
 *-------------------------------------------------------------------*/
int ext2_mknod (struct inode *g, const char *gg, int ggg, int gggg, int ggggg)
{
    kprintf("in mknod\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_rename
 *-------------------------------------------------------------------*/
int ext2_rename (struct inode *h, const char *hh, int hhh,
            struct inode *hhhh, const char *hhhhh, int hhhhhh)
{
    kprintf("in ext2_rename\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_permission
 *-------------------------------------------------------------------*/
int ext2_permission (struct inode * inode, int mask)
{
    unsigned short mode = inode->i_mode;

#if 0
	kprintf("in ext2_permission\n");
#endif
    /*
     * Special case, access is always granted for root
     */
    if (suser ())
        return 1;
    /*
     * If no ACL, checks using the file mode
     */
    else if (current->euid == inode->i_uid)
        mode >>= 6;
    else if (in_group_p (inode->i_gid))
        mode >>= 3;
    if (((mode & mask & S_IRWXO) == mask))
        return 1;
    else
        return 0;
}

/*-------------------------------------------------------------------*
 * ext2_truncate
 *-------------------------------------------------------------------*/
void ext2_truncate (struct inode *j)
{
    kprintf("in ext2_truncate\n");
}

/*-------------------------------------------------------------------*
 * ext2_ioctl
 *-------------------------------------------------------------------*/
int ext2_ioctl (struct inode *a, struct file *b, unsigned int c,
               unsigned long d)
{
    kprintf("in ext2_ioctl\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_dir_read
 *-------------------------------------------------------------------*/
static int ext2_dir_read (struct inode * inode, struct file * filp,
                char * buf, int count)
{
    kprintf("in ext2_dir_read\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_readdir
 *-------------------------------------------------------------------*/
static int ext2_readdir (struct inode *a, struct file *b, struct dirent *c, int
d)
{
    kprintf("in ext2_readdir\n");
return 1;
}

/*-------------------------------------------------------------------*
 * file_fsync
 *-------------------------------------------------------------------*/
int file_fsync (struct inode *inode, struct file *filp)
{
    kprintf("in file_fsync\n");
return 1;
}

/*-------------------------------------------------------------------*
 * find_first_zero_byte
 *-------------------------------------------------------------------*/
static inline char * find_first_zero_byte (char * addr, int size)
{
    char *res;

    if (!size)
        return 0;
    __asm__("
        cld
        mov $0,%%eax
        repnz; scasb
        jnz 1f
        dec %%edi
1:      "
        : "=D" (res)
        : "0" (addr), "c" (size)
        : "ax");
    return res;
}

/*-------------------------------------------------------------------*
 * find_first_zero_bit
 *-------------------------------------------------------------------*/
static inline int find_first_zero_bit (unsigned long * addr, unsigned size)
{
    int res;

    if (!size)
        return 0;
    __asm__("
        cld
        movl $-1,%%eax
        repe; scasl
        je 1f
        subl $4,%%edi
        movl (%%edi),%%eax
        notl %%eax
        bsfl %%eax,%%edx
        jmp 2f
1:      xorl %%edx,%%edx
2:      subl %%ebx,%%edi
        shll $3,%%edi
        addl %%edi,%%edx"
        :"=d" (res)
        :"c" ((size + 31) >> 5), "D" (addr), "b" (addr)
        :"ax");
    return res;
}

/*-------------------------------------------------------------------*
 * find_next_zero_bit
 *-------------------------------------------------------------------*/
static inline int find_next_zero_bit (unsigned long * addr, int size,
                      int offset)
{
    unsigned long * p = ((unsigned long *) addr) + (offset >> 5);
    int set = 0, bit = offset & 31, res;

    if (bit) {
        /*
         * Look for zero in first byte
         */
        __asm__("
            bsfl %1,%0
            jne 1f
            movl $32, %0
1:          "
            : "=r" (set)
            : "r" (~(*p >> bit)));
        if (set < (32 - bit))
            return set + offset;
        set = 32 - bit;
        p++;
    }
    /*
     * No zero yet, search remaining full bytes for a zero
     */
    res = find_first_zero_bit (p, size - 32 * (p - addr));
    return (offset + set + res);
}

/*-------------------------------------------------------------------*
 * read_block_bitmap
 *-------------------------------------------------------------------*/
static void read_block_bitmap (struct super_block * sb,
                   unsigned int block_group,
                   unsigned long bitmap_nr)
{
    struct ext2_group_desc * gdp;
    struct buffer_head * bh;
   
    gdp = get_group_desc (sb, block_group, NULL);
    bh = block_read(sb->s_dev, gdp->bg_block_bitmap, 1024);
    if (!bh)
        panic ("read_block_bitmap:
                Cannot read block bitmap\n
                block_group = %d, block_bitmap = %lu",
                block_group, gdp->bg_block_bitmap);
    sb->u.ext2_sb.s_block_bitmap_number[bitmap_nr] = block_group;
    sb->u.ext2_sb.s_block_bitmap[bitmap_nr] = bh;
}

/*-------------------------------------------------------------------*
 * load_block_bitmap
 *-------------------------------------------------------------------*/
static inline int load_block_bitmap (struct super_block * sb,
                     unsigned int block_group)
{
    int i, j;
    unsigned long block_bitmap_number;
    struct buffer_head * block_bitmap;

    if (block_group >= sb->u.ext2_sb.s_groups_count)
        panic ("load_block_bitmap: 
                block_group >= groups_count\n
                block_group = %d, groups_count = %lu",
                block_group, sb->u.ext2_sb.s_groups_count);


    if (sb->u.ext2_sb.s_groups_count <= EXT2_MAX_GROUP_LOADED) {
        if (sb->u.ext2_sb.s_block_bitmap[block_group]) {
            if (sb->u.ext2_sb.s_block_bitmap_number[block_group] !=
                block_group)
                panic ("load_block_bitmap: 
                        block_group != block_bitmap_number");
            else
                return block_group;
        } else {
            read_block_bitmap (sb, block_group, block_group);
            return block_group;
        }
    }
    for (i = 0; i < sb->u.ext2_sb.s_loaded_block_bitmaps &&
            sb->u.ext2_sb.s_block_bitmap_number[i] != block_group; i++)
        ;
    if (i < sb->u.ext2_sb.s_loaded_block_bitmaps &&
        sb->u.ext2_sb.s_block_bitmap_number[i] == block_group) {
        block_bitmap_number = sb->u.ext2_sb.s_block_bitmap_number[i];
        block_bitmap = sb->u.ext2_sb.s_block_bitmap[i];
        for (j = i; j > 0; j--) {
            sb->u.ext2_sb.s_block_bitmap_number[j] =
                sb->u.ext2_sb.s_block_bitmap_number[j - 1];
            sb->u.ext2_sb.s_block_bitmap[j] =
                sb->u.ext2_sb.s_block_bitmap[j - 1];
        }
        sb->u.ext2_sb.s_block_bitmap_number[0] = block_bitmap_number;
        sb->u.ext2_sb.s_block_bitmap[0] = block_bitmap;
    } else {
        if (sb->u.ext2_sb.s_loaded_block_bitmaps < EXT2_MAX_GROUP_LOADED)
            sb->u.ext2_sb.s_loaded_block_bitmaps++;
        else
            brelse (sb->u.ext2_sb.s_block_bitmap[EXT2_MAX_GROUP_LOADED - 1]);
        for (j = sb->u.ext2_sb.s_loaded_block_bitmaps - 1; j > 0;  j--) {
            sb->u.ext2_sb.s_block_bitmap_number[j] =
                sb->u.ext2_sb.s_block_bitmap_number[j - 1];
            sb->u.ext2_sb.s_block_bitmap[j] =
                sb->u.ext2_sb.s_block_bitmap[j - 1];
        }
        read_block_bitmap (sb, block_group, 0);
    }
    return 0;
}

/*-------------------------------------------------------------------*
 * ext2_new_block
 *-------------------------------------------------------------------*/
int ext2_new_block (struct super_block * sb, unsigned long goal,
            unsigned long * prealloc_count,
            unsigned long * prealloc_block)
{
    struct buffer_head * bh;
    struct buffer_head * bh2;
    char * p, * r;
    int i, j, k, tmp;
    unsigned long lmap;
    int bitmap_nr;
    struct ext2_group_desc * gdp;
    struct ext2_super_block * es;

	kprintf("in ext2_new_block\n");
    if (!sb) {
        kprintf ("ext2_new_block: nonexistent device");
        return 0;
    }
    es = sb->u.ext2_sb.s_es;
    if (es->s_free_blocks_count <= es->s_r_blocks_count && !suser()) {
        return 0;
    }

#ifdef DEBUG
    kprintf ("goal=%lu.\n", goal);
#endif

repeat:
    /*
     * First, test whether the goal block is free.
     */
    if (goal < es->s_first_data_block || goal >= es->s_blocks_count)
        goal = es->s_first_data_block;
    i = (goal - es->s_first_data_block) / EXT2_BLOCKS_PER_GROUP(sb);
    gdp = get_group_desc (sb, i, &bh2);
    if (gdp->bg_free_blocks_count > 0) {
        j = ((goal - es->s_first_data_block) % EXT2_BLOCKS_PER_GROUP(sb));
        bitmap_nr = load_block_bitmap (sb, i);
        bh = sb->u.ext2_sb.s_block_bitmap[bitmap_nr];
#ifdef DEBUG
        kprintf ("goal is at %d:%d.\n", i, j);
#endif

        if (!test_bit(j, bh->b_data)) {
            goto got_block;
        }
        if (j) {
            lmap = ((((unsigned long *) bh->b_data)[j >> 5]) >>
                ((j & 31) + 1));
            if (j < EXT2_BLOCKS_PER_GROUP(sb) - 32)
                lmap |= (((unsigned long *) bh->b_data)[(j >> 5) + 1]) <<
                 (31 - (j & 31));
            else
                lmap |= 0xffffffff << (31 - (j & 31));
            if (lmap != 0xffffffffl) {
                __asm__ ("bsfl %1,%0"
                     : "=r" (k)
                     : "r" (~lmap));
                k++;
                if ((j + k) < EXT2_BLOCKS_PER_GROUP(sb)) {
                    j += k;
                    goto got_block;
                }
            }
        }

#ifdef DEBUG
        kprintf ("Bit not found near goal\n");
#endif
        p = ((char *) bh->b_data) + (j >> 3);
        r = find_first_zero_byte (p,
                      (EXT2_BLOCKS_PER_GROUP(sb) - j + 7) >> 3);
        k = (r - ((char *) bh->b_data)) << 3;
        if (k < EXT2_BLOCKS_PER_GROUP(sb)) {
            j = k;
            goto search_back;
        }
        k = find_next_zero_bit ((unsigned long *) bh->b_data,
                    EXT2_BLOCKS_PER_GROUP(sb),
                    j);
        if (k < EXT2_BLOCKS_PER_GROUP(sb)) {
            j = k;
            goto got_block;
        }
    }
#ifdef DEBUG
    kprintf ("Bit not found in block group %d.\n", i);
#endif

    for (k = 0; k < sb->u.ext2_sb.s_groups_count; k++) {
        i++;
        if (i >= sb->u.ext2_sb.s_groups_count)
            i = 0;
        gdp = get_group_desc (sb, i, &bh2);
        if (gdp->bg_free_blocks_count > 0)
            break;
    }
    if (k >= sb->u.ext2_sb.s_groups_count) {
        return 0;
    }

    bitmap_nr = load_block_bitmap (sb, i);
    bh = sb->u.ext2_sb.s_block_bitmap[bitmap_nr];
    r = find_first_zero_byte (bh->b_data,
                  EXT2_BLOCKS_PER_GROUP(sb) >> 3);
    j = (r - bh->b_data) << 3;
    if (j < EXT2_BLOCKS_PER_GROUP(sb))
        goto search_back;
    else
        j = find_first_zero_bit ((unsigned long *) bh->b_data,
                     EXT2_BLOCKS_PER_GROUP(sb));
    if (j >= EXT2_BLOCKS_PER_GROUP(sb)) {
        kprintf ("ext2_new_block: 
                Free blocks count corrupted for block group %d", i);
        return 0;
    }


search_back:
    /*
     * We have succeeded in finding a free byte in the block
     * bitmap.  Now search backwards up to 7 bits to find the
     * start of this group of free blocks.
     */
    for (k = 0; k < 7 && j > 0 && !test_bit (j - 1, bh->b_data); k++, j--);

got_block:

#ifdef DEBUG
    kprintf ("using block group %d(%d)\n", i, gdp->bg_free_blocks_count);
#endif

    tmp = j + i * EXT2_BLOCKS_PER_GROUP(sb) + es->s_first_data_block;

    if (test_opt (sb, CHECK_STRICT) &&
        (tmp == gdp->bg_block_bitmap ||
         tmp == gdp->bg_inode_bitmap ||
         in_range (tmp, gdp->bg_inode_table, sb->u.ext2_sb.s_itb_per_group)))
        panic ("ext2_new_block: 
                Allocating block in system zone\n
                block = %u", tmp);

    if (set_bit (j, bh->b_data)) {
        kprintf ("ext2_new_block:
                  bit already set for block %d", j);
        goto repeat;
    }

#ifdef DEBUG
    kprintf ("found bit %d\n", j);
#endif

    /*
     * Do block preallocation now if required.
     */
    j = tmp;

    bh->b_dirt = 1;
    if (sb->s_flags & MS_SYNC) {
        ll_rw_block (WRITE, 1, &bh);
		kprintf("-_- deleted wait_on_buffer(bh), ");
        /* wait_on_buffer (bh); */
    }

    if (j >= es->s_blocks_count) {
        kprintf ("ext2_new_block
                block >= blocks count\n
                block_group = %d, block=%d", i, j);
        return 0;
    }
    if (!(bh = get_block(sb->s_dev, j, 1024))) {
        kprintf ("ext2_new_block: cannot get block %d", j);
        return 0;
    }
    clear_block (bh->b_data, sb->s_blocksize);
    bh->b_uptodate = 1;
    bh->b_dirt = 1;
    brelse (bh);

    gdp->bg_free_blocks_count--;
    bh2->b_dirt = 1;
    es->s_free_blocks_count--;
    sb->u.ext2_sb.s_sbh->b_dirt = 1;
    sb->s_dirt = 1;
	return j;
	return 0;
}

/*-------------------------------------------------------------------*
 * ext2_alloc_block
 *-------------------------------------------------------------------*/
static int ext2_alloc_block (struct inode * inode, unsigned long goal)
{
    unsigned long result;

	result = ext2_new_block (inode->i_sb, goal, 0, 0);

	return result;
}

/*-------------------------------------------------------------------*
 * ext2_sync_inode
 *-------------------------------------------------------------------*/
int ext2_sync_inode (struct inode *inode)
{
	kprintf("in ext2_sync_inode\n");
	return 0;
}

/*-------------------------------------------------------------------*
 * ext2_free_blocks
 *-------------------------------------------------------------------*/
void ext2_free_blocks (struct super_block * sb, unsigned long block,
               unsigned long count)
{
	kprintf("in ext2_free_blocks\n");
}

/*-------------------------------------------------------------------*
 * inode_getblk
 *-------------------------------------------------------------------*/
static struct buffer_head * inode_getblk (struct inode * inode, int nr,
                      int create, int new_block, int * err)
{
    int tmp, goal = 0;
    unsigned long * p;
    struct buffer_head * result;
    int blocks = inode->i_sb->s_blocksize / 512;

	kprintf("in inode_getblk\n");
    p = inode->u.ext2_i.i_data + nr;
	kprintf("inode_getblk: nr=%d, p=%08X, create=%d, ", nr, p, create);

repeat:
    tmp = *p;
    if (tmp) {
		kprintf("Oh~~ tmp is not NULL tmp=%d\n", tmp);
		/* in weix block size 1024 */
		result = get_block(inode->i_dev, (tmp), 1024);
        if (tmp == *p)
            return result;
        brelse (result);
        goto repeat;
    }

    if (!create) {
        *err = -EFBIG;
        return NULL;
    }

    if (inode->u.ext2_i.i_next_alloc_block == new_block)
        goal = inode->u.ext2_i.i_next_alloc_goal;

#ifdef DEBUG
    kprintf ("hint = %d,", goal);
#endif

    if (!goal) {
        for (tmp = nr - 1; tmp >= 0; tmp--) {
            if (inode->u.ext2_i.i_data[tmp]) {
                goal = inode->u.ext2_i.i_data[tmp];
                break;
            }
        }
        if (!goal)
            goal = (inode->u.ext2_i.i_block_group *
                EXT2_BLOCKS_PER_GROUP(inode->i_sb)) +
                   inode->i_sb->u.ext2_sb.s_es->s_first_data_block;
    }

#ifdef DEBUG
    kprintf ("goal = %d.\n", goal);
#endif

    tmp = ext2_alloc_block (inode, goal);
    if (!tmp)
        return NULL;
    result = get_block(inode->i_dev, tmp, 1024);
    if (*p) {
        ext2_free_blocks (inode->i_sb, tmp, 1);
        brelse (result);
        goto repeat;
    }
    *p = tmp;
    inode->u.ext2_i.i_next_alloc_block = new_block;
    inode->u.ext2_i.i_next_alloc_goal = tmp;
    inode->i_ctime = CURRENT_TIME;
    inode->i_blocks += blocks;
    if (IS_SYNC(inode))
        ext2_sync_inode (inode);
    else
        inode->i_dirt = 1;
    return result;
}

/*-------------------------------------------------------------------*
 * block_getblk
 *-------------------------------------------------------------------*/
static struct buffer_head * block_getblk (struct inode * inode,
                      struct buffer_head * bh, int nr,
                      int create, int blocksize,
                      int new_block, int * err)
{
    int tmp, goal = 0;
    unsigned long * p;
    struct buffer_head * result;
    int blocks = inode->i_sb->s_blocksize / 512;

	kprintf("in block_getblk\n");
    if (!bh)
        return NULL;
    if (!bh->b_uptodate) {
        ll_rw_block (READ, 1, &bh);
        wait_on_buffer (bh);
        if (!bh->b_uptodate) {
            brelse (bh);
            return NULL;
        }
    }
    p = (unsigned long *) bh->b_data + nr;
repeat:
    tmp = *p;
    if (tmp) {
        result = get_block(bh->b_dev, tmp, 1024);
        if (tmp == *p) {
            brelse (bh);
            return result;
        }
        brelse (result);
        goto repeat;
    }
    if (!create) {
        brelse (bh);
        *err = -EFBIG;
        return NULL;
    }
    if (inode->u.ext2_i.i_next_alloc_block == new_block)
        goal = inode->u.ext2_i.i_next_alloc_goal;
    if (!goal) {
        for (tmp = nr - 1; tmp >= 0; tmp--) {
            if (((unsigned long *) bh->b_data)[tmp]) {
                goal = ((unsigned long *)bh->b_data)[tmp];
                break;
            }
        }
        if (!goal)
            goal = bh->b_blocknr;
    }
    tmp = ext2_alloc_block (inode, goal);
    if (!tmp) {
        brelse (bh);
        return NULL;
    }
    result = get_block(bh->b_dev, tmp, 1024);
    if (*p) {
        ext2_free_blocks (inode->i_sb, tmp, 1);
        brelse (result);
        goto repeat;
    }
    *p = tmp;
    bh->b_dirt = 1;
    if (IS_SYNC(inode)) {
        ll_rw_block (WRITE, 1, &bh);
        wait_on_buffer (bh);
    }
    inode->i_ctime = CURRENT_TIME;
    inode->i_blocks += blocks;
    inode->i_dirt = 1;
    inode->u.ext2_i.i_next_alloc_block = new_block;
    inode->u.ext2_i.i_next_alloc_goal = tmp;
    brelse (bh);
    return result;
}

/*-------------------------------------------------------------------*
 * ext2_getblk
 *-------------------------------------------------------------------*/
struct buffer_head * ext2_getblk (struct inode * inode, long block,
                  int create, int * err)
{
    struct buffer_head * bh;
    unsigned long b;
    unsigned long addr_per_block = EXT2_ADDR_PER_BLOCK(inode->i_sb);

	kprintf("in ext2_getblk\n");
    *err = -EIO;

    if (block < 0) {
        kprintf ("ext2_getblk: block < 0");
        return NULL;
    }

    if (block > EXT2_NDIR_BLOCKS + addr_per_block  +
            addr_per_block * addr_per_block +
            addr_per_block * addr_per_block * addr_per_block) {
        kprintf ("ext2_getblk: block > big");
        return NULL;
    }

#ifdef DEBUG
    kprintf ("block %lu, next %lu, goal %lu.\n", block,
            inode->u.ext2_i.i_next_alloc_block,
            inode->u.ext2_i.i_next_alloc_goal);
#endif

    if (block == inode->u.ext2_i.i_next_alloc_block + 1) {
        inode->u.ext2_i.i_next_alloc_block++;
        inode->u.ext2_i.i_next_alloc_goal++;
    }


    *err = -ENOSPC;
    b = block;
    if (block < EXT2_NDIR_BLOCKS)
        return inode_getblk (inode, block, create, b, err);

    block -= EXT2_NDIR_BLOCKS;
    if (block < addr_per_block) {
        bh = inode_getblk (inode, EXT2_IND_BLOCK, create, b, err);
        return block_getblk (inode, bh, block, create,
                     inode->i_sb->s_blocksize, b, err);
    }
    block -= addr_per_block;
    if (block < addr_per_block * addr_per_block) {
        bh = inode_getblk (inode, EXT2_DIND_BLOCK, create, b, err);
        bh = block_getblk (inode, bh, block / addr_per_block, create,
                   inode->i_sb->s_blocksize, b, err);
        return block_getblk (inode, bh, block & (addr_per_block - 1),
                     create, inode->i_sb->s_blocksize, b, err);
    }
    block -= addr_per_block * addr_per_block;
    bh = inode_getblk (inode, EXT2_TIND_BLOCK, create, b, err);
    bh = block_getblk (inode, bh, block/(addr_per_block * addr_per_block),
               create, inode->i_sb->s_blocksize, b, err);
    bh = block_getblk (inode, bh, (block/addr_per_block) & (addr_per_block - 1),
               create, inode->i_sb->s_blocksize, b, err);
    return block_getblk (inode, bh, block & (addr_per_block - 1), create,
                 inode->i_sb->s_blocksize, b, err);
	return (bh);
}

/*-------------------------------------------------------------------*
 * ext2_check_dir_entry
 *-------------------------------------------------------------------*/
int ext2_check_dir_entry (struct inode * dir, struct ext2_dir_entry * de, struct buffer_head * bh,
              unsigned long offset)
{
    char * error_msg = NULL;

    if (de->rec_len < EXT2_DIR_REC_LEN(1))
        error_msg = "rec_len is smaller than minimal";
    else if (de->rec_len % 4 != 0)
        error_msg = "rec_len % 4 != 0";
    else if (((char *) de - bh->b_data) + de->rec_len >
         dir->i_sb->s_blocksize)
        error_msg = "directory entry across blocks";

    if (error_msg != NULL)
        panic("bad directory entry: %s\n
                offset=%lu, inode=%lu, rec_len=%d, name_len=%d",
                error_msg, offset, de->inode, de->rec_len,
                de->name_len);
    return error_msg == NULL ? 1 : 0;
}

/*-------------------------------------------------------------------*
 * ext2_lookup
 *-------------------------------------------------------------------*/
int ext2_lookup (struct inode * dir, const char * name, int len,
         struct inode ** result)
{
    unsigned long ino;
	unsigned long offset;
	unsigned long * addr;
	char * dlimit;
	int i, tmp, block=0;
	int ok_sign=0;
    struct buffer_head * bh;
	struct buffer_head * bh2;
	struct ext2_dir_entry * de;
	struct super_block * sb;

#if 0
	kprintf("in ext2_lookup\n");
#endif
    *result = NULL;

	/* ̳  0  */
    if (!dir)
        return -ENOENT;

	sb = dir->i_sb;

	/* Ȯҷ ϴ ̳尡 丮  */
    if (!S_ISDIR(dir->i_mode)) {
        iput (dir);
        return -ENOENT;
    }

#if 0
	kprintf("\nSearching real block = %d\n", dir->u.ext2_i.i_data[block]);
#endif

    bh = block_read(dir->i_sb->s_dev, dir->u.ext2_i.i_data[block], 1024);
	if(!bh)
		return -1;
	de = (struct ext2_dir_entry *) bh->b_data;
	//dump_memory((ulong *)bh->b_data, 64);

	tmp=0;
	offset=0;
	block=0;

#if 0
	kprintf("i_size = %d, ", dir->i_size);
	kprintf("name = %s, len = %d, ", name, len); 
#endif

	/* when direct addressing */
	while(offset < dir->i_size) {
		//kprintf("block = %d\n", block);
		/* if block < 12, run this */
		if(block < 12) {
            bh = block_read(dir->i_sb->s_dev, dir->u.ext2_i.i_data[block++], 1024);
            if(!bh) {
				kprintf("BREAD return NULL, ");
                return -1;
			}
			//dump_memory((ulong *)bh->b_data, 64);
			de = (struct ext2_dir_entry *) bh->b_data;
			dlimit = bh->b_data + sb->s_blocksize;
			while((char *)de < dlimit) {
				if(dir->i_ino != 2) {
					if (!ext2_check_dir_entry (dir, de, bh, offset)) {
						kprintf("ext2_check_dir_entry error\n");
                		return -1;
					}
				}
				if(de->inode !=0) {
					if(ext2_match(len, name, de)) {
#if 0
						kprintf("!!FOUND!!");
#endif
						ok_sign=1;
						if(ok_sign)
							goto ok_ok_sk;
					}
				}
                if(de->name[0] == '.' && de->name_len == 1) {
                    offset += 12;
                    de = (struct ext2_dir_entry *)
                         ((char *) de + 12);
                } else if (de->name[0] == '.' && de->name[1] == '.' && de->name_len == 2) {
                    offset += 12;
                    de = (struct ext2_dir_entry *)
                        ((char *) de + 12);
                } else {
                    offset += de->rec_len;
                    de = (struct ext2_dir_entry *)
                        ((char *) de + de->rec_len);
                }
			}
        	brelse (bh);
		}

		/* second-order array of logical block numbers */
		if((block) == 12) {
#if 0
			kprintf("in block 12\n");
#endif
			bh = block_read(dir->i_sb->s_dev, dir->u.ext2_i.i_data[12], 1024);
			if(!bh) {
				kprintf("BREAD return NULL, ");
				return -1;
			}
			addr = (unsigned long *) bh->b_data;
			for(i=0 ; i<268; i++) {
				bh2 = block_read(dir->i_sb->s_dev, *(addr+i), 1024);
				if(!bh2) {
					kprintf("shit!! bh2 is NULL\n");
					return -1;
				}
				de = (struct ext2_dir_entry *) bh2->b_data;
				dlimit = bh2->b_data + sb->s_blocksize;
        		while((char *)de < dlimit) {
            		if(dir->i_ino != 2) {
                		if (!ext2_check_dir_entry (dir, de, bh2, offset))
                    		return -1;
            		}
					if(de->inode !=0 && ext2_match(len, name, de)) {
#if 0
						kprintf("!!FOUND!!\n");
#endif
						ok_sign=1;
						brelse(bh2);
						goto ok_ok_sk;
					}
					if(de->name[0] == '.' && de->name_len == 1) {
						offset += 12;
                        de = (struct ext2_dir_entry *)
                            ((char *) de + 12);
					} else if (de->name[0] == '.' && de->name[1] == '.' && de->name_len == 2) {
						offset += 12;
                        de = (struct ext2_dir_entry *)
                            ((char *) de + 12);
					} else {
           	 			offset += de->rec_len;
        	    		de = (struct ext2_dir_entry *)
            	    		((char *) de + de->rec_len);
					}
				}
				brelse(bh2);
        	}
        	brelse (bh);
			block++;
		}

		/* third-order array of logical block numbers */
		if(block == 13) {
			kprintf("in block 13\n");
			kprintf("T_T plz make more third-oder array of logical\n");
        	bh = block_read(dir->i_sb->s_dev, dir->u.ext2_i.i_data[13], 1024);
        	if(!bh)
            	return -1;
        	dlimit = bh->b_data + sb->s_blocksize;
        	while((char *)de < dlimit) {
			}	
		}
	}	/* while(offset < dir->i_size) { */

ok_ok_sk: 				/* :-) mobile company SK */
	if(ok_sign) {
    	ino = de->inode;
#if 0
		kprintf("inode num=%d, ", ino);
#endif

    	if (!(*result = iget (dir->i_sb, ino))) {
        	iput (dir);
        	return -EACCES;
    	}
    	iput (dir);
		brelse(bh);
    	return 0;
	} else {
		panic("lookup fail: ok_sign == 0\n");
	}
	return -1;
}

/*-------------------------------------------------------------------*
 * ext2_create
 *-------------------------------------------------------------------*/
int ext2_create (struct inode *a,const char *aa, int aaaa, int aaaaa,
                struct inode **aaa)
{
    kprintf("in ext2_create\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_mkdir
 *-------------------------------------------------------------------*/
int ext2_mkdir (struct inode *b, const char *bb, int bbb, int bbbb)
{
    kprintf("in ext2_create\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_rmdir
 *-------------------------------------------------------------------*/
int ext2_rmdir (struct inode *c, const char *cc, int ccc)
{
    kprintf("in ext2_rmdir\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_unlink
 *-------------------------------------------------------------------*/
int ext2_unlink (struct inode *d, const char *dd, int ddd)
{
    kprintf("in ext2_unlink\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_symlink
 *-------------------------------------------------------------------*/
int ext2_symlink (struct inode *e, const char *ee, int eee, const char *eeee)
{
    kprintf("in ext2_symlink\n");
return 1;
}

/*-------------------------------------------------------------------*
 * ext2_link
 *-------------------------------------------------------------------*/
int ext2_link (struct inode *f, struct inode *ff, const char *fff, int ffff)
{
    kprintf("in ext2_link\n");
return 1;
}

static struct file_operations ext2_dir_operations = {
    NULL,           /* lseek - default */
    ext2_dir_read,      /* read */
    NULL,           /* write - bad */
    ext2_readdir,       /* readdir */
    NULL,           /* select - default */
    ext2_ioctl,     /* ioctl */
    NULL,           /* mmap */
    NULL,           /* no special open code */
    NULL,           /* no special release code */
    file_fsync      /* fsync */
};

struct inode_operations ext2_dir_inode_operations = {
    &ext2_dir_operations,   /* default directory file-ops */
    ext2_create,        /* create */
    ext2_lookup,        /* lookup */
    ext2_link,      /* link */
    ext2_unlink,        /* unlink */
    ext2_symlink,       /* symlink */
    ext2_mkdir,     /* mkdir */
    ext2_rmdir,     /* rmdir */
    ext2_mknod,     /* mknod */
    ext2_rename,        /* rename */
    NULL,           /* readlink */
    NULL,           /* follow_link */
    NULL,           /* bmap */
    ext2_truncate,      /* truncate */
    ext2_permission     /* permission */
};
