/*-------------------------------------------------------------------*
 * kernel/fs/ext2/file.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/fs.h>
#include <weix/ext2_fs.h>
#include <weix/types.h>
#include <weix/stat.h>
#include <weix/string.h>
#include <weix/dirent.h>
#include <asm/bitops.h>

#include <weix/kdebug.h>

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
static int ext2_file_read (struct inode * inode, struct file * filp,
            char * buf, int count);
static int ext2_file_write (struct inode * inode, struct file * filp,
                char * buf, int count);
int generic_mmap(struct inode * inode, struct file * file,
    unsigned long addr, size_t len, int prot, unsigned long off);

/*-------------------------------------------------------------------*
 * ext2_release_file
 *-------------------------------------------------------------------*/
static void ext2_release_file (struct inode * inode, struct file * filp)
{
    kprintf("in ext2_release_file\n");
}

/*-------------------------------------------------------------------*
 * generic_mmap
 *-------------------------------------------------------------------*/
int generic_mmap(struct inode * inode, struct file * file,
    unsigned long addr, size_t len, int prot, unsigned long off)
{
    kprintf("in generic_mmap\n");
    return 0;
}

/*-------------------------------------------------------------------*
 * ext2_sync_file
 *-------------------------------------------------------------------*/
int ext2_sync_file (struct inode * inode, struct file * file)
{
    kprintf("in ext2_sync_file\n");
    return 0;
}

/*-------------------------------------------------------------------*
 * ext2_bmap
 *-------------------------------------------------------------------*/
int ext2_bmap (struct inode *a, int b)
{
    kprintf("in ext2_bmap\n");
    return 0;
}

static struct file_operations ext2_file_operations = {
    NULL,           /* lseek - default */
    ext2_file_read,     /* read */
    ext2_file_write,    /* write */
    NULL,           /* readdir - bad */
    NULL,           /* select - default */
    ext2_ioctl,     /* ioctl */
    generic_mmap,       /* mmap */
    NULL,           /* no special open is needed */
    ext2_release_file,  /* release */
    ext2_sync_file      /* fsync */
};

struct inode_operations ext2_file_inode_operations = {
    &ext2_file_operations,/* default file operations */
    NULL,           /* create */
    NULL,           /* lookup */
    NULL,           /* link */
    NULL,           /* unlink */
    NULL,           /* symlink */
    NULL,           /* mkdir */
    NULL,           /* rmdir */
    NULL,           /* mknod */
    NULL,           /* rename */
    NULL,           /* readlink */
    NULL,           /* follow_link */
    ext2_bmap,      /* bmap */
    ext2_truncate,      /* truncate */
    ext2_permission     /* permission */
};

/*-------------------------------------------------------------------*
 * ext2_file_read
 *-------------------------------------------------------------------*/
static int ext2_file_read (struct inode * inode, struct file * filp,
            char * buf, int count)
{
    kprintf("in ext2_file_read\n");
    return 0;
}

/*-------------------------------------------------------------------*
 * ext2_file_write
 *-------------------------------------------------------------------*/
static int ext2_file_write (struct inode * inode, struct file * filp,
                char * buf, int count)
{
    kprintf("in ext2_file_write\n");
    return 0;
}
