/*-------------------------------------------------------------------*
 * kernel/fs/ext2/super.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/fs.h>
#include <weix/ext2_fs.h>
#include <weix/types.h>
#include <weix/stat.h>
#include <weix/string.h>
#include <weix/dirent.h>
#include <asm/bitops.h>
#include <weix/kdebug.h>
#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
static struct buffer_head * fst_group_desc_ptr;

/*-------------------------------------------------------------------*
 * ext2_read_inode
 *-------------------------------------------------------------------*/
void ext2_read_inode (struct inode *inode)
{
    struct buffer_head * bh;
    struct ext2_inode * raw_inode;
    unsigned long block_group;
    unsigned long group_desc;
    unsigned long desc;
    unsigned long block;
    struct ext2_group_desc * gdp;

#if 0
	kprintf("ext2_read_inode");
#endif

    if ((inode->i_ino != EXT2_ROOT_INO && inode->i_ino != EXT2_ACL_IDX_INO &&
         inode->i_ino != EXT2_ACL_DATA_INO && inode->i_ino < EXT2_FIRST_INO) ||
        inode->i_ino > inode->i_sb->u.ext2_sb.s_es->s_inodes_count) {
        panic("ext2_read_inode: bad inode number: %lu", inode->i_ino);
        return;
    }

    block_group = (inode->i_ino - 1) / EXT2_INODES_PER_GROUP(inode->i_sb);
    if (block_group >= inode->i_sb->u.ext2_sb.s_groups_count) {
        panic("ext2_read_inode: group >= groups count");
    }  

    group_desc = block_group / EXT2_DESC_PER_BLOCK(inode->i_sb);

    desc = block_group % EXT2_DESC_PER_BLOCK(inode->i_sb);
    bh = inode->i_sb->u.ext2_sb.s_group_desc[group_desc];
    if (!bh) {
        panic("ext2_read_inode: Descriptor not loaded");
    }
    gdp = (struct ext2_group_desc *) fst_group_desc_ptr->b_data;
    block = gdp[desc].bg_inode_table +
        (((inode->i_ino - 1) % EXT2_INODES_PER_GROUP(inode->i_sb))
         / EXT2_INODES_PER_BLOCK(inode->i_sb));

	kprintf("block=%d, ", block);
    if (!(bh = block_read(inode->i_dev, block, 1024)))
        panic("ext2_read_inode: unable to read i-node block\ninode=%lu, block=%lu", inode->i_ino, block);
	/* struct ext2_inode size is 128bytes in my computer */
    raw_inode = ((struct ext2_inode *) bh->b_data) +
        (inode->i_ino - 1) % EXT2_INODES_PER_BLOCK(inode->i_sb);
	//dump_memory((ulong *)bh->b_data, 64);
    inode->i_mode = raw_inode->i_mode;
    inode->i_uid = raw_inode->i_uid;
    inode->i_gid = raw_inode->i_gid;
    inode->i_nlink = raw_inode->i_links_count;
    inode->i_size = raw_inode->i_size;
    inode->i_atime = raw_inode->i_atime;
    inode->i_ctime = raw_inode->i_ctime;
    inode->i_mtime = raw_inode->i_mtime;
    inode->u.ext2_i.i_dtime = raw_inode->i_dtime;
    inode->i_blksize = inode->i_sb->s_blocksize;
    inode->i_blocks = raw_inode->i_blocks;
    inode->u.ext2_i.i_flags = raw_inode->i_flags;
    inode->u.ext2_i.i_faddr = raw_inode->i_faddr;
    inode->u.ext2_i.i_frag = raw_inode->i_frag;
    inode->u.ext2_i.i_fsize = raw_inode->i_fsize;
    inode->u.ext2_i.i_file_acl = raw_inode->i_file_acl;
    inode->u.ext2_i.i_dir_acl = raw_inode->i_dir_acl;
    inode->u.ext2_i.i_version = raw_inode->i_version;
    inode->u.ext2_i.i_block_group = block_group;
    inode->u.ext2_i.i_next_alloc_block = 0;
    inode->u.ext2_i.i_next_alloc_goal = 0;
    if (inode->u.ext2_i.i_prealloc_count)
        kprintf("ext2_read_inode: New inode has non-zero prealloc count!");
    if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode)) {
        inode->i_rdev = raw_inode->i_block[0];
	} else for (block = 0; block < EXT2_N_BLOCKS; block++)
        inode->u.ext2_i.i_data[block] = raw_inode->i_block[block];
    brelse (bh);
    inode->i_op = NULL;
    if (inode->i_ino == EXT2_ACL_IDX_INO ||
        inode->i_ino == EXT2_ACL_DATA_INO)
        /* Nothing to do */ ;
    else if (S_ISREG(inode->i_mode)) {
        kprintf("this inode is file\n");
        inode->i_op = &ext2_file_inode_operations;
    } else if (S_ISDIR(inode->i_mode)) {
        kprintf("this inode is directory!!!!!!!!\n");
        inode->i_op = &ext2_dir_inode_operations;
    } else if (S_ISLNK(inode->i_mode)) {
        kprintf("this inode is symbolic link\n");
        inode->i_op = &ext2_symlink_inode_operations;
    } else if (S_ISCHR(inode->i_mode)) {
        kprintf("this inode is character device\n");
        inode->i_op = &chrdev_inode_operations;
    } else if (S_ISBLK(inode->i_mode)) {
        kprintf("this inode is block device\n");
        inode->i_op = &blkdev_inode_operations;
    } else if (S_ISFIFO(inode->i_mode)) {
        kprintf("this inode is FIFO\n");
        //init_fifo(inode);
    } else {
		kprintf("SHITSHITSHITSHIT!!! what is this?\n");
	}
    if (inode->u.ext2_i.i_flags & EXT2_SYNC_FL)
        inode->i_flags |= MS_SYNC;
}

/*-------------------------------------------------------------------*
 * ext2_write_inode
 *-------------------------------------------------------------------*/
void ext2_write_inode (struct inode *ii)
{
    kprintf("in ext2_write_inode\n");
}

/*-------------------------------------------------------------------*
 * set_inode_dtime
 *-------------------------------------------------------------------*/
static void set_inode_dtime (struct inode * inode,
                 struct ext2_group_desc * gdp)
{
    unsigned long inode_block;
    struct buffer_head * bh;
    struct ext2_inode * raw_inode;
    inode_block = gdp->bg_inode_table + (((inode->i_ino - 1) %
            EXT2_INODES_PER_GROUP(inode->i_sb)) /
            EXT2_INODES_PER_BLOCK(inode->i_sb));
    bh = block_read(inode->i_sb->s_dev, inode_block, inode->i_sb->s_blocksize);
    if (!bh)
        panic ("set_inode_dtime: 
                Cannot load inode table block\n
                inode=%lu, inode_block=%lu",
                inode->i_ino, inode_block);
    raw_inode = ((struct ext2_inode *) bh->b_data) +
            (((inode->i_ino - 1) %
            EXT2_INODES_PER_GROUP(inode->i_sb)) %
            EXT2_INODES_PER_BLOCK(inode->i_sb));
    raw_inode->i_links_count = 0;
    raw_inode->i_dtime = CURRENT_TIME;
    bh->b_dirt = 1;
    if (IS_SYNC(inode)) {
        ll_rw_block (WRITE, 1, &bh);
        wait_on_buffer (bh);
    }
    brelse (bh);
}

/*-------------------------------------------------------------------*
 * ext2_free_inode
 *-------------------------------------------------------------------*/
void ext2_free_inode (struct inode * inode)
{
    struct super_block * sb;
    struct buffer_head * bh;
    struct buffer_head * bh2;
    unsigned long block_group;
    unsigned long bit;
    int bitmap_nr;
    struct ext2_group_desc * gdp;
    struct ext2_super_block * es;

    if (!inode)
        return;
    if (!inode->i_dev) {
        kprintf ("ext2_free_inode: inode has no device\n");
        return;
    }
    if (inode->i_count > 1) {
        kprintf ("ext2_free_inode: inode has count=%d\n",
            inode->i_count);
        return;
    }
    if (inode->i_nlink) {
        kprintf ("ext2_free_inode: inode has nlink=%d\n",
            inode->i_nlink);
        return;
    }
    if (!inode->i_sb) {
        kprintf("ext2_free_inode: inode on nonexistent device\n");
        return;
    }
#ifdef DEBUG
	kprintf("freeing inode %lu\n", inode->i_ino);
#endif

    sb = inode->i_sb;
    if (inode->i_ino < EXT2_FIRST_INO ||
        inode->i_ino > sb->u.ext2_sb.s_es->s_inodes_count) {
        kprintf("free_inode: 
                reserved inode or nonexistent inode");
        return;
    }
    es = sb->u.ext2_sb.s_es;
    block_group = (inode->i_ino - 1) / EXT2_INODES_PER_GROUP(sb);
    bit = (inode->i_ino - 1) % EXT2_INODES_PER_GROUP(sb);
    bitmap_nr = load_inode_bitmap (sb, block_group);
    bh = sb->u.ext2_sb.s_inode_bitmap[bitmap_nr];
    if (!clear_bit (bit, bh->b_data))
        kprintf("ext2_free_inode:
                  bit already cleared for inode %lu", inode->i_ino);
    else {
        gdp = get_group_desc (sb, block_group, &bh2);
        gdp->bg_free_inodes_count++;
        if (S_ISDIR(inode->i_mode))
            gdp->bg_used_dirs_count--;
        bh2->b_dirt = 1;
        es->s_free_inodes_count++;
        sb->u.ext2_sb.s_sbh->b_dirt = 1;
        set_inode_dtime (inode, gdp);
    }
    bh->b_dirt = 1;
    if (sb->s_flags & MS_SYNC) {
        ll_rw_block (WRITE, 1, &bh);
        wait_on_buffer (bh);
    }

    sb->s_dirt = 1;
    clear_inode (inode);
}

/*-------------------------------------------------------------------*
 * ext2_put_inode
 *-------------------------------------------------------------------*/
void ext2_put_inode (struct inode *inode)
{
    if (inode->i_nlink || inode->i_ino == EXT2_ACL_IDX_INO ||
        inode->i_ino == EXT2_ACL_DATA_INO)
        return;
    inode->i_size = 0;
    if (inode->i_blocks)
        ext2_truncate (inode);
    ext2_free_inode (inode);
}

/*-------------------------------------------------------------------*
 * ext2_put_super
 *-------------------------------------------------------------------*/
void ext2_put_super (struct super_block *b)
{
    kprintf("in ext2_put_super\n");
}

/*-------------------------------------------------------------------*
 * ext2_write_super
 *-------------------------------------------------------------------*/
void ext2_write_super (struct super_block *b)
{
    kprintf("in ext2_write_super\n");
}

/*-------------------------------------------------------------------*
 * ext2_statfs
 *-------------------------------------------------------------------*/
void ext2_statfs (struct super_block *b, struct statfs *f)
{
    kprintf("in ext2_statfs\n");
}

/*-------------------------------------------------------------------*
 * ext2_remount
 *-------------------------------------------------------------------*/
int ext2_remount (struct super_block *b, int *i, char *c)
{
    kprintf("in ext2_remount\n");
    return 1;
}

static struct super_operations ext2_sops = {
    ext2_read_inode,
    NULL,
    ext2_write_inode,
    ext2_put_inode,
    ext2_put_super,
    ext2_write_super,
    ext2_statfs,
    ext2_remount
};

/*-------------------------------------------------------------------*
 * ext2_setup_super
 *-------------------------------------------------------------------*/
static void ext2_setup_super (struct super_block * sb,
                  struct ext2_super_block * es)
{
    if (!(sb->s_flags & MS_RDONLY)) {
        if (!(sb->u.ext2_sb.s_mount_state & EXT2_VALID_FS))
            kprintf ("EXT2-fs warning: mounting unchecked fs, "
                "running e2fsck is recommended\n");
        else if ((sb->u.ext2_sb.s_mount_state & EXT2_ERROR_FS))
            kprintf ("EXT2-fs warning: mounting fs with errors, "
                "running e2fsck is recommended\n");
        else if (es->s_max_mnt_count >= 0 &&
                 es->s_mnt_count >= (unsigned short) es->s_max_mnt_count)
            kprintf ("EXT2-fs warning: maximal mount count reached, "
                "running e2fsck is recommended\n");
        else if (es->s_checkinterval &&
            (es->s_lastcheck + es->s_checkinterval <= CURRENT_TIME))
            kprintf ("EXT2-fs warning: checktime reached, "
                "running e2fsck is recommended\n");
        es->s_state &= ~EXT2_VALID_FS;
        if (!es->s_max_mnt_count)
            es->s_max_mnt_count = EXT2_DFL_MAX_MNT_COUNT;
        es->s_mnt_count++;
        es->s_mtime = CURRENT_TIME;
        sb->u.ext2_sb.s_sbh->b_dirt = 1;
        sb->s_dirt = 1;
        if (test_opt (sb, DEBUG))
            kprintf ("[EXT II FS %s, %s, bs=%lu, fs=%lu, gc=%lu, "
                "bpg=%lu, ipg=%lu, mo=%04lx]\n",
                EXT2FS_VERSION, EXT2FS_DATE, sb->s_blocksize,
                sb->u.ext2_sb.s_frag_size,
                sb->u.ext2_sb.s_groups_count,
                EXT2_BLOCKS_PER_GROUP(sb),
                EXT2_INODES_PER_GROUP(sb),
                sb->u.ext2_sb.s_mount_opt);
        if (test_opt (sb, CHECK)) {
            ext2_check_blocks_bitmap (sb);
            ext2_check_inodes_bitmap (sb);
        }
    }
}

/*-------------------------------------------------------------------*
 * ext2_check_descriptors
 *-------------------------------------------------------------------*/
static int ext2_check_descriptors (struct super_block * sb)
{
    int i;
    int desc_block = 0;
    unsigned long block = sb->u.ext2_sb.s_es->s_first_data_block;
    struct ext2_group_desc * gdp = NULL;

    for(i=0 ; i<sb->u.ext2_sb.s_groups_count; i++) {
        if ((i % EXT2_DESC_PER_BLOCK(sb)) == 0) {
            gdp = (struct ext2_group_desc *) sb->u.ext2_sb.s_group_desc[desc_block++]->b_data;
#ifdef DEBUG2
            kprintf("gdp=%08X\n", gdp);
#endif
        }
        if (gdp->bg_block_bitmap < block ||
            gdp->bg_block_bitmap >= block + EXT2_BLOCKS_PER_GROUP(sb))
        {
            kprintf("ext2_check_desciptors: block bitmap for group %d not in gou
rp (block %lu)!\n", i, gdp->bg_block_bitmap);
            return 0;
        }
        if (gdp->bg_inode_bitmap < block ||
            gdp->bg_inode_bitmap >= block + EXT2_BLOCKS_PER_GROUP(sb))
        {
            kprintf("ext2_check_desciptors: Inode bitmap for group %d ot in grou
p (block %lu)!\n", i, gdp->bg_inode_bitmap);
            return 0;
        }
        if (gdp->bg_inode_table < block ||
            gdp->bg_inode_table + sb->u.ext2_sb.s_itb_per_group >=
            block + EXT2_BLOCKS_PER_GROUP(sb))
        {
            kprintf("ext2_check_desciptors: Inode table for group %d not in grou
p (block %lu)!\n", i, gdp->bg_inode_table);
            return 0;
        }
        block += EXT2_BLOCKS_PER_GROUP(sb);
        gdp++;
    }
    return 1;
}

/*-------------------------------------------------------------------*
 * do_mount
 *-------------------------------------------------------------------*/
struct super_block * do_mount(int dev)
{
	struct super_block * sb;

	sb = ext2_read_super(NULL, NULL, dev);
	
	return sb;
}

/*-------------------------------------------------------------------*
 * ext2_read_super
 *-------------------------------------------------------------------*/
struct super_block * ext2_read_super (struct super_block * tmp_a, void * tmp_b, int dev)
{
    struct super_block * sb;
    struct buffer_head * bh;
    struct ext2_super_block *es;
    int i, j, bh_count;
    unsigned long sb_block = 1;
    unsigned long logic_sb_block = 1;

    for(sb = &super_blocks[0] ; sb < &super_blocks[NR_SUPER] ; sb++ )
        if (!(sb->s_dev))
            break;
    sb->s_dev = -1;     /* mark it in use */
    if (sb >= &super_blocks[NR_SUPER])
        return NULL;
    /* read super block */
#ifdef DEBUG2
    kprintf("do_mount: reading super block\n");
#endif
    if (!(bh = block_read(dev,sb_block, BLOCK_SIZE))) {
		kprintf("Oh~~~ NO~~~ bh = NULL~~~");
        return NULL;
	} 
    /*
     * Note: s_es must be initialized s_es as soon as possible because
     * some ext2 macro-instructions depend on its value
     */
    es = (struct ext2_super_block *) bh->b_data;
    sb->u.ext2_sb.s_es = es;
    sb->s_magic = es->s_magic;

    if (sb->s_magic != EXT2_SUPER_MAGIC) {
        sb->s_dev = 0;
        kprintf ("WEIX: Can't find an ext2 filesystem on dev %d/%d.\n",
                MAJOR(dev), MINOR(dev));
        brelse(bh);
        return NULL;
    } else {
		kprintf("SUPER_MAGIC!!");
	}

    sb->s_blocksize = EXT2_MIN_BLOCK_SIZE << es->s_log_block_size;
    sb->s_blocksize_bits = EXT2_BLOCK_SIZE_BITS(sb);
    sb->u.ext2_sb.s_frag_size = EXT2_MIN_FRAG_SIZE <<
                   es->s_log_frag_size;
    sb->u.ext2_sb.s_blocks_per_group = es->s_blocks_per_group;
    sb->u.ext2_sb.s_frags_per_group = es->s_frags_per_group;
    sb->u.ext2_sb.s_inodes_per_group = es->s_inodes_per_group;
    sb->u.ext2_sb.s_inodes_per_block = sb->s_blocksize /
                       sizeof (struct ext2_inode);

    sb->u.ext2_sb.s_itb_per_group = sb->u.ext2_sb.s_inodes_per_group /
                        sb->u.ext2_sb.s_inodes_per_block;
    sb->u.ext2_sb.s_desc_per_block = sb->s_blocksize /
                     sizeof (struct ext2_group_desc);
    sb->u.ext2_sb.s_sbh = bh;
    sb->u.ext2_sb.s_es = es;
    sb->u.ext2_sb.s_mount_state = es->s_state;
    sb->u.ext2_sb.s_rename_lock = 0;
    sb->u.ext2_sb.s_rename_wait = NULL;

    sb->u.ext2_sb.s_groups_count = (es->s_blocks_count -
                        es->s_first_data_block +
                       EXT2_BLOCKS_PER_GROUP(sb) - 1) /
                       EXT2_BLOCKS_PER_GROUP(sb);
    for (i = 0; i < EXT2_MAX_GROUP_DESC; i++)
        sb->u.ext2_sb.s_group_desc[i] = NULL;
    bh_count = (sb->u.ext2_sb.s_groups_count + EXT2_DESC_PER_BLOCK(sb) - 1) /
           EXT2_DESC_PER_BLOCK(sb);

    if (bh_count > EXT2_MAX_GROUP_DESC) {
        sb->s_dev = 0;
        brelse (bh);
        kprintf("EXT2-fs: file system is too big\n");
        return NULL;
    }

#if 0
	kprintf("sb->s_blocksize = %d, ", sb->s_blocksize);
#endif

    for (i = 0; i < bh_count; i++) {
#if 0
        kprintf("read: block_read(0x%X, %d)\n", dev, logic_sb_block+i+1+3);
#endif
        sb->u.ext2_sb.s_group_desc[i] = block_read(dev, logic_sb_block+i+1 , sb->s_blocksize);

#if 0
        kprintf("s_group_desc[%d] = %08X\n", i, *(sb->u.ext2_sb.s_group_desc[i])
);
        kprintf("s_group_desc[%d]->b_data = %08X\n", i, (sb->u.ext2_sb.s_group_desc[i])->b_data);
#endif
        if (!sb->u.ext2_sb.s_group_desc[i]) {
            sb->s_dev = 0;
            for (j = 0; j < i; j++)
                brelse (sb->u.ext2_sb.s_group_desc[j]);
            brelse (bh);
            kprintf ("EXT2-fs: unable to read group descriptors\n");
            return NULL;
        }
    }

	fst_group_desc_ptr = sb->u.ext2_sb.s_group_desc[0];

    if (!ext2_check_descriptors (sb)) {
        sb->s_dev = 0;
        for (j = 0; j < i; j++)
            brelse (sb->u.ext2_sb.s_group_desc[j]);
        brelse (bh);
        kprintf("EXT2-fs: group descriptors corrupted !\n");
        return NULL;
    }
    for (i = 0; i < EXT2_MAX_GROUP_LOADED; i++) {
        sb->u.ext2_sb.s_inode_bitmap_number[i] = 0;
        sb->u.ext2_sb.s_inode_bitmap[i] = NULL;
        sb->u.ext2_sb.s_block_bitmap_number[i] = 0;
        sb->u.ext2_sb.s_block_bitmap[i] = NULL;
    }
    sb->u.ext2_sb.s_loaded_inode_bitmaps = 0;
    sb->u.ext2_sb.s_loaded_block_bitmaps = 0;

    sb->s_dev = dev;
    sb->s_op = &ext2_sops;

    if (!(sb->s_mounted = iget (sb, EXT2_ROOT_INO))) {
        sb->s_dev = 0;
        for (i = 0; i < EXT2_MAX_GROUP_DESC; i++)
            if (sb->u.ext2_sb.s_group_desc[i])
                brelse (sb->u.ext2_sb.s_group_desc[i]);
        brelse (bh);
        kprintf ("EXT2-fs: get root inode failed\n");
        return NULL;
    }

    ext2_setup_super (sb, es);
    return sb;
}

