/*-------------------------------------------------------------------*
 * kernel/fs/ext2/symlink.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/sched.h>
#include <weix/kernel.h>
#include <weix/fs.h>
#include <weix/ext2_fs.h>
#include <weix/types.h>
#include <weix/stat.h>
#include <weix/string.h>
#include <weix/dirent.h>
#include <asm/bitops.h>
#include <weix/kdebug.h>

#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
static struct inode_hash_entry {
    struct inode * inode;
    int updating;
} hash_table[NR_IHASH];

static struct inode * first_inode;
static int nr_free_inodes = 0;

/*-------------------------------------------------------------------*
 * hashfn
 *-------------------------------------------------------------------*/
static inline int const hashfn(dev_t dev, unsigned int i)
{
    return (dev ^ i) % NR_IHASH;
}

/*-------------------------------------------------------------------*
 * hash
 *-------------------------------------------------------------------*/
static inline struct inode_hash_entry * const hash(dev_t dev, int i)
{
    return hash_table + hashfn(dev, i);
}

/*-------------------------------------------------------------------*
 * remove_inode_free
 *-------------------------------------------------------------------*/
static void remove_inode_free(struct inode *inode)
{
    if (first_inode == inode)
        first_inode = first_inode->i_next;
    if (inode->i_next)
        inode->i_next->i_prev = inode->i_prev;
    if (inode->i_prev)
        inode->i_prev->i_next = inode->i_next;
    inode->i_next = inode->i_prev = NULL;
}

/*-------------------------------------------------------------------*
 * remove_inode_hash
 *-------------------------------------------------------------------*/
static void remove_inode_hash(struct inode *inode)
{
    struct inode_hash_entry *h;
    h = hash(inode->i_dev, inode->i_ino);

    if (h->inode == inode)
        h->inode = inode->i_hash_next;
    if (inode->i_hash_next)
        inode->i_hash_next->i_hash_prev = inode->i_hash_prev;
    if (inode->i_hash_prev)
        inode->i_hash_prev->i_hash_next = inode->i_hash_next;
    inode->i_hash_prev = inode->i_hash_next = NULL;
}

/*-------------------------------------------------------------------*
 * insert_inode_free
 *-------------------------------------------------------------------*/
static void insert_inode_free(struct inode *inode)
{
    inode->i_next = first_inode;
    inode->i_prev = first_inode->i_prev;
    inode->i_next->i_prev = inode;
    inode->i_prev->i_next = inode;
    first_inode = inode;
}

/*-------------------------------------------------------------------*
 * clear_inode
 *-------------------------------------------------------------------*/
void clear_inode(struct inode * inode)
{
    struct my_life_schedule * wait;

    wait_on_inode(inode);
    remove_inode_hash(inode);
    remove_inode_free(inode);
    wait = ((volatile struct inode *) inode)->i_wait;
    if (inode->i_count)
        nr_free_inodes++;
    memset(inode,0,sizeof(*inode));
    ((volatile struct inode *) inode)->i_wait = wait;
    insert_inode_free(inode);
}

/*-------------------------------------------------------------------*
 * ext2_follow_link
 *-------------------------------------------------------------------*/
static int ext2_follow_link(struct inode * dir, struct inode * inode,
                int flag, int mode, struct inode ** res_inode)
{
    kprintf("in ext2_follow_link\n");
    return 0;
}

/*-------------------------------------------------------------------*
 * ext2_readlink
 *-------------------------------------------------------------------*/
static int ext2_readlink (struct inode * inode, char * buffer, int buflen)
{
    kprintf("in ext2_readlink\n");
    return 0;
}

struct inode_operations ext2_symlink_inode_operations = {
    NULL,           /* no file-operations */
    NULL,           /* create */
    NULL,           /* lookup */
    NULL,           /* link */
    NULL,           /* unlink */
    NULL,           /* symlink */
    NULL,           /* mkdir */
    NULL,           /* rmdir */
    NULL,           /* mknod */
    NULL,           /* rename */
    ext2_readlink,      /* readlink */
    ext2_follow_link,   /* follow_link */
    NULL,           /* bmap */
    NULL,           /* truncate */
    NULL            /* permission */
};
