/*-------------------------------------------------------------------*
 * kernel/fs/fcntl.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/string.h>
#include <weix/errno.h>
#include <weix/sched.h>
#include <weix/kernel.h>
#include <asm/fs_func.h>

#include <weix/fcntl.h>
#include <weix/stat.h>
#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * dupfd
 *-------------------------------------------------------------------*/
static int dupfd(unsigned int fd, unsigned int arg)
{
	if (fd >= NR_OPEN || !current->filp[fd])
		return -EBADF;
	if (arg >= NR_OPEN)
		return -EINVAL;
	while (arg < NR_OPEN)
		if (current->filp[arg])
			arg++;
		else
			break;
	if (arg >= NR_OPEN)
		return -EMFILE;
	current->close_on_exec &= ~(1<<arg);
	(current->filp[arg] = current->filp[fd])->f_count++;
	return arg;
}

/*-------------------------------------------------------------------*
 * sys_dup2
 *-------------------------------------------------------------------*/
int sys_dup2(unsigned int oldfd, unsigned int newfd)
{
	sys_close(newfd);
	return dupfd(oldfd,newfd);
}

/*-------------------------------------------------------------------*
 * sys_dup
 *-------------------------------------------------------------------*/
int sys_dup(unsigned int fildes)
{
	return dupfd(fildes,0);
}

/*-------------------------------------------------------------------*
 * sys_fcntl
 *-------------------------------------------------------------------*/
int sys_fcntl(unsigned int fd, unsigned int cmd, unsigned long arg)
{	
	kprintf("in sys_fcntl\n");
	return -1;
}
