/*-------------------------------------------------------------------*
 * kernel/fs/open.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/system.h>
#include <weix/string.h>
#include <weix/errno.h>
#include <weix/fcntl.h>
#include <weix/types.h>
#include <weix/utime.h>
#include <weix/stat.h>
#include <weix/sched.h>
#include <weix/tty.h>
#include <weix/kernel.h>
#include <asm/fs_func.h>

#include <weix/extern_func_T_T.h>

/*-------------------------------------------------------------------*
 * defines
 *-------------------------------------------------------------------*/
#define MAY_EXEC 1
#define MAY_WRITE 2
#define NOTIFY_SIZE     1
#define NGROUPS         32
#define NOGROUP         (-1)
#define ACC_MODE(x) ("\000\004\002\006"[(x)&O_ACCMODE])
#define suser() (current->euid == 0)

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
static int nr_free_inodes = 0;
/* current file struct in weix */
struct file file_table[NR_FILE];

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
void iput(struct inode * inode);
int lookup(struct inode * dir,const char * name, int len,
    struct inode ** result);
int follow_link(struct inode * dir, struct inode * inode,
    int flag, int mode, struct inode ** res_inode);

/*-------------------------------------------------------------------*
 * sys_pipe
 *-------------------------------------------------------------------*/
int sys_pipe(unsigned long * fildes)
{
    return -1;
}

/*-------------------------------------------------------------------*
 * sys_utime
 *-------------------------------------------------------------------*/
int sys_utime(char * filename, struct utimbuf * times)
{
	return -1;
}

/*-------------------------------------------------------------------*
 * sys_access
 *-------------------------------------------------------------------*/
int sys_access(const char * filename,int mode)
{
	return -1;
}

/*-------------------------------------------------------------------*
 * sys_chdir
 *-------------------------------------------------------------------*/
int sys_chdir(const char * filename)
{
	return -1;
}

/*-------------------------------------------------------------------*
 * sys_chroot
 *-------------------------------------------------------------------*/
int sys_chroot(const char * filename)
{
	return -1;
}

/*-------------------------------------------------------------------*
 * sys_chmod
 *-------------------------------------------------------------------*/
int sys_chmod(const char * filename,int mode)
{
	return -1;
}

/*-------------------------------------------------------------------*
 * sys_chown
 *-------------------------------------------------------------------*/
int sys_chown(const char * filename,int uid,int gid)
{
	return -1;
}

/*-------------------------------------------------------------------*
 * notify_change
 *-------------------------------------------------------------------*/
int notify_change(int flags, struct inode * inode)
{
#if 0
	kprintf("in notify_change\n");
#endif
    if (inode->i_sb && inode->i_sb->s_op  &&
        inode->i_sb->s_op->notify_change)
        return inode->i_sb->s_op->notify_change(flags, inode);
    return 0;
}

/*-------------------------------------------------------------------*
 * dir_namei
 *-------------------------------------------------------------------*/
static int dir_namei(const char * pathname, int * namelen, const char ** name,
    struct inode * base, struct inode ** res_inode)
{
    char c;
    const char * thisname;
    int len,error;
    struct inode * inode;

#if 0
	kprintf("in dir_namei ");
#endif
    *res_inode = NULL;
    if (!base) {
        base = current->pwd;
        base->i_count++;
    }
    if ((c = *pathname) == '/') {
        iput(base);
        base = current->root;
        pathname++;
        base->i_count++;
    }

    while (1) {
        thisname = pathname;
        for(len=0;(c = *(pathname++))&&(c != '/');len++)
            /* nothing */ ;
        if (!c)
            break;
        base->i_count++;
        error = lookup(base,thisname,len,&inode);
        if (error) {
			kprintf("lookup error, ");
            iput(base);
            return error;
        }
        error = follow_link(base,inode,0,0,&base);
        if (error)
            return error;
    }

#if 0
	kprintf("base->i_op = %08X, base->i_op->lookup=%08X, ", base->i_op, base->i_op->lookup);
#endif

	if(S_ISDIR(base->i_mode)) {
    	if (!base->i_op || !base->i_op->lookup) {
			kprintf("!!_ _!!HUK!!!\n");
        	iput(base);
        	return -ENOTDIR;
    	}
	}
    *name = thisname;
    *namelen = len;
    *res_inode = base;
#if 0
	kprintf("*name=%s, *namelen=%d, ", *name, *namelen);
#endif
    return 0;
}

/*-------------------------------------------------------------------*
 * in_group_p
 *-------------------------------------------------------------------*/
int in_group_p(gid_t grp)
{
    int i;

    if (grp == current->egid)
        return 1;

    for (i = 0; i < NGROUPS; i++) {
		kprintf("Oh~~~~ fix me plz i'am in in_group_p function ");
		/*
        if (current->groups[i] == NOGROUP)
            break;
        if (current->groups[i] == grp)
            return 1;
		*/
    }
    return 0;
}

/*-------------------------------------------------------------------*
 * permission
 *-------------------------------------------------------------------*/
int permission(struct inode * inode,int mask)
{
    int mode = inode->i_mode;

#if 0
	kprintf("in permission\n");
#endif
    if (inode->i_op && inode->i_op->permission)
        return inode->i_op->permission(inode, mask);
    else if (current->euid == inode->i_uid)
        mode >>= 6;
    else if (in_group_p(inode->i_gid))
        mode >>= 3;
    if (((mode & mask & 0007) == mask) || suser())
        return 1;
    return 0;
}

/*-------------------------------------------------------------------*
 * up
 *-------------------------------------------------------------------*/
inline void up(struct semaphore * sem)
{
	kprintf("in weix, semaphore not made :-( let's wait, ");
}

/*-------------------------------------------------------------------*
 * down
 *-------------------------------------------------------------------*/
inline void down(struct semaphore * sem)
{
	kprintf("in weix, semaphore not made :-( let's wait, ");	
}

/*-------------------------------------------------------------------*
 * lookup
 *-------------------------------------------------------------------*/
int lookup(struct inode * dir,const char * name, int len,
    struct inode ** result)
{
    struct super_block * sb;
    int perm;

#if 0
	kprintf("in lookup, ");
#endif
    *result = NULL;
    if (!dir)
        return -ENOENT;
/* check permissions before traversing mount-points */
    perm = permission(dir,MAY_EXEC);
    if (len==2 && name[0] == '.' && name[1] == '.') {
        if (dir == current->root) {
            *result = dir;
            return 0;
        } else if ((sb = dir->i_sb) && (dir == sb->s_mounted)) {
            sb = dir->i_sb;
            iput(dir);
            dir = sb->s_covered;
            if (!dir)
                return -ENOENT;
            dir->i_count++;
        }
    }

	/* check inode operation and lookup operation be set */
	if(S_ISDIR(dir->i_mode)) {
	    if (!dir->i_op || !dir->i_op->lookup) {
    	    iput(dir);
        	return -ENOTDIR;
    	}
	}

	/* check permission */
    if (!perm) {
        iput(dir);
        return -EACCES;
    }

	/*
	 * Ȯҷ ̳ ̸ ȮѴ.
	 * ( ̽ð{01/11/21}) Ȯҷ ϴ  
	 * "dev/tty0" ̱  len 3   ִ.
	 */ 
    if (!len) {
        *result = dir;
        return 0;
    }
    return dir->i_op->lookup(dir,name,len,result);
}

/*-------------------------------------------------------------------*
 * follow_link
 *-------------------------------------------------------------------*/
int follow_link(struct inode * dir, struct inode * inode,
    int flag, int mode, struct inode ** res_inode)
{
#if 0
	kprintf("in follow_link, ");
#endif
    if (!dir || !inode) {
        iput(dir);
        iput(inode);
        *res_inode = NULL;
        return -ENOENT;
    }
    if (!inode->i_op || !inode->i_op->follow_link) {
        iput(dir);
        *res_inode = inode;
        return 0;
    }
    return inode->i_op->follow_link(dir,inode,flag,mode,res_inode);
}

/*-------------------------------------------------------------------*
 * open_namei
 *-------------------------------------------------------------------*/
int open_namei(const char * pathname, int flag, int mode,
    struct inode ** res_inode, struct inode * base)
{
    const char * basename;
    int namelen,error;
    struct inode * dir, *inode;
    struct my_life_schedule ** p;

#if 0
	kprintf("in open_namei, ");
#endif
    mode &= S_IALLUGO & ~current->umask;
    mode |= S_IFREG;

    error = dir_namei(pathname,&namelen,&basename,base,&dir);
    if (error) {
		kprintf("oh~ error, ");
        return error;
	}

    if (!namelen) {         /* special case: '/usr/' etc */
        if (flag & 2) {
            iput(dir);
            return -EISDIR;
        }
        /* thanks to Paul Pluzhnikov for noticing this was missing.. */
        if (!permission(dir,ACC_MODE(flag))) {
            iput(dir);
            return -EACCES;
        }
        *res_inode=dir;
        return 0;
    }
    dir->i_count++;     /* lookup eats the dir */
    if (flag & O_CREAT) {
        down(&dir->i_sem);
        error = lookup(dir,basename,namelen,&inode);
        if (!error) {
            if (flag & O_EXCL) {
                iput(inode);
                error = -EEXIST;
            }
        } else if (!permission(dir,MAY_WRITE | MAY_EXEC))
            error = -EACCES;
        else if (!dir->i_op || !dir->i_op->create)
            error = -EACCES;
        else if (IS_RDONLY(dir))
            error = -EROFS;
        else {
            dir->i_count++;     /* create eats the dir */
            error = dir->i_op->create(dir,basename,namelen,mode,res_inode);
            up(&dir->i_sem);
            iput(dir);
            return error;
        }
        up(&dir->i_sem);
    } else {
    	kprintf("increse i_count -_- by aso\n");
    	dir->i_count += 2;
        error = lookup(dir,basename,namelen,&inode);
	}
    if (error) {
        iput(dir);
        return error;
    }
	kprintf("increse i_count -_- by aso\n");
	dir->i_count += 5;
    error = follow_link(dir,inode,flag,mode,&inode);
    if (error) {
		kprintf("-follow_link error-");
        return error;
	}
    if (S_ISDIR(inode->i_mode) && (flag & 2)) {
		kprintf("-strange error-");
        iput(inode);
        return -EISDIR;
    }
    if (!permission(inode,ACC_MODE(flag))) {
		kprintf("-permission error-");
        iput(inode);
        return -EACCES;
    }
    if (S_ISBLK(inode->i_mode) || S_ISCHR(inode->i_mode)) {
        if (IS_NODEV(inode)) {
            iput(inode);
            return -EACCES;
        }
    }

    if ((inode->i_count > 1) && (flag & 2)) {
        for (p = &LAST_TASK ; p > &FIRST_TASK ; --p) {
			/* fix me plz */
                //struct vm_area_struct * mpnt;
            if (!*p)
                continue;
			kprintf("Oh~~~ sorry. this here not made function all in weix, wait for later version");
			/*
            if (inode == (*p)->executable) {
                iput(inode);
                return -ETXTBSY;
            }
			*/
			/*
            for(mpnt = (*p)->mmap; mpnt; mpnt = mpnt->vm_next) {
                if (mpnt->vm_page_prot & PAGE_RW)
                    continue;
                if (inode == mpnt->vm_inode) {
                    iput(inode);
                    return -ETXTBSY;
                }
            }
			*/
        }
    }
    if (flag & O_TRUNC) {
          inode->i_size = 0;
          if (inode->i_op && inode->i_op->truncate)
               inode->i_op->truncate(inode);
          if ((error = notify_change(NOTIFY_SIZE, inode))) {
           iput(inode);
           return error;
          }
          inode->i_dirt = 1;
    }
#if 0
	kprintf("end of open_namei, ");
#endif
    *res_inode = inode;

    return 0;
}

/*-------------------------------------------------------------------*
 * __wait_on_inode
 *-------------------------------------------------------------------*/
void __wait_on_inode(struct inode *inode)
{
    cli();
    while (inode->i_lock)
        sleep_on(&inode->i_wait);
    sti();
}

/*-------------------------------------------------------------------*
 * wait_on_inode
 *-------------------------------------------------------------------*/
void wait_on_inode(struct inode * inode)
{
    __wait_on_inode(inode);
}

/*-------------------------------------------------------------------*
 * unlock_inode
 *-------------------------------------------------------------------*/
static inline void unlock_inode(struct inode * inode)
{
#if 0
	kprintf("in unlock_inode, ");
#endif
    inode->i_lock = 0;
	kprintf("Oh~~~~~~ wake_up func omit in this function (unlock_inode), fix me plz, ");
    //wake_up(&inode->i_wait);
}

/*-------------------------------------------------------------------*
 * write_inode
 *-------------------------------------------------------------------*/
static void write_inode(struct inode * inode)
{
#if 0
	kprintf("in write_inode, ");
#endif
    if (!inode->i_dirt)
        return;
    wait_on_inode(inode);
    if (!inode->i_dirt)
        return;
    if (!inode->i_sb || !inode->i_sb->s_op || !inode->i_sb->s_op->write_inode) {
        inode->i_dirt = 0;
        return;
    }
    inode->i_lock = 1;
    inode->i_sb->s_op->write_inode(inode);
    unlock_inode(inode);
}

/*-------------------------------------------------------------------*
 * iput
 *-------------------------------------------------------------------*/
void iput(struct inode * inode)
{
#if 0
	kprintf("in iput, i_count=%d, ", inode->i_count);
#endif
    if (!inode) {
        return;
	}
    wait_on_inode(inode);
    if (!inode->i_count) {
        kprintf("WEIX: iput: trying to free free inode\n");
        kprintf("WEIX: device %d/%d, inode %lu, mode=0%07o\n",
            MAJOR(inode->i_rdev), MINOR(inode->i_rdev),
                    inode->i_ino, inode->i_mode);
        return;
    }
repeat:
    if (inode->i_count>1) {
        inode->i_count--;
        return;
    }

    if (inode->i_sb && inode->i_sb->s_op && inode->i_sb->s_op->put_inode) {
        inode->i_sb->s_op->put_inode(inode);
        if (!inode->i_nlink)
            return;
    } else {
		kprintf("Hmm.. can't put_inode(), ");
	}
    if (inode->i_dirt) {
        write_inode(inode); /* we can sleep - so do again */
        wait_on_inode(inode);
        goto repeat;
    }
    inode->i_count--;
    nr_free_inodes++;
#if 0
	kprintf("end of iput, ");
#endif
    return;
}

/*-------------------------------------------------------------------*
 * sys_open
 *-------------------------------------------------------------------*/
int sys_open(const char * filename,int flags,int mode)
{
    struct inode * inode;
    struct file * f;
    int i, flag,error,fd;

#if 0
	kprintf("in sys_open :0-), ");
#endif

    for(fd=0 ; fd<NR_OPEN ; fd++)
        if (!current->filp[fd])
            break;
    if (fd>=NR_OPEN)
        return -EMFILE;

	FD_CLR(fd,&current->close_on_exec);

	f=0+file_table;
	for (i=0 ; i<NR_FILE ; i++,f++)
        if (!f->f_count) break;
    if (i>=NR_FILE)
        return -EINVAL;

	current->filp[fd] = f;
    f->f_flags = flag = flags;
    f->f_mode = (flag+1) & O_ACCMODE;
    if (f->f_mode)
        flag++;
    if (flag & (O_TRUNC | O_CREAT))
        flag |= 2;

	error = open_namei(filename,flag,mode,&inode,NULL);
    if (error) {
		kprintf("Oh~ error, ");
        current->filp[fd]=NULL;
        f->f_count--;
        return error;
    }
    if (S_ISCHR(inode->i_mode)) {
        if (MAJOR(inode->i_rdev)==4) {
            if (current->leader && current->tty<0) {
                current->tty = MINOR(inode->i_rdev);
                tty_table[current->tty].pgrp = current->pgrp;
            }
        } else if (MAJOR(inode->i_rdev)==5) {
            if (current->tty<0) {
                iput(inode);
                current->filp[fd]=NULL;
                f->f_count=0;
                return -EPERM;
            }
        }
    }
    f->f_mode = inode->i_mode;
    f->f_flags = flag;
    f->f_count = 1;
    f->f_inode = inode;
    f->f_pos = 0;
    f->f_reada = 0;
    f->f_op = NULL;
    if (inode->i_op)
        f->f_op = inode->i_op->default_file_ops;
    if (f->f_op && f->f_op->open) {
        error = f->f_op->open(inode,f);
        if (error) {
            iput(inode);
            f->f_count--;
            current->filp[fd]=NULL;
            return error;
        }
    }
    f->f_flags &= ~(O_CREAT | O_EXCL | O_NOCTTY | O_TRUNC);
    return (fd);
}

/*-------------------------------------------------------------------*
 * sys_create
 *-------------------------------------------------------------------*/
int sys_creat(const char * pathname, int mode)
{
	kprintf("called sys_creat\n");
	return sys_open(pathname, O_CREAT | O_TRUNC, mode);
}

/*-------------------------------------------------------------------*
 * sys_close
 *-------------------------------------------------------------------*/
int sys_close(unsigned int fd)
{	
    struct file * filp;

    if (fd >= NR_OPEN)
        return -EINVAL;
    current->close_on_exec &= ~(1<<fd);
    if (!(filp = current->filp[fd]))
        return -EINVAL;
    current->filp[fd] = NULL;
    if (filp->f_count == 0)
        panic("Close: file count is 0");
    if (--filp->f_count)
        return (0);
    iput(filp->f_inode);
    return (0);
}
