/*-------------------------------------------------------------------*
 * /kernel/gdt.c
 *
 * Description:
 * Global Descriptor Table functions
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/system.h>
#include <asm/segment.h>
#include <weix/types.h>
#include <weix/params.h>
#include <weix/gdt.h>
#include <weix/string.h>

/*-------------------------------------------------------------------*
 * gloval variables
 *-------------------------------------------------------------------*/
descriptor *GDT = (descriptor *)0x70000;	// 448k .

/*-------------------------------------------------------------------*
 * setup_GDT
 *-------------------------------------------------------------------*/
void setup_GDT(void)
{
	dword GDTR[2];

	cli();
	memset((void *)GDT, 0, GDT_SIZE * sizeof(descriptor));
	setup_descriptor(&GDT[1], 0, LIMIT_4GB, 8, 0, 1, 1); /* 0x08 -- KERNEL CODE */
	setup_descriptor(&GDT[2], 0, LIMIT_4GB, 2, 0, 1, 1); /* 0x10 -- KERNEL DATA */
	setup_descriptor(&GDT[3], 0, LIMIT_4GB, 8, 3, 1, 1); /* 0x1B -- USER CODE */
	setup_descriptor(&GDT[4], 0, LIMIT_4GB, 2, 3, 1, 1); /* 0x23 -- USER DATA */

	/* the following computation of the GDTR
	 * could/should be done much more readable by
	 * describing the memory layout of the GDTR
	 * in a packed struct in gdt.h */
	GDTR[0] = ((GDT_SIZE * sizeof(descriptor)) - 1) << 16;
	GDTR[1] = (dword)GDT;

	/* load the GDTR - it's only 48-bits and therefore the
	 * two first bytes are not used */
	__asm__ __volatile__("lgdt (%0)" : : "r" (((byte *)GDTR) + 2));

	/* flush the instruction queue and set CS */
	__asm__ __volatile__("ljmp $0x8, $__flush\n"
				         "__flush:");

	/* set DS, SS, ES, FS, GS */
	__asm__ __volatile__("
		movl %0, %%eax\n\t
		movw %%ax, %%ss\n\t
		movl %1, %%eax\n\t
		movw %%ax, %%ds\n\t
		movw %%ax, %%es\n\t
		movw %%ax, %%fs\n\t
		movw %%ax, %%gs"
		:
		: "i" (KERNEL_DS), "i" (USER_DS));

	sti();
}

/*-------------------------------------------------------------------*
 * setup_descriptor
 *-------------------------------------------------------------------*/
void setup_descriptor(descriptor *d, dword base, dword limit, byte type, byte dpl, byte s, byte db)
{
  /* calculate limit and granularity */
  if (limit > LIMIT_1MB)
    {
      limit = limit >> 12;
      d->g = 1;
    }
  else
    d->g = 0;

  /* set base address */
  d->base0_15 = base;
  d->base16_23 = base >> 16;
  d->base24_31 = base >> 24;

  /* set segment limit */
  d->limit0_15 = limit;
  d->limit16_19 = limit >> 16;

  /* set rest of descriptor */
  d->type = type;
  d->s = s;
  d->dpl = dpl;
  d->p = 1;
  d->avl = 0;
  d->zero = 0;
  d->db = db;
}
