/*-------------------------------------------------------------------*
 * idt.c
 *
 * Description:
 * Interrupt Descriptor Table functions
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/segment.h>
#include <asm/system.h>
#include <weix/types.h>
#include <weix/params.h> /* for IDT_SIZE */
#include <asm/io.h>
#include <weix/idt.h>
#include <weix/gdt.h> /* for struct descriptior */
#include <weix/string.h>

/*-------------------------------------------------------------------*
 * gloval viriables
 *-------------------------------------------------------------------*/
/* 0x70000  GDT ּҰ 0x70000  ϱ Դϴ. */
gate *IDT = (gate *)(0x70000 + (GDT_SIZE * sizeof(descriptor)));

/*-------------------------------------------------------------------*
 * setup_IDT
 *-------------------------------------------------------------------*/
void setup_IDT(void)
{
  dword IDTR[2];

	cli();
  memset(IDT, 0, IDT_SIZE * sizeof(gate));
  IDTR[0] = ((IDT_SIZE * sizeof(gate)) - 1) << 16;
  IDTR[1] = (dword)IDT;
  /* load the IDTR - it's only 48-bits and therefore the
   * two first bytes are not used */
  __asm__ __volatile__("lidt (%0)" : : "r" (((byte *)IDTR) + 2));
	sti();
}

/*-------------------------------------------------------------------*
 * interrupt_IDT
 *-------------------------------------------------------------------*/
void interrupt_IDT(gate *g, word selector, dword offset, byte dpl)
{
  g->offset0_15 = offset;
  g->offset16_31 = offset >> 16;
  g->selector = selector;
  g->zero = 0;
  g->p = 1;
  g->dpl = dpl;
  g->type = 0xE; /* assuming 32-bit interrupt gate */
}

/*-------------------------------------------------------------------*
 * trap_IDT
 *-------------------------------------------------------------------*/
void trap_IDT(gate *g, word selector, dword offset, byte dpl)
{
  g->offset0_15 = offset;
  g->offset16_31 = offset >> 16;
  g->selector = selector;
  g->zero = 0;
  g->p = 1;
  g->dpl = dpl;
  g->type = 0xF; /* assuming 32-bit trap gate */
}

/*-------------------------------------------------------------------*
 * task_IDT
 *-------------------------------------------------------------------*/
void task_IDT(gate *g, word selector, byte dpl)
{
  g->offset0_15 = 0;
  g->offset16_31 = 0;
  g->selector = selector;
  g->zero = 0;
  g->p = 1;
  g->dpl = dpl;
  g->type = 5;
}
