/*-------------------------------------------------------------------**
 * /kernel/kdebug.c
 *
 * Description:
 * Contains various printing functions for use with
 * kernel debugging.  These routines are not
 * optimized since we're only debugging :-)
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/types.h>
#include <weix/types.h>
#include <weix/stdarg.h>
#include <weix/string.h>
#include <weix/idt.h>
#include <weix/gdt.h>
#include <weix/kdebug.h>
#include <weix/sched.h>
#include <weix/params.h>
#include <weix/tty.h>

/*-------------------------------------------------------------------*
 * defines
 *-------------------------------------------------------------------*/
/* personal definitions */
#define SHOW_MEM_SIZE	4
/* parameters about video card */
#define ROWS        25
#define COLS        80
#define VID_SIZE    (ROWS*COLS)
#define ROWSIZE     0xA0

#define VID_PTR     0xB8000

#define BLNK_MSK    0x7F
#define BG_MSK      0x8F
#define FG_MSK      0xF8
#define INTSTY_MSK  0xF7

#define DIG_INI     0x30 /* ASCII '0' */
#define TAB_SIZE    8

/*-------------------------------------------------------------------*
 * grobal variables
 *-------------------------------------------------------------------*/
// Attribute:
//            MSB           LSB
//             0 0 0 0 0 0 0 0
//             + +-+-+ + +-+-+
//             |   |   |   |------> Foreground
//             |   |   |----------> Intensity
//             |   |--------------> Background
//             |------------------> Blink
byte	char_attribute;
struct 	current_tty_buffer tty_buffer[4];
int 	tty_buffer_ptr=0;
struct 	cursor {
  byteptr p;
  byte    x,y;
};
struct cursor char_cursor;

/*-------------------------------------------------------------------*
 * console_scroll
 *
 * numline ŭ scroll մϴ.
 *-------------------------------------------------------------------*/
static void console_scroll(int numlines)
{
  int i,j;
  char *dest, *src;

  for( i=0; i<numlines; i++ ) {
     for( j=1; j<ROWS; j++ ) {
	 // Setup pointers
	    src = (char *)(VID_PTR + j*ROWSIZE);
	    memcpy( src - ROWSIZE, src, sizeof(byte)*ROWSIZE );
	 }
     // Clear last line
     dest = (char *)(VID_PTR + (ROWS-1)*ROWSIZE);
     for( j=0; j<COLS; j++ ) {
	    *dest++ = ' '; *dest++ = char_attribute;
     }
  }
}


/*-------------------------------------------------------------------*
 * console_cleartoline
 *
 *    ٷ Ѿϴ.
 *-------------------------------------------------------------------*/
void console_cleartoline(void)
{
  // Do a line feed
  while( char_cursor.x < COLS ) {
      *char_cursor.p++ = ' ';
      *char_cursor.p++ = char_attribute;
      char_cursor.x++;
  }
  char_cursor.x = 0;

  // May need to scroll a line
  if( ++char_cursor.y > ROWS-1 ) {
      console_scroll(1);
      char_cursor.y = ROWS-1;
      char_cursor.p = (char *)(VID_PTR + (ROWS-1)*ROWSIZE);
  }
}


/*-------------------------------------------------------------------*
 * console_reset_cursor
 *
 * Ŀ մϴ. 
 *-------------------------------------------------------------------*/
static void console_reset_cursor(void)
{
  char_cursor.p = (byteptr)VID_PTR;
  char_cursor.x = char_cursor.y = 0;
}

/*-------------------------------------------------------------------*
 * console_init
 *
 * ȭ ־   ȭ  մϴ.
 *-------------------------------------------------------------------*/
void console_init(byte bg, byte fg)
{
  // Setup attribute byte
  char_attribute = 0x00;
  console_set_background(bg);
  console_set_foregournd(fg);
  console_clear_screen();
}

/*-------------------------------------------------------------------*
 * console_clear_screen()
 *
 *  ȭ ϴ.
 *-------------------------------------------------------------------*/
void console_clear_screen(void)
{
  int i;
  char *p = char_cursor.p;
  console_reset_cursor();
  // clear the screen to the background
  for( i=0; i<VID_SIZE; *p++=' ', *p++=char_attribute, i++ );
}

/*-------------------------------------------------------------------*
 * console_putchar
 *
 *  Ŀ ġ  ڸ ¶  ĭ ̵
 * մϴ.
 *-------------------------------------------------------------------*/
void console_putchar( char c )
{
  *char_cursor.p++ = c;
  *char_cursor.p++ = char_attribute;
   char_cursor.x++;
}

/*-------------------------------------------------------------------*
 * console_set_background
 *
 *  մϴ.
 *-------------------------------------------------------------------*/
void console_set_background( byte bg )
{
  //clear the current background
  char_attribute &= BG_MSK;
  //set the background
  char_attribute |= (bg << 4);
}

/*-------------------------------------------------------------------*
 * console_set_foregournd
 *
 *  մϴ.
 *-------------------------------------------------------------------*/
void console_set_foregournd( byte fg )
{
  //clear the current foreground
  char_attribute &= FG_MSK;
  //set the foreground
  char_attribute |= fg;
}

/*-------------------------------------------------------------------*
 * console_set_blink
 *
 * blink() մϴ.
 *-------------------------------------------------------------------*/
void console_set_blink( bool set_bit )
{
  //clear the current blink bit
  char_attribute &= BLNK_MSK;
  //set it
  char_attribute |= ( set_bit ? (0x1 << 7) : (0x0) );
}

/*-------------------------------------------------------------------*
 * console_set_intensity
 *
 * intensity Ʈ մϴ.
 *-------------------------------------------------------------------*/
void console_set_intensity( bool set_bit )
{
  //clear the current intensity bit
  char_attribute &= INTSTY_MSK;
  //set it
  char_attribute |= ( set_bit ? (0x1 << 3) : (0x0) );
}

/*-------------------------------------------------------------------*
 * console_goto_xy
 *
 * ȭ x, y  Ŀ ̵մϴ.
 *-------------------------------------------------------------------*/
void console_goto_xy( byte x, byte y )
{
  size_t offset = 2*((int)x*ROWS + (int)y*COLS);
  if( (x < COLS-1) && (y < ROWS-1) ) {
     char_cursor.p = (byteptr)VID_PTR + offset;
     char_cursor.x = x;
     char_cursor.y = y;
  }
}

/*-------------------------------------------------------------------*
 * console_get_x --  x ġ մϴ.
 *-------------------------------------------------------------------*/
inline byte console_get_x(void) { return char_cursor.x; }


/*-------------------------------------------------------------------*
 * console_get_y --  y ġ մϴ.
 *-------------------------------------------------------------------*/
inline byte console_get_y(void) { return char_cursor.y; }


/*-------------------------------------------------------------------*
 * kprintf -- string.h  ִ vprintf մϴ.
 *
 * ֻܼ  Ģ ڿ մϴ.
 * ⺻  ڸ ٷ  ֽϴ.
 *-------------------------------------------------------------------*/
void kprintf( const char *fmt, ... )
{
  	va_list args;
  	char buf[256];
  	int i, j, n;
	struct current_tty_buffer * tt;

	tt = tty_buffer;

  	va_start(args, fmt);
  	n = vsprintf(buf, fmt, args);
  	va_end(args);
  	for (i = 0; i < n; i++) {
      	switch (buf[i]) {
          	case '\t': for (j = char_cursor.x % TAB_SIZE; j < TAB_SIZE; j++) {
							console_putchar(' ');
							/* ȭ鿡 µǴ ڵ buffer  ϴ.
							 * ̴ view_prev_screen Ȥ view_next_screen Լ 
							 * Ŀ ޼   ߿ ˴ϴ. */
							tt->buffer[tty_buffer_ptr] = ' ';
							/*  µ ڰ  buffer  ũ   
							 *  ó ư  ǰ ˴ϴ.  weix
							 * 1024*8  ڸ մϴ. */
							if(tty_buffer_ptr >= ((1024*8)-1))
								tty_buffer_ptr=0;
							else
								tty_buffer_ptr++;
						}
                	     	break;
          	case '\n':
          	case '\r': console_cleartoline();
						tt->buffer[tty_buffer_ptr] = '\n';
                        if(tty_buffer_ptr >= ((1024*8)-1))
                            tty_buffer_ptr=0;
                        else
                            tty_buffer_ptr++;
                     	break;
          	default  : if (char_cursor.x >= 79) {			/*  üũմϴ. */
							console_cleartoline();
							console_putchar(buf[i]);
							tt->buffer[tty_buffer_ptr] = buf[i];
                            if(tty_buffer_ptr >= ((1024*8)-1)) 
                                tty_buffer_ptr=0;
                            else
								tty_buffer_ptr++;
					 	} else {
							console_putchar(buf[i]);
                            tt->buffer[tty_buffer_ptr] = buf[i];
                            if(tty_buffer_ptr >= ((1024*8)-1)) 
                                tty_buffer_ptr=0;
                            else
                            	tty_buffer_ptr++;
					 	}
					 	break;
      }
  	}
}

/*-------------------------------------------------------------------*
 * k_error
 *-------------------------------------------------------------------*/
void k_error(const char * fmt, ...) {
    va_list args;
	int n;
    char buf[1024];

    va_start(args, fmt);
    n = vsprintf(buf, fmt, args);
    va_end(args);
	console_set_foregournd(COL_RED);
	kprintf("ERROR: %s\n", buf);
	console_set_foregournd(COL_GRAY);
}

/*-------------------------------------------------------------------*
 * Name : void show_CPU_reg()
 * describe : show current Registers status
 *-------------------------------------------------------------------*/
void show_CPU_reg()
{
	__asm__ ("
		push %ds\n
		push %es\n
		push %fs\n
		push %gs\n
		pusha\n
		pushfl\n

		movl $0x23, %eax\n
		mov  %eax, %ds\n

		call registers_handler\n

		popfl\n
		popa\n
		pop %gs\n
		pop %fs\n
		pop %es\n
		pop	%ds\n
	");
}

/*-------------------------------------------------------------------*
 * registers_handler
 *-------------------------------------------------------------------*/
void registers_handler(dword eflags, dword edi, dword esi, dword ebp, dword esp, 
						dword ebx, dword edx, dword ecx, dword eax, 
						dword gs, dword fs, dword es, dword ds)
{
	int i;
	ulong *p = (ulong *) esp;
	dword reg_cr0, reg_cr2, reg_cr3, reg_cr4;

	kprintf("Current Registers status.\n");

   	kprintf("EAX=0x%08X  EBX=0x%08X  ECX=0x%08X  EDX=0x%08X\n", \
       	     eax, ebx, ecx, edx);
 	kprintf("EBP=0x%08X  ESI=0x%08X  EDI=0x%08X  EFLAGS=0x%08X\n", \
   		   	  ebp, esi, edi, eflags);
  	kprintf("DS =0x%04X      ES =0x%04X      FS =0x%04X      GS =0x%04X\n", \
      		  ds & 0xffff, es & 0xffff, fs & 0xffff, gs & 0xffff);

	for(i=-12;i<24;i+=6) {
		kprintf("STACK[%02d]=%08X %08X %08X %08X %08X %08X\n", i, p[i], p[i+1], p[i+2], p[i+3], p[i+4], p[i+5]);
	}

	__asm__ __volatile__ ("
		movl 	%%cr0, %%ebx\n\t
		movl 	%%ebx, %0\n\t"
		: "=g" (reg_cr0)
		: 
		: "ebx"
		);
	kprintf("CR0=0x%08X  ", reg_cr0);

    __asm__ __volatile__ ("
        movl   %%cr2, %%ebx\n\t
        movl   %%ebx, %0\n\t"
        : "=g" (reg_cr2)
        :
		: "ebx"
        );
    kprintf("CR2=0x%08X  ", reg_cr2);

    __asm__ __volatile__ ("
        movl   %%cr3, %%ebx\n\t
        movl   %%ebx, %0\n\t"
        : "=g" (reg_cr3)
        :
		: "ebx"
        );
    kprintf("CR3=0x%08X  ", reg_cr3);

    __asm__ __volatile__ ("
        movl   %%cr4, %%ebx\n\t
        movl   %%ebx, %0\n\t"
        : "=g" (reg_cr4)
        :
		: "ebx"
        );
    kprintf("CR4=0x%08X\n", reg_cr4);
}

void kprintchar(unsigned char c)
{
	if( c >= 0x20 )
		kprintf("%c", c);
	else
		kprintf(".");
}


/*-------------------------------------------------------------------*
 * Name : dump_memory
 * Desc : dump memory from s_addr to (s_addr+size).
 *-------------------------------------------------------------------*/
void dump_memory(ulong *buff, int len)
{
	int i;
	unsigned char *ptr=(uchar *)buff;


	kprintf("Memory dump : start_addr[0x%08X], last_addr[0x%08X]\n", buff, (buff + len));

	for( i = 0; i<len; i++)
	{
		if ( i % 16 == 0 )
			kprintf("0x%08x  ", &buff[i]);

		kprintf("%02x ", ptr[i]);

		if ( i % 16 - 15 == 0 )
		{
			int	j;
			kprintf("  ");
			for(j=i-15; j<=i; j++)
				kprintchar(buff[j]);
			kprintf("\n");
		}
	}
	if ( i%16 != 0 )
	{
		int j;
		int spaces=(len=i+16-i%16)*3+2;
		for(j = 0; j < spaces; j++)
			kprintf("  ");

		for(j=i - i%16; j < len; j++)
			kprintchar(buff[j]);
	}

	kprintf("\n");
}

/*-------------------------------------------------------------------*
 * show_IDT_table
 *-------------------------------------------------------------------*/
void show_IDT_table(int idt_start, int idt_end)
{
	gate *i = (gate *) (GDT_SIZE * sizeof(descriptor) + idt_start * sizeof(gate));
	dword i_offset;

	while (idt_start <= idt_end) {
		i_offset = i->offset16_31 << 16;
		i_offset += i->offset0_15;

		kprintf("- IDT[%d] = %08X --------------------------------------------------\n", idt_start, &i[idt_start]);
		kprintf("select=%04X\t", i->selector);
		kprintf("offset=%08X\t", i_offset);
		kprintf("p=%d\t", i->p);
		kprintf("dpl=%d\t", i->dpl);
		kprintf("type=%d\n", i->type);

		idt_start++;
		i++;
	}
}

/*-------------------------------------------------------------------*
 * show_GDT_table
 *-------------------------------------------------------------------*/
void show_GDT_table(int gdt_start, int gdt_end)
{
    descriptor *g = (descriptor *) (0x0 + gdt_start * sizeof(descriptor));
    dword g_base;
	dword tmp;
	dword g_limit;

    while (gdt_start <= gdt_end) {
        g_base = g->base24_31 << 24;
		tmp    = g->base16_23 << 16;
		g_base += tmp;
		g_base += g->base0_15;

        g_limit = g->limit16_19 << 16;
        g_limit += g->limit0_15;
		g_limit = g_limit << 12;

        kprintf("- GDT[%d] = %08X -------------------------------------------------\n", gdt_start, &g[gdt_start]);
        kprintf("base=%08X\t", g_base);
        kprintf("limit=%08X\t", g_limit);
        kprintf("p=%d\t", g->p);
        kprintf("dpl=%d\t", g->dpl);
        kprintf("type=%d\t", g->type);
		kprintf("db=%d\t", g->db);
		kprintf("s=%d\n", g->s);

        gdt_start++;
        g++;
    }
}

/*-------------------------------------------------------------------*
 * show_task_struct
 *-------------------------------------------------------------------*/
void show_task_struct(struct my_life_schedule *p)
{
	kprintf("+ my_life_schedule p addr = %08X , p->tss addr=%08X, p->ldt addr=%08X-----\n", p, &(p->tss), &(p->ldt));
	kprintf("- state=%X, counter=%X, my_priority=%X, signal=%X\n", p->state, p->counter, p->my_priority, p->signal);
	kprintf("- exit_code=%d, end_of_code=%08X, end_of_data=%08X, brk=%08X\n", p->exit_code, p->end_of_code, p->end_of_data, p->brk);
	kprintf("- start_stack=%08X, pid=%X, my_father=%X, pgrp=%X, session=%X, leader=%X\n", p->start_stack, p->pid, p->my_father, p->pgrp, p->session, p->leader);
	kprintf("- uid=%u, euid=%u, suid=%u\n", p->uid, p->euid, p->suid); 
	kprintf("- gid=%u, egid=%u, sgid=%u\n", p->gid, p->egid, p->sgid);
	kprintf("- alarm=%X utime=%X, stime=%X, cutime=%X, cstime=%X, start_time=%X\n", p->alarm, p->utime, p->stime, p->cutime, p->cstime, p->start_time);
	kprintf("- tty=%d, umask=%u\n", p->tty, p->umask);
	kprintf("- tss.esp0=%08X, tss.ss0=%04X, tss.esp1=%08X, tss.ss1=%04X\n", p->tss.esp0, p->tss.ss0, p->tss.esp1, p->tss.ss1); 
	kprintf("- tss.esp2=%08X, tss.ss2=%04X, tss.cr3=%08X, tss.eip=%08X\n", p->tss.esp, p->tss.ss2, p->tss.cr3, p->tss.eip);
	kprintf("- tss.eflags=%08X, tss.eax=%08X, tss.ecx=%08X, tss.edx=%08X\n", p->tss.eflags, p->tss.eax, p->tss.ecx, p->tss.edx);
	kprintf("- tss.ebx=%08X, tss.esp=%08X, tss.ebp=%08X, tss.esi=%08X\n", p->tss.ebx, p->tss.esp, p->tss.ebp, p->tss.esi);
	kprintf("- tss.edi=%08X, tss.es=%04X, tss.cs=%04X, tss.ss=%04X\n", p->tss.edi, p->tss.es, p->tss.cs, p->tss.ss);
	kprintf("- tss.ds=%04X, tss.fs=%04X, tss.gs=%04X, tss.ldt=%04X\n", p->tss.ds, p->tss.fs, p->tss.gs, p->tss.ldt);
}

/*-------------------------------------------------------------------*
 * where_is1, where_is2, where_is3
 *
 * test functions in assembly envirments.
 * where_is3 can use "find address" :-)
 *-------------------------------------------------------------------*/
void where_is1()
{
	kprintf("Oh!!! shiT!!!!!!!!~~~~~~ I'm here in 1\n");
}
void where_is2()
{
    kprintf("Oh!!! shiT!!!!!!!!~~~~~~ I'm here in 2\n");
}
void where_is3(ulong w_value)
{
    kprintf("Oh!!! shiT!!!!!!!!~~~~~~ I'm here in 3\n");
	kprintf("w_value = %08X\n", w_value);
}
