/*-------------------------------------------------------------------*
 * kernel/mm/paging.c
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/page.h>
#include <weix/mm.h>
#include <weix/params.h>
#include <weix/types.h>
#include <asm/io.h>
#include <weix/idt.h> /* for struct gate */
#include <weix/gdt.h> /* for struct descriptior */
#include <weix/string.h>

#ifdef DEBUG
#include <weix/kdebug.h>
#endif

/*-------------------------------------------------------------------*
 * defines
 *-------------------------------------------------------------------*/
#define invalidate() \
__asm__ __volatile__("movl %%cr3,%%eax\n\tmovl %%eax,%%cr3": : :"ax")

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
/* Page Directory   ּ
 * 0x800  ߽ϴ.   Page Directory  4k  
 * ־ ϱ Դϴ.
 *  ġ ּ = 0x81000
*/
page_dir_entry *PDE = (page_dir_entry *)((GDT_SIZE * sizeof(descriptor)) + (IDT_SIZE * sizeof(gate)) + 0x800 + 0x70000);

/* Page Table  ּ
 *  ġ ּ = 0x82000
 */
page_table_entry *PTE = (page_table_entry *)((GDT_SIZE * sizeof(descriptor)) + (IDT_SIZE * sizeof(gate)) + (PDE_SIZE * sizeof(page_dir_entry)) + 0x800 + 0x70000);

/*-------------------------------------------------------------------*
 * setup_paging
 *-------------------------------------------------------------------*/
void setup_paging(ulong mem_start, ulong mem_end)
{
	int i;

	/*  丮,  ̺  0 ʱȭմϴ. */
	/*  ġ 0x81000  0x82000 Դϴ. */
  	memset((void *)PDE, 0, PDE_SIZE * sizeof(page_dir_entry));
	/*  ̺  ּҴ 0x82000  0x86000 Դϴ. */
	memset((void *)PTE, 0, 4 * PTE_SIZE * sizeof(page_table_entry));

	/*  丮  P/W/U/0/0/0/0/page_table_start_addr  մϴ. */
	/* ̰ PDE[0], PDE[1] Ʒ  ϴ. -----------------------
	 * PDE[0] = 0x00082007
	 * PDE[1] = 0x00083007 */
	/* 0~4M */
	seting_PDE(&PDE[0], 1, 1, 1, 0, 0, 0, 0, (dword)(&PTE[0*1024]));
                                /* PTE[1024] 0x83000 ŵϴ. */
	/* 4~8M */
	seting_PDE(&PDE[1], 1, 1, 1, 0, 0, 0, 0, (dword)(&PTE[1*1024]));
	/* 0xc0000000 ~ 0xc0000000+8M */
	seting_PDE(&PDE[768], 1, 1, 1, 0, 0, 0, 0, (dword)(&PTE[2*1024]));
	seting_PDE(&PDE[769], 1, 1, 1, 0, 0, 0, 0, (dword)(&PTE[3*1024]));

	/* dump memory */
	// dump_memory((dword *) 0x81000, 8 * sizeof(page_dir_entry));

	/*  ̺ P/W/U/0/0/0/0/kernel_start_addr  մϴ.
	 * Ŀ ϵ ýÿ 0x100000(1M) ε ǰ ÷ ýÿ
	 * 0x0 ġ εǰ ˴ϴ.
	 * PTE[0]    = 00001007
	 * PTE[1]    = 00002007
	 *     ~
	 * PTE[2047] = 00800007 */
	
	/*  ½ϴ. PTE   1024 Դϴ.
	 * 2  丮 شϴ  ̺ ϱ  
	 * ̷ մϴ. ׷  ó (2048 * sizeof(page_table_entry)) 
	 * Ҵ Դϴ. ׷ PTE[] 1024 Ѿϴ. ̻  ϴ. :-) */
	for(i=0 ; i < (PTE_SIZE*2) ; i++) 
		seting_PTE(&PTE[i], 1, 1, 1, 0, 0, 0, 0, (i * 0x1000));

	/* dump memory */
	// dump_memory((dword *) 0x82000, 8 * sizeof(page_table_entry));

	/*  ¡ enable ŵϴ. */
	enable_paging();

	invalidate();
}

/*-------------------------------------------------------------------*
 * seting_PDE
 *-------------------------------------------------------------------*/
void seting_PDE(page_dir_entry *pd, byte p, byte w, byte u, byte pwt, byte pcd, byte a, byte ps, dword pdaddr)
{
	pd->p = p;	
	pd->w = w;
	pd->u = u;
	pd->pwt = pwt;
	pd->pcd = pcd;
	pd->a = a;
	pd->ps = ps;
	pd->pdaddr_12_23 = pdaddr >> 12;
	pd->pdaddr_24_31 = pdaddr >> 24;
}

/*-------------------------------------------------------------------*
 * seting_PTE
 *-------------------------------------------------------------------*/
void seting_PTE(page_table_entry *pt, byte p, byte w, byte u, byte pwt, byte pcd, byte a, byte d, dword ptaddr)
{
    pt->p = p;
    pt->w = w;
    pt->u = u;
    pt->pwt = pwt;
    pt->pcd = pcd;
    pt->a = a;
    pt->d = d;
    pt->ptaddr_12_23 = ptaddr >> 12;
    pt->ptaddr_24_31 = ptaddr >> 24;
}

/*-------------------------------------------------------------------*
 * enable_paging
 *-------------------------------------------------------------------*/
void enable_paging(void)
{
	__asm__ ("
		/* i put address of PDE directly */
		/* 0x81000  PDE ּԴϴ. */
		movl $0x81000, %%eax\n\t
		movl %%eax, %%cr3\n\t
		movl %%cr0, %%eax\n\t
		orl $0x80000000, %%eax\n\t
		movl %%eax, %%cr0\n\t
		jmp	 1f
	1:
		movl $1f, %%eax
		jmp	 *%%eax
	1:
	"
	:
	:
	: "eax"	/* clobber list */
	);
}
