/*-------------------------------------------------------------------*
 * /kernel/pic.c
 *
 * Description:
 * Programable Interrupt Controller functions
 *-------------------------------------------------------------------*/

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <weix/types.h>
#include <asm/io.h>
#include <weix/pic.h>

/*-------------------------------------------------------------------*
 * global variables
 *-------------------------------------------------------------------*/
/* 16-bits of IRQ masking */
static word IRQmask;

/*-------------------------------------------------------------------*
 * defines
 *-------------------------------------------------------------------*/
#define SLAVE_IRQ 8
#define MASTER_SLAVE 2
#define	ICU0 0x20
#define	ICU1 0xA0
#define	ICU_RESET 0x11

/*-------------------------------------------------------------------*
 * reprogramPIC() Լ  ڵ带  ϰ ֽϴ.
 *-------------------------------------------------------------------*/
void reprogram_PIC(void)
{
	ushort master=0x20;
	ushort slave=0x28;

  /* move irqs to 0x20-0x2F
   * this makes room for the exceptions 0x0-0x1F */

  outb(ICU0, ICU_RESET);
  io_wait();
  outb(ICU1, ICU_RESET);
  io_wait();
  outb(ICU0+1, master);
  io_wait();
  outb(ICU1+1, slave);
  io_wait();
  outb(ICU0+1, 0x04);
  io_wait();
  outb(ICU1+1, 0x02);
  io_wait();
  outb(ICU0+1, 0x01);
  io_wait();
  outb(ICU1+1, 0x01);
  io_wait();
  outb(ICU0+1, 0xFB);
  io_wait();
  outb(ICU1+1, 0xFF);

  IRQmask = 0xFFFB; /* all IRQs are masked except IRQ2 */
}

/*-------------------------------------------------------------------*
 * unmask_IRQ
 *-------------------------------------------------------------------*/
void unmask_IRQ(int irq)
{
  IRQmask &= ~(1 << irq);
  if(irq < 8)
    outb(ICU0+1, IRQmask & 0xFF);
  else
    outb(ICU1+1, IRQmask >> 8);
}

/*-------------------------------------------------------------------*
 * mask_IRQ
 *-------------------------------------------------------------------*/
void mask_IRQ(int irq)
{
  IRQmask |= (1 << irq);
  if(irq < 8)
    outb(ICU0+1,IRQmask & 0xFF);
  else
    outb(ICU1+1,IRQmask >> 8);
}

/*-------------------------------------------------------------------*
 * refresh_IRQ
 *-------------------------------------------------------------------*/
void refresh_IRQ(void)
{
  outb(ICU0+1,IRQmask & 0xFF);
  outb(ICU1+1,IRQmask >> 8);
}
