/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/system.h>
#include <asm/io.h>
#include <weix/signal.h>
#include <weix/head.h>
#include <weix/interrupt.h>
#include <weix/k_stat.h>
#include <weix/linkage.h>
#include <weix/page.h>
#include <weix/sched.h>
#include <weix/sys.h>
#include <weix/timer.h>
#include <weix/idt.h>
#include <weix/gdt.h>
#include <weix/kdebug.h>
#include <weix/wait.h>
#include <weix/extern_func_T_T.h>

#define LATCH (1193180/HZ)
#define TIMER_IRQ 0

/*-------------------------------------------------------------------*
 * preferences
 *-------------------------------------------------------------------*/
void schedule(void);
int add_run_queue(struct my_life_schedule * );

/*-------------------------------------------------------------------*
 * global variables 
 *-------------------------------------------------------------------*/
unsigned long timer_active = 0;
struct timer_struct timer_table[32];
struct kernel_stat kstat =
    { 0, 0, 0, { 0, 0, 0, 0 }, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

/*  ü [my_life_schedule + stack]  մϴ. */
union my_life_union {
    struct my_life_schedule task;
    char stack[PAGE_SIZE];			/*  μ ڽ   ֽϴ. */
};

/* initial taskԴϴ. */
static union my_life_union init_task = {INIT_TASK,};
/* δ  μ մϴ. */
struct weix_goverment goverment = INIT_GOVER;

long volatile jiffies=0;
/* ý ۽ startup_time  õ˴ϴ. */
long startup_time=0;
/*  need_resched  1̸, schedule Լ ȣ˴ϴ. */
int need_resched = 0;

/* current_gover  θ ŵϴ. */
struct weix_goverment * current_gover = &(goverment);
/* current point  task ŵϴ. */
struct my_life_schedule *current = &(init_task.task), *last_task_used_math = NULL;
/* run_queue pointer */
struct wait_queue * run_queue;
/* run_queue_prt  run_queue   task ã Ű Դϴ. */
struct wait_queue * run_queue_ptr;

/*-------------------------------------------------------------------*
 * find_best_task
 *
 *  Լ run_queue 꿡 ִ task ߿  ˸
 * task    task  (ڽ  pid) 
 * մϴ.
 *-------------------------------------------------------------------*/
int find_best_task(void)
{
	int i=1, j=0, z=0, c=-1, ret_a;
	struct wait_queue * t = run_queue_ptr, * k=run_queue_ptr, * tmp;
	struct my_life_schedule * best;

	/* alarm üũϿ signal  ִ interruptible task  ϴ. */
	while(k->task != NULL) {
		if(k->task->alarm && k->task->alarm < jiffies) {
			k->task->signal |= (1<<(SIGALRM-1));
			k->task->alarm = 0;
		}
		if(k->task->signal && k->task->state==TASK_INTERRUPTIBLE) {
			k->task->state=TASK_RUNNING;
			if(!(ret_a=add_run_queue(k->task))) {
				kprintf("\nHmm... can't add task in run_queue maybe run_queue FULL\n");
			}
		}
		if(z>MAX_NR_RUN_QUEUE)
			break;
		k=k->next;
		z++;
	}

	while(1) {
		/* run_queue->task ׸ NULL ƴ  ãҽϴ! */
		if((t->task != NULL) && t->task->state == TASK_RUNNING && t->task->counter > c) {
			best = t->task;		/* Best  ϴ  մϴ. */
			t->task = NULL;		/* run_queue  ִ  NULL · ϴ. */
			tmp = t->next;
			/* run_queue task  ġ  task մϴ. */
			while(1) {
				if(tmp->task == NULL) {
					tmp->task = best;
					break;
				}
				tmp = tmp->next;
			}
			/* citizen[NR_TASKS] ִ  õ task ȣ ãϴ. */
			for(j=0;j<NR_TASKS;j++) {
				if(best == current_gover->citizen[j])
					break;
			}
			c = best->counter;
			if(c) {
				run_queue_ptr = run_queue_ptr->next;    /*  schedule ؼ run_queue_ptr   űϴ. */
				return j;
			}
			best->counter = (best->counter >> 1) + best->my_priority;
		}

		if(i>MAX_NR_RUN_QUEUE)
			break;
		t = t->next;
		i++;
	}

	/*  run_queue  ƹ  ٸ idle task մϴ.
	 * 0 Ѵٴ  citizen[0] == idle  Ѵٴ ǹԴϴ. */
	return 0;
}

/*-------------------------------------------------------------------*
 * add_run_queue
 *
 * run_queue_ptr Ͽ run_queue  task ׸
 * NULL  ׸ ã  μ task մϴ. 
 *-------------------------------------------------------------------*/
int add_run_queue(struct my_life_schedule * p)
{
	int i=0;
	struct wait_queue *t=run_queue_ptr;
	while(1) {
		if(t->task == NULL) {
			t->task = p;
			return 1;
		}
		t=t->next;
		if(i>MAX_NR_RUN_QUEUE)
			break;
		i++;
	}
	return 0;
}

/*-------------------------------------------------------------------*
 * schedule
 *
 * Task 0  ̻ ų μ   Ǵ
 * idle task Դϴ. ̰ kill   , sleep  
 * ϴ. goverment->citizen[0]  state ׸ 
 *  ʽϴ. ( :  :-))
 *-------------------------------------------------------------------*/
void schedule(void)
{
    int next;

	cli();
#if 0
	kprintf("in schedule, ");
#endif

	/* clean need_resched */
	need_resched=0;

	/* Ʒ ִ  ο ٷ Դϴ. */
	/*   task ٸ idle task մϴ. */
    next = find_best_task();

#if 0
	kprintf("next = %d\n", next);
#endif
	sti();
	switch_to(next);

#if 0
	kprintf("task switch success :-)\n");
	show_task_struct(current);
#endif
}

/*-------------------------------------------------------------------*
 * sys_pause
 *-------------------------------------------------------------------*/
int sys_pause(void)
{
	need_resched=1;
    return 0;
}

/*-------------------------------------------------------------------*
 * __sleep_on
 *-------------------------------------------------------------------*/
static inline void __sleep_on(struct my_life_schedule **p, int state)
{
    struct my_life_schedule *tmp;

    if (!p)
        return;
    if (current == &(init_task.task))
        panic("task[0] trying to sleep");
    tmp = *p;
    *p = current;
    current->state = state;
repeat: schedule();
    if (*p && *p != current) {
        (**p).state = 0;
        current->state = TASK_UNINTERRUPTIBLE;
        goto repeat;
    }
    if (!*p)
        kprintf("Warning: *P = NULL\n\r");
    if ((*p = tmp))
        tmp->state=0;
}

/*-------------------------------------------------------------------*
 * interruptible_sleep_on
 *-------------------------------------------------------------------*/
void interruptible_sleep_on(struct my_life_schedule **p)
{
	__sleep_on(p,TASK_INTERRUPTIBLE);
}

/*-------------------------------------------------------------------*
 * sleep_on
 *-------------------------------------------------------------------*/
void sleep_on(struct my_life_schedule **p)
{
	kprintf("sleep_on : in fuction\n");
    __sleep_on(p,TASK_UNINTERRUPTIBLE);
}

/*-------------------------------------------------------------------*
 * wake_up
 *-------------------------------------------------------------------*/
void wake_up(struct my_life_schedule **p)
{
    if (p && *p) {
        if ((**p).state == TASK_STOPPED)
            kprintf("wake_up: TASK_STOPPED");
        if ((**p).state == TASK_ZOMBIE)
            kprintf("wake_up: TASK_ZOMBIE");
        (**p).state=TASK_RUNNING;
		add_run_queue(*p);
    }
}

/*-------------------------------------------------------------------*
 * do_timer
 *-------------------------------------------------------------------*/
void do_timer(long cpl)
{
    if (cpl) {
        current->utime++;
    } else {
        current->stime++;
	}
    if ((--current->counter)>0) {
		return;
	}
    current->counter=0;
    if (!cpl) 
		return;
    schedule();
}

/*-------------------------------------------------------------------*
 * sys_alarm
 *-------------------------------------------------------------------*/
int sys_alarm(long seconds)
{
    current->alarm = (seconds>0)?(jiffies+HZ*seconds):0;
    return seconds;
}

/*-------------------------------------------------------------------*
 * sys_getpid
 *-------------------------------------------------------------------*/
int sys_getpid(void)
{
    return current->pid;
}

/*-------------------------------------------------------------------*
 * sys_getppid
 *-------------------------------------------------------------------*/
int sys_getppid(void)
{
    return current->my_father->pid;
}

/*-------------------------------------------------------------------*
 * sys_getuid
 *-------------------------------------------------------------------*/
int sys_getuid(void)
{
    return current->uid;
}

/*-------------------------------------------------------------------*
 * sys_geteuid
 *-------------------------------------------------------------------*/
int sys_geteuid(void)
{
    return current->euid;
}

/*-------------------------------------------------------------------*
 * sys_getgid
 *-------------------------------------------------------------------*/
int sys_getgid(void)
{
    return current->gid;
}

/*-------------------------------------------------------------------*
 * sys_getegid
 *-------------------------------------------------------------------*/
int sys_getegid(void)
{
    return current->egid;
}

/*-------------------------------------------------------------------*
 * sys_nice
 *-------------------------------------------------------------------*/
int sys_nice(long increment)
{
    if (current->my_priority-increment>0)
        current->my_priority -= increment;
    return 0;
}

/*-------------------------------------------------------------------*
 * sys_signal
 *-------------------------------------------------------------------*/
int sys_signal(long signal,long addr,long restorer)
{
	kprintf("FIXME : sys_signal\n");
	return -1;
}

/*-------------------------------------------------------------------*
 * timer_bh
 *-------------------------------------------------------------------*/
static void timer_bh(void * unused)
{
	kprintf("FIXME : timer bottom half clalled\n");
	return;
}

/*-------------------------------------------------------------------*
 * sched_init
 *-------------------------------------------------------------------*/
void sched_init(void)
{
	int i;

	/* TSS  LDT GDT մϴ. */
	setup_descriptor(&GDT[FIRST_TSS_ENTRY], (ulong )(&(init_task.task.tss)), sizeof(init_task.task.tss), 9, 0, 0, 0);
	setup_descriptor(&GDT[FIRST_LDT_ENTRY], (ulong )(&(init_task.task.ldt)), sizeof(init_task.task.ldt), 2, 0, 0, 0);

	/* GDT + 8 ŵϴ.
	 * --  GDT  --
	 * GDT[0] = dummy code	= 0x0
	 * GDT[1] = KERNEL CS 	= 0x8
	 * GDT[2] = KERNEL DS 	= 0x10
	 * GDT[3] = USER CS 	= 0x18
	 * GDT[4] = USER DS 	= 0x20
	 * GDT[5] = DUMMY
	 * GDT[6] = TSS0 
	 * GDT[7] = LDT0 
	 */

	/* μ ̺  ûմϴ. */
    for(i=1; i<NR_TASKS; i++)
        current_gover->citizen[i] = NULL;
	/*  ںе  ûմϴ. */
	for(i=0; i<NR_GOVER_STAFF; i++)
		current_gover->gover_staff[i] = NULL;
    /* run_queue ͸ մϴ. */
    run_queue = current_gover->run_queue_list;
	/* run queue  û԰ Ҿ linked-list  ϴ. */
	for(i=0; i<MAX_NR_RUN_QUEUE; i++) {
		if(i<(MAX_NR_RUN_QUEUE-1)) {	/* 1  (MAX_NR_RUN_QUEUE-1) մϴ. */
			(run_queue+i)->task = NULL;
			/* run_queue->next  run_queue  ŵϴ. */
			(run_queue+i)->next = run_queue+i+1;
		}
		/*   run_queue  ó run_queue Ű մϴ. */
		if(i == (MAX_NR_RUN_QUEUE-1)) {
			(run_queue+i)->task = NULL;
			(run_queue+i)->next = run_queue;
		}
	}
	/* run_queue ͸ մϴ. */
	run_queue_ptr = run_queue;

	/* Eflags  NT Ʈ ϴ. */
	__asm__("pushfl ; andl $0xffffbfff,(%esp) ; popfl");
	/* TR ͸ մϴ. */
    ltr(0);
	/* LDT  ٽ մϴ. */
    lldt(0);
	
	/* Ŭ time  ٲߴϴ. */
    outb_p(0x43, 0x36);      	/* binary, mode 3, LSB/MSB, ch 0 */
								/* 0x36 = command for 16-bit */
							 	/* 0x43h -> mode 3, binary operation */
    outb_p(0x40, LATCH & 0xff);    	/* LSB */ /* counter value = LATCH */
								   	/* load timer 0 count LSB */
    outb(0x40, LATCH >> 8);    		/* outb(port, value) */ /* MSB */
								   	/* load timer 0 count MSB */
	/* IDT[128] = system_call */
	trap_IDT(&IDT[0x80], KERNEL_CS,(ulong ) &system_call, 3);

	/* Ÿ̸Ӹ մϴ. */
	bh_base[TIMER_BH].routine = timer_bh;
	if (request_irq(TIMER_IRQ,(void (*)(int)) do_timer)!=0)
        panic("Could not allocate timer IRQ!");
}
