/* Modified by andsoon  */

/*
 * Weix Operating System
 * Copyright (C) 2001 Jung Won Gyo
 *
 * This program is  free software; you can  redistribute it and/or modify
 * it under the terms  of the GNU General Public  License as published by
 * the  Free Software Foundation; either version  2,  or (at your option)
 * any later version.
 *
 * This program is distributed  in the hope that  it will be  useful, but
 * WITHOUT    ANY  WARRANTY;  without   even   the  implied  warranty  of
 * MERCHANTABILITY  or  FITNESS FOR  A PARTICULAR   PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should  have received a  copy of  the GNU  General Public  License
 * along with   this  program; if   not,  write  to   the Free   Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * sys_call.S  contains the system-call and fault low-level handling routines.
 * This also contains the timer-interrupt handler, as well as all interrupts
 * and faults that can result in a task-switch.
 *
 * NOTE: This code handles signal-recognition, which happens every time
 * after a timer-interrupt and after each system call.
 *
 * I changed all the .align's to 4 (16 byte alignment), as that's faster
 * on a 486.
 *
 * Stack layout in 'ret_from_system_call':
 * 	ptrace needs to have all regs on the stack.
 *	if the order here is changed, it needs to be 
 *	updated in fork.c:copy_process, signal.c:do_signal,
 *	ptrace.c and ptrace.h
 *
 *	 0(%esp) - %ebx
 *	 4(%esp) - %ecx
 *	 8(%esp) - %edx
 *       C(%esp) - %esi
 *	10(%esp) - %edi
 *	14(%esp) - %ebp
 *	18(%esp) - %eax
 *	1C(%esp) - %ds
 *	20(%esp) - %es
 *      24(%esp) - %fs
 *	28(%esp) - %gs
 *	2C(%esp) - orig_eax
 *	30(%esp) - %eip
 *	34(%esp) - %cs
 *	38(%esp) - %eflags
 *	3C(%esp) - %oldesp
 *	40(%esp) - %oldss
 */

#include <asm/segment.h>

EBX		= 0x00
ECX		= 0x04
EDX		= 0x08
ESI		= 0x0C
EDI		= 0x10
EBP		= 0x14
EAX		= 0x18
DS		= 0x1C
ES		= 0x20
FS		= 0x24
GS		= 0x28
ORIG_EAX	= 0x2C
EIP		= 0x30
CS		= 0x34
EFLAGS		= 0x38
OLDESP		= 0x3C
OLDSS		= 0x40

CF_MASK		= 0x00000001
IF_MASK		= 0x00000200
NT_MASK		= 0x00004000
VM_MASK		= 0x00020000

/*
 * these are offsets into the task-struct.
 */
state		=  0
counter		=  4
priority	=  8
signal		= 12
blocked		= 16
flags		= 20
errno		= 24
dbgreg6		= 52
dbgreg7		= 56

ENOSYS = 38

.globl system_call, hd_interrupt, sys_fork
.globl ret_from_sys_call, sys_execve
.extern current

#define SAVE_ALL \
	cld; \
	push 	%gs; \
	push 	%fs; \
	push 	%es; \
	push 	%ds; \
	pushl 	%eax; \
	pushl 	%ebp; \
	pushl 	%edi; \
	pushl 	%esi; \
	pushl 	%edx; \
	pushl 	%ecx; \
	pushl 	%ebx; \
	movl 	$(KERNEL_DS),%edx; \
	mov 	%dx,%ds; \
	mov 	%dx,%es; \
	movl 	$(USER_DS),%edx; \
	mov 	%dx,%fs;

#define RESTORE_ALL \
	cmpw 	$(KERNEL_CS),CS(%esp); \
	je 		1f;   \
	movl 	current,%eax; \
	movl 	dbgreg7(%eax),%ebx; \
	movl 	%ebx,%db7;	\
1:	popl 	%ebx; \
	popl 	%ecx; \
	popl 	%edx; \
	popl 	%esi; \
	popl 	%edi; \
	popl 	%ebp; \
	popl 	%eax; \
	pop 	%ds; \
	pop 	%es; \
	pop 	%fs; \
	pop 	%gs; \
	addl	$4, %esp; \
	iret

/*-------------------------------------------------------------------*
 *                           handle_bottom_half
 *-------------------------------------------------------------------*/
.align 4
handle_bottom_half:
	pushfl
	incl 	intr_count
	sti
	call 	do_bottom_half
	popfl
	decl 	intr_count
	jmp 	9f

/*-------------------------------------------------------------------*
 *                           reschedule
 *-------------------------------------------------------------------*/
.align 4
reschedule:
	pushl 	$ret_from_sys_call
	jmp 	schedule

/*-------------------------------------------------------------------*
 *                           system_call
 *
 * ̺κ   Ǳ ñ ٶϴ.
 *-------------------------------------------------------------------*/
.align 4
system_call:
	pushl 	%eax			# save orig_eax
	SAVE_ALL
	/* Current stack :
	  +-------+
	  | DS    |		+
	  | ESP   |		|	Before interrupt call push DS, ESP, EFLAG, CS
	  | EFLAG |		|	ret_addr in stack
	  | CS    |		|
	  | retadd|		+
	  | EAX   |
	  | GS    |
	  | FS    |
	  | ES    |
	  | DS    |
	  | EAX   |
	  | EBP   |
	  | EDI   |
	  | ESI   |
	  | EDX   |
	  | ECX   |
	  | EBX   |
	  +-------+
	*/
	/* a validity check is then performed on the system call number
	 * passed by ther User Mode process. If it is greater than or equal
	 * to NR_syscalls, the system call handler terminaters. */
	movl 	$-ENOSYS,EAX(%esp)
	cmpl 	NR_syscalls,%eax
	jae 	ret_from_sys_call

	/* ther system_call() function checks whether the PF_TRACESYS flag
	 * included in the flags field of current is equal to 1, that is,
	 * whether the system call invocations of the executed program
	 * are being traced by some debugger. if this is the case, 
	 * system_call() invokes the syscall_trace() function twice,
	 * once right before and once right after the execution of the 
	 * system call service routine. */
	movl 	current,%ebx
	andl 	$~CF_MASK,EFLAGS(%esp)	# clear carry - assume no errors
	movl 	$0,errno(%ebx)
	movl 	%db6,%edx
	movl 	%edx,dbgreg6(%ebx)  # save current hardware debugging status

	/* Finally, the sepcific service routine associated with the system
 	 * call number contained in eax is invoked.
	 * since each entry in the dispatch table is 4 bytes long, the kernel
	 * finds the address of the service routine to be invoked by first
	 * multiplying the system call number by 4, adding the initial 
	 * address of the sys_call_table dispatch table, and extractin a
	 * pointer to the serevice routine from that slot in th table */
	call *sys_call_table(, %eax, 4)

	/* when the service routine terminates, system_call() gets its
	 * return code from eax and stores it in the stack location where
	 * the User Mode value of the eax register has been saved, It then
	 * jumps to ret_from_sys_call(), which terminates the execution of 
	 * the system call handle. */
	movl 	%eax,EAX(%esp)		# save the return value
	movl 	errno(%ebx),%edx
	negl 	%edx
	je 		ret_from_sys_call

	movl 	%edx,EAX(%esp)
	orl 	$(CF_MASK),EFLAGS(%esp)	# set carry to indicate error
	jmp 	ret_from_sys_call

/*-------------------------------------------------------------------*
 *                           ret_from_sys_call
 *-------------------------------------------------------------------*/
	.align 4,0x90
ret_from_sys_call:
	cmpl 	$0,intr_count
	jne 	2f
	/* first, the bh_mask and bh_active variables are checked to determine
	 * whether active unmasked bottom halves exist. if no bottom half 
	 * must be executed, a jump is made to the reschedule address, 
	 * otherwise, the handle_bottom_half fuction is invoked, then control 
	 * is transferred to ret_from_intr. */

	movl 	bh_mask,%eax
	andl 	bh_active,%eax
	jne 	handle_bottom_half

9:	movl 	EFLAGS(%esp),%eax		# check VM86 flag: CS/SS are
	testl 	$(VM_MASK),%eax			# different then
	jne 	1f
	cmpw 	$(KERNEL_CS),CS(%esp)	# was old code segment supervisor ?
	je 		2f
1:	sti
	orl 	$(IF_MASK),%eax		# these just try to make sure
	andl 	$~NT_MASK,%eax		# the program doesn't do anything
	movl 	%eax, EFLAGS(%esp)	# stupid
	cmpl 	$0, need_resched
	jne 	reschedule
	movl 	current, %eax
	# Hmm.. this below line change orginal line are
	# cmpl 	task, %eax
	cmpl 	current_gover , %eax			# task[0] cannot have signals
	je 		2f
	cmpl 	$0,state(%eax)		# state
	jne 	reschedule
	cmpl 	$0, counter(%eax)		# counter
	je 		reschedule
	movl 	blocked(%eax), %ecx
	movl 	%ecx, %ebx			# save blocked in %ebx for signal handling
	notl 	%ecx
	andl 	signal(%eax), %ecx
	jne 	signal_return
2:	
	RESTORE_ALL
/*-------------------------------------------------------------------*
 *                           signal_return
 *-------------------------------------------------------------------*/
.align 4
signal_return:
	movl 	%esp,%ecx
	pushl 	%ecx
	testl 	$(VM_MASK),EFLAGS(%ecx)
	jne 	v86_signal_return
	pushl 	%ebx
	call	where_is2
	# call do_signal
	popl 	%ebx
	popl 	%ebx
	RESTORE_ALL
/*-------------------------------------------------------------------*
 *                           v86_signal_return
 *-------------------------------------------------------------------*/
.align 4
v86_signal_return:
	# call 	_save_v86_state
	movl 	%eax,%esp
	pushl 	%eax
	pushl 	%ebx
	call	where_is2
	# call do_signal
	popl 	%ebx
	popl 	%ebx
	RESTORE_ALL
/*-------------------------------------------------------------------*
 *							sys_execve
 *-------------------------------------------------------------------*/
.align 4
sys_execve:
	lea		EIP(%esp), %eax
	pushl	%eax
	call	do_execve
	addl	$4, %esp
	ret
